/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************/

#include "rt_config.h"

/*  
    ==========================================================================
    Description: 
        association state machine init, including state transition and timer init
    Parameters: 
        S - pointer to the association state machine
    Note:
        The state machine looks like the following 
        
                                    ASSOC_IDLE             
        MT2_MLME_DISASSOC_REQ    mlme_disassoc_req_action 
        MT2_PEER_DISASSOC_REQ    peer_disassoc_action     
        MT2_PEER_ASSOC_REQ       drop                     
        MT2_PEER_REASSOC_REQ     drop                     
        MT2_CLS3ERR              cls3err_action           
    ==========================================================================
 */
VOID AssocStateMachineInit(
    IN	PRTMP_ADAPTER	pAd, 
    IN  STATE_MACHINE *S, 
    OUT STATE_MACHINE_FUNC Trans[]) 
{
    StateMachineInit(S, (STATE_MACHINE_FUNC*)Trans, MAX_ASSOC_STATE, MAX_ASSOC_MSG, (STATE_MACHINE_FUNC)Drop, ASSOC_IDLE, ASSOC_MACHINE_BASE);

    StateMachineSetAction(S, ASSOC_IDLE, MT2_MLME_DISASSOC_REQ, (STATE_MACHINE_FUNC)MlmeDisassocReqAction);
    StateMachineSetAction(S, ASSOC_IDLE, MT2_PEER_DISASSOC_REQ, (STATE_MACHINE_FUNC)PeerDisassocReqAction);
    StateMachineSetAction(S, ASSOC_IDLE, MT2_PEER_ASSOC_REQ, (STATE_MACHINE_FUNC)PeerAssocReqAction);
    StateMachineSetAction(S, ASSOC_IDLE, MT2_PEER_REASSOC_REQ, (STATE_MACHINE_FUNC)PeerReassocReqAction);
//    StateMachineSetAction(S, ASSOC_IDLE, MT2_CLS3ERR, Cls3errAction);
}

/*
    ==========================================================================
    Description:
        peer assoc req handling procedure
    Parameters:
        Adapter - Adapter pointer
        Elem - MLME Queue Element
    Pre:
        the station has been authenticated and the following information is stored
    Post  :
        -# An association response frame is generated and sent to the air
    ==========================================================================
 */
VOID PeerAssocReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR          Addr2;
    MACHDR           AssocRspHdr;
    UCHAR            SsidIe = IE_SSID, RateIe = IE_SUPP_RATES;
    USHORT           ListenInterval;
    USHORT           CapabilityInfo;
    USHORT          StatusCode = 0;
    USHORT          Aid;
    UCHAR           *OutBuffer = NULL;
    NDIS_STATUS      NStatus;
    ULONG            FrameLen = 0;
    char             Ssid[32];
    UCHAR            SsidLen;
    UCHAR            SupportedRatesLen;
    UCHAR            SupportedRates[8];
    MAC_TABLE_ENTRY       *pEntry;
    UCHAR            RSNIE_Len;
    UCHAR            RSN_IE[MAX_LEN_OF_RSNIE];

    // 1. frame sanity check
    if (! PeerAssocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &ListenInterval, &SsidLen, &Ssid[0], &SupportedRatesLen, &SupportedRates[0],RSN_IE, &RSNIE_Len)) 
        return;
        
	// for hidden SSID sake, SSID in AssociateRequest should be fully verified
    if ((SsidLen != pAd->PortCfg.SsidLen) || (NdisEqualMemory(Ssid, pAd->PortCfg.Ssid, SsidLen)==0))
        return;

    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - receive ASSOC request from %02x:%02x:%02x:%02x:%02x:%02x\n",
        Addr2.Octet[0],Addr2.Octet[1],Addr2.Octet[2],Addr2.Octet[3],Addr2.Octet[4],Addr2.Octet[5]);

    // 2. qualify this STA's auth_asoc status in the MAC table, decide StatusCode
    StatusCode = BuildAssociation(pAd, &Addr2, CapabilityInfo, RSN_IE, &RSNIE_Len,&Aid);
    // 3. send Association Response
    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
    if(OutBuffer == NULL)
        return;
        
    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Send ASSOC response (Status=%d)...\n", StatusCode);
    Aid |= 0xc000; // 2 most signifiant bits should be ON
    MgtMacHeaderInit(pAd, &AssocRspHdr, SUBTYPE_ASSOC_RSP, &Addr2);
    MakeOutgoingFrame(OutBuffer, &FrameLen,
                      sizeof(MACHDR), (CHAR *)&AssocRspHdr,
                      2, &CapabilityInfo,                   // capability
                      2, &StatusCode,                       // Status code
                      2, &Aid,                              // AID
                      1, &RateIe,                           // supported rates
                      1, &pAd->PortCfg.SupportedRatesLen,
                      SupportedRatesLen, pAd->PortCfg.SupportedRates,
                      END_OF_ARGS);
    MiniportMMRequest(pAd, OutBuffer, FrameLen);
}

/*
    ==========================================================================
    Description:
        mlme reassoc req handling procedure
    Parameters:
        Elem - 
    Pre:
        -# SSID  (Adapter->PortCfg.ssid[])
        -# BSSID (AP address, Adapter->PortCfg.bssid)
        -# Supported rates (Adapter->PortCfg.supported_rates[])
        -# Supported rates length (Adapter->PortCfg.supported_rates_len)
        -# Tx power (Adapter->PortCfg.tx_power)
    ==========================================================================
 */
VOID PeerReassocReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR          ApAddr, Addr2;
    MACHDR           ReassocRspHdr;
    UCHAR            SsidIe = IE_SSID, RateIe = IE_SUPP_RATES;
    USHORT           CapabilityInfo, ListenInterval;
    USHORT           StatusCode = MLME_SUCCESS;
    USHORT           Aid = 1;
    ULONG            FrameLen = 0;
    NDIS_STATUS      NStatus;
    UCHAR            *OutBuffer = NULL;
    char             Ssid[32];
    UCHAR            SsidLen;
    UCHAR            SupportedRatesLen;
    UCHAR            SupportedRates[8];
    MAC_TABLE_ENTRY       *pEntry;
    UCHAR            RSNIE_Len;
    UCHAR            RSN_IE[MAX_LEN_OF_RSNIE];
   
    // 1. frame sanity check
    if (! PeerReassocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &ListenInterval, &ApAddr, &SsidLen, &Ssid[0], &SupportedRatesLen, &SupportedRates[0],RSN_IE, &RSNIE_Len)) 
        return;
    
    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - receive RE-ASSOC request ...\n");
    
    // 2. qualify this STA's auth_asoc status in the MAC table, decide StatusCode
    StatusCode = BuildAssociation(pAd, &Addr2, CapabilityInfo, RSN_IE, &RSNIE_Len,&Aid);

    // 3. reply Re-association Response
    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
    if(OutBuffer == NULL)
        return;

    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Send RE-ASSOC response (Status = %d)...\n", StatusCode);
    Aid |= 0xc000; // 2 most signifiant bits should be ON
    MgtMacHeaderInit(pAd, &ReassocRspHdr, SUBTYPE_REASSOC_RSP, &Addr2);
    MakeOutgoingFrame(OutBuffer,            &FrameLen,
                      sizeof(MACHDR),       (CHAR *)&ReassocRspHdr,
                      2,                    &CapabilityInfo,          // capability info
                      2,                    &StatusCode,              // status code
                      2,                    &Aid,                     // AID
                      1,                    &RateIe,                  // supported rates
                      1,                    &pAd->PortCfg.SupportedRatesLen, 
                      pAd->PortCfg.SupportedRatesLen,   pAd->PortCfg.SupportedRates,
                      END_OF_ARGS);
    MiniportMMRequest(pAd, OutBuffer, FrameLen);
}

/*
    ==========================================================================
    Description:
        left part of IEEE 802.11/1999 p.374 
    Parameters:
        Elem - MLME message containing the received frame
    ==========================================================================
 */
VOID PeerDisassocReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR       Addr2;
    USHORT        Reason;
    MAC_TABLE_ENTRY       *pEntry;

    if (! PeerDisassocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &Reason)) 
        return;

    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - receive DIS-ASSOC request ...\n");
    
    pEntry = MacTableLookup(pAd, &Addr2);
    if (pEntry)
    {
        ApLogEvent(pAd, &Addr2, EVENT_DISASSOCIATED);
        MacTableDeleteEntry(pAd, &Addr2);
    }
}


/*
    ==========================================================================
    Description:
        Upper layer orders to disassoc s STA
    Parameters:
        Elem -
    ==========================================================================
 */
VOID MlmeDisassocReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MLME_DISASSOC_REQ_STRUCT *DisassocReq;
    MACHDR                DisassocHdr;
    CHAR                 *OutBuffer = NULL;
    ULONG                 FrameLen = 0;
    NDIS_STATUS           NStatus;
    MAC_TABLE_ENTRY       *pEntry;

    DisassocReq = (MLME_DISASSOC_REQ_STRUCT *)(Elem->Msg);

    pEntry = MacTableLookup(pAd, &DisassocReq->Addr);
    if (pEntry)
    {
        ApLogEvent(pAd, &DisassocReq->Addr, EVENT_DISASSOCIATED);
        MacTableDeleteEntry(pAd, &DisassocReq->Addr);
    }

    // 2. send out a DISASSOC request frame
    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
    if(OutBuffer == NULL)
        return;
    
    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - MLME disassociates a STA; Send DISASSOC request\n");
    MgtMacHeaderInit(pAd, &DisassocHdr, SUBTYPE_DISASSOC, &DisassocReq->Addr);
    MakeOutgoingFrame(OutBuffer, &FrameLen, sizeof(MACHDR), (CHAR *)&DisassocHdr, 2, (char*)&DisassocReq->Reason, END_OF_ARGS);
    MiniportMMRequest(pAd, OutBuffer, FrameLen);
}


/*
    ==========================================================================
    Description:
        right part of IEEE 802.11/1999 page 374
    Note: 
        This event should never cause ASSOC state machine perform state
        transition, and has no relationship with CNTL machine. So we separate
        this routine as a service outside of ASSOC state transition table.
    ==========================================================================
 */
VOID Cls3errAction(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR      pAddr) 
{
    MACHDR                DisassocHdr;
    CHAR                 *OutBuffer = NULL;
    ULONG                 FrameLen = 0;
    NDIS_STATUS           NStatus;
    USHORT                Reason = REASON_CLS3ERR;
    MAC_TABLE_ENTRY       *pEntry;

    pEntry = MacTableLookup(pAd, pAddr);
    if (pEntry)
    {
        //ApLogEvent(pAd, &pAddr, EVENT_DISASSOCIATED);
        MacTableDeleteEntry(pAd, pAddr);
    }
    
    // 2. send out a DISASSOC request frame
    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
    if(OutBuffer == NULL)
        return;
    
    DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Class 3 Error, Send DISASSOC frame\n");
    MgtMacHeaderInit(pAd, &DisassocHdr, SUBTYPE_DISASSOC, pAddr);
    MakeOutgoingFrame(OutBuffer, &FrameLen, sizeof(MACHDR), (CHAR *)&DisassocHdr, 2, (char*)&Reason, END_OF_ARGS);
    MiniportMMRequest(pAd, OutBuffer, FrameLen);
}
 
/*
    ==========================================================================
    Description:
        procedures on IEEE 802.11/1999 p.376 
    Parametrs:
    ==========================================================================
 */
VOID AssocPostProc(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR pAddr2, 
    IN USHORT CapabilityInfo, 
    IN USHORT Aid, 
    IN UCHAR Rates[], 
    IN UCHAR RatesLen) 
{
    MAC_TABLE_ENTRY *pEntry;

    // update MAC table
    pEntry = MacTableLookup(pAd, pAddr2);
    if (! pEntry)
        pEntry = MacTableInsertEntry(pAd, pAddr2);
    
    if (pEntry)
    {
        pEntry->Sst = SST_ASSOC;
        pEntry->Rate = RATE_11;
        pEntry->Aid = Aid;
        ApLogEvent(pAd, pAddr2, EVENT_ASSOCIATED);
    }
    
}

// search for entire MAC table to find next available AID to be used
// if none is available, return 0
USHORT AssignAid(
    IN PRTMP_ADAPTER pAd)
{
    USHORT Aid=0, i;
    UCHAR AidUsed[MAX_LEN_OF_MAC_TABLE];

    // find all used AID    
    NdisZeroMemory(AidUsed, MAX_LEN_OF_MAC_TABLE);
    for (i=0;i<MAX_LEN_OF_MAC_TABLE;i++)
        if (pAd->MacTab.Content[i].Valid)
            AidUsed[pAd->MacTab.Content[i].Aid] = 1;

    // find smallest unused AID
    for (i=1; i<MAX_LEN_OF_MAC_TABLE; i++)
        if (AidUsed[i]==0)
        {
            Aid = i;
            break;
        }

    return Aid;
}

/*
    ==========================================================================
    Description:
       assign a new AID to the newly associated/re-associated STA and
       decide its MaxSupportedRate and CurrTxRate. Both rates should not
       exceed AP's capapbility
    Return:
       MLME_SUCCESS - association successfully built
       others - association failed due to resource issue
    ==========================================================================
 */
USHORT BuildAssociation(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR      pAddr,
    IN USHORT        CapabilityInfo,
    IN UCHAR         *RSN,
    IN UCHAR         *pRSNLen,
    OUT USHORT       *pAid)
{
    USHORT           StatusCode = MLME_SUCCESS;
    PMAC_TABLE_ENTRY pEntry;
    UCHAR            MaxSupportedRate = RATE_11;

    pEntry = MacTableLookup(pAd, pAddr);
    if (pEntry && ((pEntry->Sst == SST_AUTH) || (pEntry->Sst == SST_ASSOC)))
    {
        // TODO:
        // should qualify other parameters, for example - capablity, supported rates, listen interval, ... etc
        // to decide the Status Code
        *pAid = AssignAid(pAd);
        pEntry->Aid = *pAid;
        
        NdisMoveMemory(pEntry->RSN_IE, RSN, *pRSNLen);
        pEntry->RSNIE_Len = *pRSNLen;
        pEntry->Flag = 0;

        if (*pAid == 0)
            StatusCode = MLME_ASSOC_REJ_UNABLE_HANDLE_STA;
        else
        {
            pEntry->Sst = SST_ASSOC;
            pEntry->CapabilityInfo = CapabilityInfo;
            if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPAPSK)
            {
                pEntry->PrivacyFilter = Ndis802_11PrivFilter8021xWEP;
                pEntry->WpaState = AS_INITPSK;
            }
            else if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPA)
            {
                pEntry->PrivacyFilter = Ndis802_11PrivFilter8021xWEP;
                pEntry->WpaState = AS_AUTHENTICATION;
            }
            pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
            StatusCode = MLME_SUCCESS;
            ApLogEvent(pAd, &pEntry->Addr, EVENT_ASSOCIATED);
            if ((pEntry->RetryTimerRunning == FALSE) && (pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPAPSK))
            {               // retry up to 3 times at interval 1 second
                DBGPRINT(RT_DEBUG_TRACE, "add_timer  Timer   ====>>>   \n");
                pEntry->RetryTimerRunning = TRUE;
                init_timer(&pEntry->RetryTimer);
                pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
                pEntry->RetryTimer.data = (unsigned long)pAd;
                pEntry->RetryTimer.function = &WPARetryExec;
                add_timer(&pEntry->RetryTimer);
                pEntry->ReTryCounter = PAIR_HS_RETRY_TIMER_CTR;
            }
            DBGPRINT(RT_DEBUG_TRACE,"assign AID=%d to STA %02x:%02x:%02x:%02x:%02x:%02x, MaxSupportedRate=%d Mbps, CurrTxRate=%d Mbps\n", 
                *pAid, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3],
                pAddr->Octet[4], pAddr->Octet[5], RateIdToMbps[pEntry->MaxSupportedRate], RateIdToMbps[pEntry->CurrTxRate]);
            DBGPRINT(RT_DEBUG_TRACE, "RSNIE_Len = 0x%x    pEntry->RSNIE_Len = %d \n\n ",*pRSNLen,pEntry->RSNIE_Len);
            DBGPRINT(RT_DEBUG_TRACE, "pEntry->PrivacyFilter = %x \n\n ",pEntry->PrivacyFilter );

        }
    }
    else // CLASS 3 error should have been handled beforehand; here should be MAC table full
        StatusCode = MLME_ASSOC_REJ_UNABLE_HANDLE_STA;
    return StatusCode;
}


