/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************/

#include "rt_config.h"
#include <stdarg.h>

UCHAR RateIdToMbps[]     = { 1, 2, 5, 11, 6, 9, 12, 18, 24, 36, 48, 54, 72, 100};
USHORT RateIdTo500Kbps[] = {2, 4, 11, 22, 12, 18, 24, 36, 48, 72, 96, 108, 144, 200};

/*
    ==========================================================================
    Description:
        initialize the MLME task and its data structure (queue, spinlock, 
        timer, state machines).
    Return:
        always return NDIS_STATUS_SUCCESS
    ==========================================================================
*/
NDIS_STATUS MlmeInit(
    IN PRTMP_ADAPTER pAd) 
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    DBGPRINT(RT_DEBUG_TRACE, "--> MLME Initialize\n");
    
    do 
    {
        Status = MlmeQueueInit(&pAd->Mlme.Queue);
        if(Status != NDIS_STATUS_SUCCESS) 
        {
            break;
        }

        spin_lock_init(&pAd->Mlme.TaskLock);
        pAd->Mlme.Running = FALSE;
        pAd->Mlme.TimBitmap = 0;
        pAd->Mlme.DtimCount = 0;
        pAd->Mlme.DtimPeriod = 3;

        // init state machines
        AssocStateMachineInit(pAd, &pAd->Mlme.AssocMachine, pAd->Mlme.AssocFunc);
        AuthStateMachineInit(pAd, &pAd->Mlme.AuthMachine, pAd->Mlme.AuthFunc);
        AuthRspStateMachineInit(pAd, &pAd->Mlme.AuthRspMachine, pAd->Mlme.AuthRspFunc);
        SyncStateMachineInit(pAd, &pAd->Mlme.SyncMachine, pAd->Mlme.SyncFunc);
	WpaStateMachineInit(pAd,&pAd->Mlme.WpaMachine,pAd->Mlme.WpaFunc);

        // Since we are using switch/case to implement it, the init is different from the above 
        // state machine init
        // MlmeCntlInit(pAd, &pAd->Mlme.CntlMachine, NULL);

        // Regularly check the timer
        init_timer(&pAd->Mlme.PeriodicTimer);
		pAd->Mlme.PeriodicTimer.expires = jiffies + MLME_TASK_EXEC_INTV;	// the time, in jiffies
		pAd->Mlme.PeriodicTimer.data = (unsigned long)pAd;
		pAd->Mlme.PeriodicTimer.function = &MlmePeriodicExec;				/* timer handler */
		add_timer(&pAd->Mlme.PeriodicTimer);

        if (pAd->PortCfg.LedMode == LED_MODE_TXRX_ACTIVITY)
        {
        	init_timer(&pAd->PortCfg.LedCntl.BlinkTimer);
		pAd->PortCfg.LedCntl.BlinkTimer.expires = jiffies + (70 * HZ)/1000;	// the time, in jiffies
		pAd->PortCfg.LedCntl.BlinkTimer.data = (unsigned long)pAd;
		pAd->PortCfg.LedCntl.BlinkTimer.function = &AsicLedPeriodicExec;				/* timer handler */
		add_timer(&pAd->PortCfg.LedCntl.BlinkTimer);
        }

        // for dynamic "R13: RX AGC VGC INIT" control
        {
            UCHAR FalseCcaHigh,FalseCcaLow;
            pAd->PortCfg.RxVgc.Duration = 1;
            pAd->PortCfg.RxVgc.FlaseCcaThreshold = 8;
            pAd->PortCfg.RxVgc.RxAgcVgcDelta = 2;
            pAd->PortCfg.RxVgc.MaxRxAgcVgc = 0x20;
            RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 39, &FalseCcaHigh);
//          RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 40, &FalseCcaLow);
            pAd->PortCfg.BbpFalseCca = FalseCcaHigh; //(FalseCcaHigh << 8) + FalseCcaLow;
        }
    } while (FALSE);

    DBGPRINT(RT_DEBUG_TRACE, "<-- MLME Initialize\n");

    return Status;
}

/*
    ==========================================================================
    Description:
        main loop of the MLME
    Pre:
        Mlme has to be initialized, and there are something inside the queue
    Note:
        This function is invoked from MPSetInformation and MPReceive;
        This task guarantee only one MlmeHandler will run. 
    ==========================================================================
 */
VOID MlmeHandler(
    IN PRTMP_ADAPTER pAd) 
{
    NDIS_STATUS             NStatus;
    MLME_QUEUE_ELEM        *Elem = NULL;

    // Only accept MLME and Frame from peer side, no other (control/data) frame should
    // get into this state machine

    RTMPAcquireSpinLock(pAd, &pAd->Mlme.TaskLock);
    if(pAd->Mlme.Running) 
    {
        RTMPFreeSpinLock(pAd, &pAd->Mlme.TaskLock);
        return;
    } 
    else 
    {
        pAd->Mlme.Running = TRUE;
    }
    RTMPFreeSpinLock(pAd, &pAd->Mlme.TaskLock);

    while (!MlmeQueueEmpty(pAd, &pAd->Mlme.Queue)) 
    {
        //From message type, determine which state machine I should drive
        if (MlmeDequeue(pAd, &pAd->Mlme.Queue, &Elem)) 
        {
            // if dequeue success
            switch (Elem->Machine) 
            {
                case ASSOC_STATE_MACHINE:
                    StateMachinePerformAction(pAd, &pAd->Mlme.AssocMachine, Elem);
                    break;
                case AUTH_STATE_MACHINE:
                    StateMachinePerformAction(pAd, &pAd->Mlme.AuthMachine, Elem);
                    break;
                case AUTH_RSP_STATE_MACHINE:
                    StateMachinePerformAction(pAd, &pAd->Mlme.AuthRspMachine, Elem);
                    break;
                case SYNC_STATE_MACHINE:
                    StateMachinePerformAction(pAd, &pAd->Mlme.SyncMachine, Elem);
                    break;
//              case MLME_CNTL_STATE_MACHINE:
//                  MlmeCntlMachinePerformAction(pAd, &pAd->Mlme.CntlMachine, Elem);
//                  break;
                case WPA_STATE_MACHINE:
                    StateMachinePerformAction(pAd, &pAd->Mlme.WpaMachine, Elem);
                    break;
                default:
                    DBGPRINT(RT_DEBUG_TRACE, "ERROR: Illegal machine in MlmeHandler()\n");
                    break;
            } // end of switch

            // free MLME element
            Elem->Occupied = FALSE;
            Elem->MsgLen = 0;
            
        }
        else {
            DBGPRINT(RT_DEBUG_ERROR, "ERROR: empty Elem in MlmeQueue\n");
        }
    }

    RTMPAcquireSpinLock(pAd, &pAd->Mlme.TaskLock);
    pAd->Mlme.Running = FALSE;
    RTMPFreeSpinLock(pAd, &pAd->Mlme.TaskLock);

}

/*
    ==========================================================================
    Description:
        Destructor of MLME (Destroy queue, state machine, spin lock and timer)
    Parameters:
        Adapter - NIC Adapter pointer
    Post:
        The MLME task will no longer work properly
    ==========================================================================
 */
VOID MlmeHalt(
    IN PRTMP_ADAPTER pAd) 
{
    BOOLEAN       Cancelled;

    DBGPRINT(RT_DEBUG_TRACE, "==> MlmeHalt\n");

    if (pAd->PortCfg.LedMode == LED_MODE_TXRX_ACTIVITY)
        del_timer_sync(&pAd->PortCfg.LedCntl.BlinkTimer);
    
    MlmeQueueDestroy(&pAd->Mlme.Queue);
//    NdisFreeSpinLock(&pAd->Mlme.TaskLock);
    del_timer_sync(&pAd->Mlme.PeriodicTimer);

	DBGPRINT(RT_DEBUG_TRACE, "<== MlmeHalt\n");
}


/*
    ==========================================================================
    Description:
        Conforirm function for out MLME packets
    Parameters:
        Adapter - NIC Adapter pointer
        Msg     - Message returned from ASIC
        MsgLen  - message length, not used in this implementation
    Pre:
        The Msg has to be allocated by NdisAllocateMemoryWithTag
    ==========================================================================
 */
VOID MlmeConfirm(
    IN PRTMP_ADAPTER pAd, 
    IN VOID *Msg, 
    IN INT MsgLen) 
{
    kfree(Msg);
}


/*
    ==========================================================================
    Description:
        This routine is executed periodically to -
        1. Decide if it's a right time to turn on PwrMgmt bit of all 
           outgoiing frames
        2. Calculate FoundedRate, and RoamCqi based on statistics of the last
           period. FoundedRate is used to constain actual TX rate, so that
           TX rate won't toggling very frequently between a successful TX 
           and a failed TX (rate too high).
        3. If the calculated RoamCqi indicated current connection not healthy,
           Then a ROAMing attempt is tried here.
    ==========================================================================
 */
VOID MlmePeriodicExec(
    IN unsigned long data)
{
    RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;
    MAC_TABLE_ENTRY *pEntry;
    int i;
    ULONG Now32;

    //
    // dynamic tune BBP R13: RX_AGC_VGC_INIT based on false CCA	
    //
    if (pAd->PortCfg.EnableRxAgcVgcTuning &&
        (pAd->Mlme.PeriodicRound < 60) &&       // turn off R13 tuning after 1 minute
        pAd->PortCfg.RxVgc.FlaseCcaThreshold && 
        pAd->PortCfg.RxVgc.RxAgcVgcDelta &&
        ((pAd->Mlme.PeriodicRound % pAd->PortCfg.RxVgc.Duration)==0))
    {
        UCHAR FalseCcaLow, FalseCcaHigh;
        USHORT FalseCca;
        LONG FalseCcaDelta;

        RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 39, &FalseCcaHigh);
//      RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 40, &FalseCcaLow);
        FalseCca = FalseCcaHigh; //      FalseCca = (FalseCcaHigh << 8) + FalseCcaLow;
        FalseCcaDelta = FalseCca - pAd->PortCfg.BbpFalseCca;
        pAd->PortCfg.BbpFalseCca = FalseCca;
        DBGPRINT(RT_DEBUG_TRACE,"FalseCca(R39)=%d, delta=%d\n",FalseCca,FalseCcaDelta);
        if (FalseCcaDelta >= pAd->PortCfg.RxVgc.FlaseCcaThreshold)
        {
            UCHAR R13;
            RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 13, &R13);
            R13 += pAd->PortCfg.RxVgc.RxAgcVgcDelta;
            if (R13 <= pAd->PortCfg.RxVgc.MaxRxAgcVgc)
    {
                RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 13, R13);
                DBGPRINT(RT_DEBUG_TRACE,"FalseCcaDelta=%d, >=%d, set R13=%d\n", 
                    FalseCcaDelta, pAd->PortCfg.RxVgc.FlaseCcaThreshold, R13);
            }
        }
    }
    
    //
    // upate channle quality, RSSI, ..etc
    //
    if (pAd->Mlme.PeriodicRound % 2)
    {
        Now32 = jiffies;
        MlmeCheckChannelQuality(pAd, Now32);
    }

    //
    // MAC table maintenance
    //
    pAd->MacTab.fAnyStationInPsm = FALSE;
    for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++) 
    {
        MAC_TABLE_ENTRY *pEntry = &pAd->MacTab.Content[i];

        if (pEntry->Valid)
        {
            // 1. check if there's any associated STA in power-save mode. this affects outgoing
            //    MCAST/BCAST frames should be stored in PSQ till DtimCount=0
            if (pEntry->PsMode == PWR_SAVE)
                pAd->MacTab.fAnyStationInPsm = TRUE;
            
            // 2. delete those MAC entry that has been idle for a long time
            pEntry->NoDataIdleCount ++;  
            if (pEntry->NoDataIdleCount >= MAC_TABLE_AGEOUT_TIME)
            {
                DBGPRINT(RT_DEBUG_TRACE, "ageout WSTA %02x:%02x:%02x:%02x:%02x:%02x after %d-sec silence\n",
                    pEntry->Addr.Octet[0],pEntry->Addr.Octet[1],pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],
                    pEntry->Addr.Octet[4],pEntry->Addr.Octet[5],MAC_TABLE_AGEOUT_TIME);
                ApLogEvent(pAd, &pEntry->Addr, EVENT_AGED_OUT);
                MacTableDeleteEntry(pAd, &pEntry->Addr);
            }
            
            // 3. garbage collect the PsQueue if the STA has being idle for a while
            if (pEntry->PsQueue.Head)
            {
                pEntry->PsQIdleCount ++;  
                if (pEntry->PsQIdleCount > 2) 
                {
                    RTMPAcquireSpinLock(pAd, &pAd->MacTabLock);
                    CleanupPsQueue(pAd, &pEntry->PsQueue);
                    RTMPFreeSpinLock(pAd, &pAd->MacTabLock);
                    pEntry->PsQIdleCount = 0;
                }
            }
            else
                pEntry->PsQIdleCount = 0;
        }
    }

    // 4. garbage collect pAd->MacTab.McastPsQueue if backlogged MCAST/BCAST frames
    //    stale in queue. Since MCAST/BCAST frames always been sent out whenever 
    //    DtimCount==0, the only case to let them stale is surprise removal of the NIC,
    //    so that ASIC-based Tbcn interrupt stops and DtimCount dead.
    if (pAd->MacTab.McastPsQueue.Head)
    {
        pAd->MacTab.PsQIdleCount ++;
        if (pAd->MacTab.PsQIdleCount > 1)
        {
            RTMPAcquireSpinLock(pAd, &pAd->MacTabLock);
            CleanupPsQueue(pAd, &pAd->MacTab.McastPsQueue);
            RTMPFreeSpinLock(pAd, &pAd->MacTabLock);
            pAd->MacTab.PsQIdleCount = 0;
        }
    }
    else
        pAd->MacTab.PsQIdleCount = 0;
    
    pAd->Mlme.PeriodicRound ++;

    pAd->Mlme.PeriodicTimer.expires = jiffies + MLME_TASK_EXEC_INTV;	// the time, in jiffies
	add_timer(&pAd->Mlme.PeriodicTimer);
}

/*
    ==========================================================================
    Description:
        This routine calculates TxPER, RxPER of the past N-sec period. And 
        according to the calculation result, change PortCfg.FoundedRate which 
        is the stable TX Rate we expect the Radio situation could sustained. 
        RoamCqi is also calculated here to decide if current AP is still doing 
        the job. If RoamCqi is not good, a ROAMing attempt may be tried later.

        PortCfg.TxRate will change dynamically within {RATE_1, FoundedRate} 
        whenever a TxResult (SUCC, RETRY, or FAIL} is obtained. In comparasion
        to that, FoundedRate changes not that fast within {RATE_1..RATE_11} 
        based on TxPER and RxPER for the past N-sec period.
    Output:
        PortCfg.RoamCqi - 0..100
        PortCfg.DrsCqi  - 0..100
        PortCfg.FoundedRate - RATE_1, RATE_2, RATE_5_5, RATE_11
    ==========================================================================
 */
VOID MlmeCheckChannelQuality(
    IN PRTMP_ADAPTER pAd,
    IN ULONG Now32)
{
    ULONG RtsFailCnt, RtsOkCnt, TxFailCnt, TxOkCnt, TxOneRetryCnt, TxMRetryCnt, TxCnt, TxPER;
    ULONG RxFailCnt, RxOkCnt, RxCnt, RxPER, TxPRR;

    //
    // monitor TX counters change for the past period
    //
    TxFailCnt     = pAd->WlanCounters.FailedCount - 
                    pAd->Mlme.PrevWlanCounters.FailedCount;
    RtsFailCnt    = pAd->WlanCounters.RTSFailureCount - 
                    pAd->Mlme.PrevWlanCounters.RTSFailureCount;
    TxOneRetryCnt = pAd->WlanCounters.RetryCount - 
                    pAd->Mlme.PrevWlanCounters.RetryCount;
    TxMRetryCnt   = pAd->WlanCounters.MultipleRetryCount - 
                    pAd->Mlme.PrevWlanCounters.MultipleRetryCount;
    TxOkCnt       = pAd->WlanCounters.TransmittedFragmentCount - 
                    pAd->Mlme.PrevWlanCounters.TransmittedFragmentCount;
    TxCnt = TxOkCnt + TxFailCnt;

    if (TxCnt)    
        pAd->Mlme.TxPER = (TxFailCnt * 100) / TxCnt;
    else
        pAd->Mlme.TxPER = 0;
    
    if (TxCnt < 10) // if less than 10 TX sample, skip TX related statistics
    {
        TxPER = 0;  // don't take TxPER into CQI consideration if too few sample
        TxPRR = 0;
    }
    else 
    {
        TxPER = pAd->Mlme.TxPER;
        TxPRR = ((TxOneRetryCnt + TxMRetryCnt + TxFailCnt) * 100) / TxCnt;
    }

    //
    // calculate RX PER
    //
    RxOkCnt   = pAd->WlanCounters.ReceivedFragmentCount - 
                pAd->Mlme.PrevWlanCounters.ReceivedFragmentCount;
    RxFailCnt = pAd->WlanCounters.FCSErrorCount - 
                pAd->Mlme.PrevWlanCounters.FCSErrorCount;
    RxCnt = RxOkCnt + RxFailCnt;

    if (RxCnt)
        pAd->Mlme.RxPER = (RxFailCnt * 100) / RxCnt;
    else
        pAd->Mlme.RxPER = 0;
    
    if (RxCnt < 10)
        RxPER = 0;  // don't take RxPER into CQI consideration if too few sample
    else
        RxPER = pAd->Mlme.RxPER;

    //
    // decide CQI based on: 1)last BEACON received time, 2)last RSSI, 3)TxPER, and 4)RxPER
    //
    // This value also decides when all roaming fails (or no roaming candidates at 
    // all), should this STA stay with original AP, or a LinkDown signal 
    // is indicated to NDIS
    //
    {
        // RoamCQI = W1*RSSI + W2*TxPRR + W3*RxPER    (RSSI 0..100), (TxPER 100..0), (RxPER 100..0)
        // DrsCQI  = W1*RSSI + W2*TxPRR + W3*RxPER
        pAd->Mlme.RoamCqi = (RSSI_WEIGHTING * pAd->PortCfg.LastRssi + 
                             TX_WEIGHTING * (100 - TxPRR) + 
                             RX_WEIGHTING* (100 - RxPER)) / 100;
        if (pAd->Mlme.RoamCqi >= 100)
            pAd->Mlme.RoamCqi = 100;
        pAd->Mlme.DrsCqi  = pAd->Mlme.RoamCqi;
    }
    
    DBGPRINT(RT_DEBUG_INFO, "MMCHK - CQI= %d, (Tx=%d/%d/%d, Rx=%d/%d, RSSI=%d)\n", 
        pAd->Mlme.RoamCqi, TxFailCnt, TxOneRetryCnt+TxMRetryCnt, TxCnt, RxFailCnt, RxCnt, pAd->PortCfg.LastRssi);

    // latch current WLAN counters for next check-for-roaming usage
    NdisMoveMemory(&pAd->Mlme.PrevWlanCounters, &pAd->WlanCounters, sizeof(COUNTER_802_11));

}

VOID MlmeSetTxPreamble(
    IN PRTMP_ADAPTER pAd, 
    IN USHORT TxPreamble)
{
    ULONG Arcsr2 = 0x00700400;     // 0x13c, ACK/CTS PLCP at 1 Mbps
    ULONG Arcsr3 = 0x00380401;     // 0x140, ACK/CTS PLCP at 2 Mbps
    ULONG Arcsr4 = 0x00150402;     // 0x144, ACK/CTS PLCP at 5.5 Mbps
    ULONG Arcsr5 = 0x000b8403;     // 0x148, ACK/CTS PLCP at 11 Mbps
    
    if (TxPreamble == Rt802_11PreambleShort)
    {
        DBGPRINT(RT_DEBUG_TRACE, "---> set to SHORT PREAMBLE\n");
        Arcsr2 |= 0x00000008;
        Arcsr3 |= 0x00000008;
        Arcsr4 |= 0x00000008;
        Arcsr5 |= 0x00000008;
        pAd->PortCfg.TxPreamble = Rt802_11PreambleShort;
        pAd->PortCfg.Dpreamble = 72;
        pAd->PortCfg.Dplcp = 24;
    }
    else
    {
        DBGPRINT(RT_DEBUG_TRACE, "---> set to LONG PREAMBLE\n");
        pAd->PortCfg.TxPreamble = Rt802_11PreambleLong;
        pAd->PortCfg.Dpreamble = 144;
        pAd->PortCfg.Dplcp = 48;
    }

    RTMP_IO_WRITE32(pAd, ARCSR2, Arcsr2);
    RTMP_IO_WRITE32(pAd, ARCSR3, Arcsr3);
    RTMP_IO_WRITE32(pAd, ARCSR4, Arcsr4);
    RTMP_IO_WRITE32(pAd, ARCSR5, Arcsr5);
}
    
VOID MlmeUpdateTxRates(
    IN PRTMP_ADAPTER pAd)
{
    int i;
    UCHAR Rate, Desire = RATE_1, Support = RATE_1, MaxBasicRate = RATE_1;

    // find max desired rate
    for (i=0; i<MAX_LEN_OF_SUPPORTED_RATES; i++)
    {
        switch (pAd->PortCfg.DesiredRates[i] & 0x7f)
        {
            case 2:  Rate = RATE_1;   break;
            case 4:  Rate = RATE_2;   break;
            case 11: Rate = RATE_5_5; break;
            case 22: Rate = RATE_11;  break;
            default: Rate = RATE_1;   break;
        }
        if (Desire < Rate)
            Desire = Rate;
    }

    // find max supported rate
    for (i=0; i<MAX_LEN_OF_SUPPORTED_RATES; i++)
    {
        switch (pAd->PortCfg.SupportedRates[i] & 0x7f)
        {
            case 2:  Rate = RATE_1;   break;
            case 4:  Rate = RATE_2;   break;
            case 11: Rate = RATE_5_5; break;
            case 22: Rate = RATE_11;  break;
            default: Rate = RATE_1;   break;
        }
        if (Support < Rate)
            Support = Rate;
        if ((pAd->PortCfg.SupportedRates[i] & 0x80) && (MaxBasicRate < Rate))
            MaxBasicRate = Rate;
    }

    // max tx rate = min {max desire rate, max supported rate}
    if (Support < Desire)
        Rate = Support;
    else
        Rate = Desire;

    pAd->PortCfg.MaxTxRate = Rate;
    pAd->PortCfg.TxRate = Rate;
    pAd->PortCfg.MaxBasicRate = MaxBasicRate;

    switch (pAd->PortCfg.MaxBasicRate)
    {
    	case RATE_1:
    		RTMP_IO_WRITE32(pAd, TXCSR1, 0x016275ff);	// ACK as 1Mb time
    		RTMP_IO_WRITE32(pAd, ARCSR1, 0x00000001);   // basic rate set {1} Mbps
			break;
    		
    	case RATE_2:
    		RTMP_IO_WRITE32(pAd, TXCSR1, 0x016205ff);	// ACK as 2Mb time
    		RTMP_IO_WRITE32(pAd, ARCSR1, 0x00000003);   // basic rate set {1, 2} Mbps
			break;
			
    	case RATE_5_5:
    		RTMP_IO_WRITE32(pAd, TXCSR1, 0x0161bfff);	// ACK as 5.5Mb time
    		RTMP_IO_WRITE32(pAd, ARCSR1, 0x00000007);   // basic rate set {1, 2, 5.5} Mbps
			break;

    	case RATE_11:
    		RTMP_IO_WRITE32(pAd, TXCSR1, 0x0161abff);	// ACK as 11Mb time
    		RTMP_IO_WRITE32(pAd, ARCSR1, 0x0000000f);   // basic rate set {1, 2, 5.5, 11} Mbps
			break;    		
    }

    DBGPRINT(RT_DEBUG_TRACE, "MMCHK - update (MaxTxRate=%d, CurrTxRate=%d, MaxBasicRate=%d)\n",
        Rate, Rate, MaxBasicRate);
}

/*! \brief init the management mac frame header
 *  \param p_hdr mac header
 *  \param subtype subtype of the frame
 *  \param p_ds destination address, don't care if it is a broadcast address
 *  \return none
 *  \pre the station has the following information in the pAd->PortCfg
 *   - bssid
 *   - station address
 *  \post
 *  \note this function initializes the following field
 */
VOID MgtMacHeaderInit(
    IN	PRTMP_ADAPTER	pAd, 
    IN OUT PMACHDR Hdr, 
    IN UCHAR Subtype, 
    IN PMACADDR pAddr1) 
{
    NdisZeroMemory(Hdr, sizeof(MACHDR));
    Hdr->Type = BTYPE_MGMT;
    Hdr->SubType = Subtype;
    COPY_MAC_ADDR(&Hdr->Addr1, pAddr1);
    COPY_MAC_ADDR(&Hdr->Addr2, &pAd->CurrentAddress);
    COPY_MAC_ADDR(&Hdr->Addr3, &pAd->PortCfg.Bssid);
}

// ===========================================================================================
// mem_mgmt.c
// ===========================================================================================

/*!***************************************************************************
 * This routine build an outgoing frame, and fill all information specified 
 * in argument list to the frame body. The actual frame size is the summation 
 * of all arguments.
 * input params:
 *      Buffer - pointer to a pre-allocated memory segment
 *      args - a list of <int arg_size, arg> pairs.
 *      NOTE NOTE NOTE!!!! the last argument must be NULL, otherwise this
 *                         function will FAIL!!!
 * return:
 *      Size of the buffer
 * usage:  
 *      MakeOutgoingFrame(Buffer, 2, &fc, 2, &dur, 6, p_addr1, 6,p_addr2, NULL);
 ****************************************************************************/
ULONG MakeOutgoingFrame(
    OUT CHAR *Buffer, 
    OUT ULONG *FrameLen, ...) 
{
    CHAR   *p;
    int     leng, i;
    ULONG   TotLeng;
    va_list Args;

    // calculates the total length
    TotLeng = 0;
    va_start(Args, FrameLen);
    do 
    {
        leng = va_arg(Args, int);
        if (leng == END_OF_ARGS) 
        {
            break;
        }
        p = va_arg(Args, PVOID);
        NdisMoveMemory(&Buffer[TotLeng], p, leng);
        TotLeng = TotLeng + leng;
    } while(TRUE);

    va_end(Args); /* clean up */
    *FrameLen = TotLeng;
    return TotLeng;
}

// ===========================================================================================
// mlme_queue.c
// ===========================================================================================

/*! \brief  Initialize The MLME Queue, used by MLME Functions
 *  \param  *Queue     The MLME Queue
 *  \return Always     Return NDIS_STATE_SUCCESS in this implementation
 *  \pre
 *  \post
 *  \note   Because this is done only once (at the init stage), no need to be locked
 */
NDIS_STATUS MlmeQueueInit(
    IN MLME_QUEUE *Queue) 
{
    INT i;

    spin_lock_init(&Queue->Lock);

    Queue->Num  = 0;
    Queue->Head = 0;
    Queue->Tail = 0;

    for (i = 0; i < MAX_LEN_OF_MLME_QUEUE; i++) 
    {
        Queue->Entry[i].Occupied = FALSE;
        Queue->Entry[i].MsgLen = 0;
        NdisZeroMemory(Queue->Entry[i].Msg, MAX_LEN_OF_MLME_QUEUE_MSG);
    }

    return NDIS_STATUS_SUCCESS;
}


/*! \brief   Enqueue a message for other threads, if they want to send messages to MLME thread
 *  \param  *Queue    The MLME Queue
 *  \param   Machine  The State Machine Id
 *  \param   MsgType  The Message Type
 *  \param   MsgLen   The Message length
 *  \param  *Msg      The message pointer
 *  \return  TRUE if enqueue is successful, FALSE if the queue is full
 *  \pre
 *  \post
 *  \note    The message has to be initialized
 */
BOOLEAN MlmeEnqueue(
    IN	PRTMP_ADAPTER	pAd,
    OUT MLME_QUEUE *Queue, 
    IN ULONG Machine, 
    IN ULONG MsgType, 
    IN ULONG MsgLen, 
    IN VOID *Msg) 
{
    INT Tail;

    if (MlmeQueueFull(pAd, Queue)) 
    {
        DBGPRINT(RT_DEBUG_ERROR, "MlmeEnqueue full, msg dropped and may corrupt MLME\n");
        return FALSE;
    }

    RTMPAcquireSpinLock(pAd, &(Queue->Lock));
    Tail = Queue->Tail;
    Queue->Tail++;
    Queue->Num++;
    if (Queue->Tail == MAX_LEN_OF_MLME_QUEUE) 
    {
        Queue->Tail = 0;
    }
    RTMPFreeSpinLock(pAd, &(Queue->Lock));
    DBGPRINT(RT_DEBUG_INFO, "MlmeEnqueue, num=%d\n",Queue->Num);
    
    Queue->Entry[Tail].Occupied = TRUE;
    Queue->Entry[Tail].Machine = Machine;
    Queue->Entry[Tail].MsgType = MsgType;
    Queue->Entry[Tail].MsgLen  = MsgLen;
    NdisMoveMemory(Queue->Entry[Tail].Msg, Msg, MsgLen);
    return TRUE;
}

/*! \brief   This function is used when Recv gets a MLME message
 *  \param  *Queue           The MLME Queue
 *  \param   TimeStampHigh   The upper 32 bit of timestamp
 *  \param   TimeStampLow    The lower 32 bit of timestamp
 *  \param   Rssi            The receiving RSSI strength
 *  \param   MsgLen          The length of the message
 *  \param  *Msg             The message pointer
 *  \return  TRUE if everything ok, FALSE otherwise (like Queue Full)
 *  \pre
 *  \post
 */
BOOLEAN MlmeEnqueueForRecv(
    IN	PRTMP_ADAPTER	pAd, 
    OUT MLME_QUEUE *Queue, 
//  IN ULONG TimeStampHigh, 
//  IN ULONG TimeStampLow,
    IN UCHAR Rssi, 
    IN ULONG MsgLen, 
    IN VOID *Msg) 
{
    INT          Tail, Machine;
    MACFRAME    *Fr = (MACFRAME *)Msg;
    USHORT       Seq;
    BOOLEAN      Result;
    ULONG        MsgType;

    if (MlmeQueueFull(pAd, Queue) || Fr->Hdr.Type == BTYPE_CNTL) 
    {
        DBGPRINT(RT_DEBUG_ERROR, "MlmeEnqueueForRecv (queue full error) \n");
        return FALSE;
    }

    if (!MsgTypeSubst(Fr, &Machine, &MsgType)) 
    {
        DBGPRINT(RT_DEBUG_ERROR, "MlmeEnqueueForRecv (MsgType error)\n");
        return FALSE;
    }
    // OK, we got all the informations, it is time to put things into queue
    RTMPAcquireSpinLock(pAd, &(Queue->Lock));
    Tail = Queue->Tail;
    Queue->Tail++;
    Queue->Num++;
    if (Queue->Tail == MAX_LEN_OF_MLME_QUEUE) 
    {
        Queue->Tail = 0;
    }
    RTMPFreeSpinLock(pAd, &(Queue->Lock));
    DBGPRINT(RT_DEBUG_INFO, "MlmeEnqueueForRecv, num=%d\n",Queue->Num);
    
    Queue->Entry[Tail].Occupied = TRUE;
    Queue->Entry[Tail].Machine = Machine;
    Queue->Entry[Tail].MsgType = MsgType;
    Queue->Entry[Tail].MsgLen  = MsgLen;
//  Queue->Entry[Tail].TimeStamp.LowPart = TimeStampLow;
//  Queue->Entry[Tail].TimeStamp.HighPart = TimeStampHigh;
    Queue->Entry[Tail].Rssi = Rssi;
    NdisMoveMemory(Queue->Entry[Tail].Msg, Msg, MsgLen);
    
    MlmeHandler(pAd);

    return TRUE;
}

/*! \brief   Dequeue a message from the MLME Queue
 *  \param  *Queue    The MLME Queue
 *  \param  *Elem     The message dequeued from MLME Queue
 *  \return  TRUE if the Elem contains something, FALSE otherwise
 *  \pre
 *  \post
 */
BOOLEAN MlmeDequeue(
    IN	PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue, 
    OUT MLME_QUEUE_ELEM **Elem) 
{
    RTMPAcquireSpinLock(pAd, &(Queue->Lock));
    *Elem = &(Queue->Entry[Queue->Head]);
    Queue->Num--;
    Queue->Head++;
    if (Queue->Head == MAX_LEN_OF_MLME_QUEUE) 
    {
        Queue->Head = 0;
    }
    RTMPFreeSpinLock(pAd, &(Queue->Lock));
    DBGPRINT(RT_DEBUG_INFO, "MlmeDequeue, num=%d\n",Queue->Num);

    return TRUE;
}

/*! \brief  test if the MLME Queue is empty
 *  \param  *Queue    The MLME Queue
 *  \return TRUE if the Queue is empty, FALSE otherwise
 *  \pre
 *  \post
 */
BOOLEAN MlmeQueueEmpty(
    IN	PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue) 
{
    BOOLEAN Ans;

    RTMPAcquireSpinLock(pAd, &(Queue->Lock));
    Ans = (Queue->Num == 0);
    RTMPFreeSpinLock(pAd, &(Queue->Lock));

    return Ans;
}

/*! \brief   test if the MLME Queue is full
 *  \param   *Queue      The MLME Queue
 *  \return  TRUE if the Queue is empty, FALSE otherwise
 *  \pre
 *  \post
 */
BOOLEAN MlmeQueueFull(
    IN PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue) 
{
    BOOLEAN Ans;

    RTMPAcquireSpinLock(pAd, &(Queue->Lock));
    Ans = (Queue->Num == MAX_LEN_OF_MLME_QUEUE);
    RTMPFreeSpinLock(pAd, &(Queue->Lock));

    return Ans;
}

/*! \brief   The destructor of MLME Queue
 *  \param 
 *  \return
 *  \pre
 *  \post
 *  \note   Free spin lock only
 */
VOID MlmeQueueDestroy(
    IN MLME_QUEUE *Queue) 
{
//    NdisFreeSpinLock(&Queue->Lock);
}

/*! \brief   To substitute the message type if the message is coming from external
 *  \param  *Fr            The frame received
 *  \param  *Machine       The state machine
 *  \param  *MsgType       the message type for the state machine
 *  \return TRUE if the substitution is successful, FALSE otherwise
 *  \pre
 *  \post
 */
BOOLEAN MsgTypeSubst(
    IN MACFRAME *Fr, 
    OUT INT *Machine, 
    OUT INT *MsgType) 
{
    USHORT Seq;
    UCHAR 	EAPType; 
    BOOLEAN     Return;

// TODO:
// only PROBE_REQ can be broadcast, all others must be unicast-to-me && is_mybssid; otherwise, 
// ignore this frame

    if (Fr->Hdr.Type == BTYPE_DATA) 
    {    
        *Machine = WPA_STATE_MACHINE;
        EAPType = *((UCHAR*)Fr+LENGTH_802_11+LENGTH_802_1_H+1);
        Return = WPAMsgTypeSubst(EAPType, MsgType);
        return Return;

    }
    else if (Fr->Hdr.Type == BTYPE_MGMT) 		
    {
    switch (Fr->Hdr.SubType) 
    {
        case SUBTYPE_ASSOC_REQ:
            *Machine = ASSOC_STATE_MACHINE;
            *MsgType = MT2_PEER_ASSOC_REQ;
            break;
//      case SUBTYPE_ASSOC_RSP:
//          *Machine = ASSOC_STATE_MACHINE;
//          *MsgType = MT2_PEER_ASSOC_RSP;
//          break;
        case SUBTYPE_REASSOC_REQ:
            *Machine = ASSOC_STATE_MACHINE;
            *MsgType = MT2_PEER_REASSOC_REQ;
            break;
//      case SUBTYPE_REASSOC_RSP:
//          *Machine = ASSOC_STATE_MACHINE;
//          *MsgType = MT2_PEER_REASSOC_RSP;
//          break;
        case SUBTYPE_PROBE_REQ:
            *Machine = SYNC_STATE_MACHINE;
            *MsgType = MT2_PEER_PROBE_REQ;
            break;
//      case SUBTYPE_PROBE_RSP:
//          *Machine = SYNC_STATE_MACHINE;
//          *MsgType = MT2_PEER_PROBE_RSP;
//          break;
//      case SUBTYPE_BEACON:
//          *Machine = SYNC_STATE_MACHINE;
//          *MsgType = MT2_PEER_BEACON;
//          break;
//      case SUBTYPE_ATIM:
//          *Machine = SYNC_STATE_MACHINE;
//          *MsgType = MT2_PEER_ATIM;
//          break;
        case SUBTYPE_DISASSOC:
            *Machine = ASSOC_STATE_MACHINE;
            *MsgType = MT2_PEER_DISASSOC_REQ;
            break;
        case SUBTYPE_AUTH:
            // get the sequence number from payload 24 Mac Header + 2 bytes algorithm
            NdisMoveMemory(&Seq, &Fr->Octet[2], sizeof(USHORT));
            if (Seq == 1 || Seq == 3) 
            {
                *Machine = AUTH_RSP_STATE_MACHINE;
                *MsgType = MT2_PEER_AUTH_ODD;
            } 
//          else if (Seq == 2 || Seq == 4) 
//          {
//              *Machine = AUTH_STATE_MACHINE;
//              *MsgType = MT2_PEER_AUTH_EVEN;
//          } 
            else 
            {
                return FALSE;
            }
            break;
        case SUBTYPE_DEAUTH:
            *Machine = AUTH_RSP_STATE_MACHINE;
            *MsgType = MT2_PEER_DEAUTH;
            break;
        default:
            return FALSE;
            break;
    }

    		return TRUE;
    }
    return FALSE;
}

// ===========================================================================================
// state_machine.c
// ===========================================================================================

/*! \brief Initialize the state machine.
 *  \param *S           pointer to the state machine 
 *  \param  Trans       State machine transition function
 *  \param  StNr        number of states 
 *  \param  MsgNr       number of messages 
 *  \param  DefFunc     default function, when there is invalid state/message combination 
 *  \param  InitState   initial state of the state machine 
 *  \param  Base        StateMachine base, internal use only
 *  \pre p_sm should be a legal pointer
 *  \post
 */

VOID StateMachineInit(
    IN STATE_MACHINE *S, 
    IN STATE_MACHINE_FUNC Trans[], 
    IN ULONG StNr, 
    IN ULONG MsgNr, 
    IN STATE_MACHINE_FUNC DefFunc, 
    IN ULONG InitState, 
    IN ULONG Base) 
{
    ULONG i, j;

    // set number of states and messages
    S->NrState = StNr;
    S->NrMsg   = MsgNr;
    S->Base    = Base;

    S->TransFunc  = Trans;
    
    // init all state transition to default function
    for (i = 0; i < StNr; i++) 
    {
        for (j = 0; j < MsgNr; j++) 
        {
            S->TransFunc[i * MsgNr + j] = DefFunc;
        }
    }
    
    // set the starting state
    S->CurrState = InitState;

}

/*! \brief This function fills in the function pointer into the cell in the state machine 
 *  \param *S   pointer to the state machine
 *  \param St   state
 *  \param Msg  incoming message
 *  \param f    the function to be executed when (state, message) combination occurs at the state machine
 *  \pre *S should be a legal pointer to the state machine, st, msg, should be all within the range, Base should be set in the initial state
 *  \post
 */
VOID StateMachineSetAction(
    IN STATE_MACHINE *S, 
    IN ULONG St, 
    IN ULONG Msg, 
    IN STATE_MACHINE_FUNC Func) 
{
    ULONG MsgIdx;
    
    MsgIdx = Msg - S->Base;

    if (St < S->NrState && MsgIdx < S->NrMsg) 
    {
        // boundary checking before setting the action
        S->TransFunc[St * S->NrMsg + MsgIdx] = Func;
    } 
}

/*! \brief   This function does the state transition
 *  \param   *Adapter the NIC adapter pointer
 *  \param   *S       the state machine
 *  \param   *Elem    the message to be executed
 *  \return   None
 */
VOID StateMachinePerformAction(
    IN	PRTMP_ADAPTER	pAd, 
    IN STATE_MACHINE *S, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    (*(S->TransFunc[S->CurrState * S->NrMsg + Elem->MsgType - S->Base]))(pAd, Elem);
}

/*
    ==========================================================================
    Description:
        The drop function, when machine executes this, the message is simply 
        ignored. This function does nothing, the message is freed in 
        StateMachinePerformAction()
    ==========================================================================
 */
VOID Drop(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
#if 0
    if ((Elem->MsgType == MT2_PEER_BEACON) ||
        (Elem->MsgType == MT2_PEER_PROBE_REQ) ||
        (Elem->MsgType == MT2_PEER_PROBE_RSP))
        ;
    else
    {
        DBGPRINT(RT_DEBUG_TRACE, ("Warn:>>Drop Msg=%d<<\n",Elem->MsgType));
    }
#endif    
}

// ===========================================================================================
// lfsr.c
// ===========================================================================================

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID LfsrInit(
    IN PRTMP_ADAPTER pAd, 
    IN ULONG Seed) 
{
    if (Seed == 0) 
        pAd->Mlme.ShiftReg = 1;
    else 
        pAd->Mlme.ShiftReg = Seed;
}

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
UCHAR RandomByte(
    IN PRTMP_ADAPTER pAd) 
{
    ULONG i;
    UCHAR R, Result;

    R = 0;

    for (i = 0; i < 8; i++) 
    {
        if (pAd->Mlme.ShiftReg & 0x00000001) 
        {
            pAd->Mlme.ShiftReg = ((pAd->Mlme.ShiftReg ^ LFSR_MASK) >> 1) | 0x80000000;
            Result = 1;
        } 
        else 
        {
            pAd->Mlme.ShiftReg = pAd->Mlme.ShiftReg >> 1;
            Result = 0;
        }
        R = (R << 1) | Result;
    }

    return R;
}

// ===========================================================================================
// asic.c
// ===========================================================================================

//
// RF register initialization set, support RT2420/RT2421
//
ULONG	RF2420RegTable[] = {
		0x94022058, 0x940c203e, 0x94000111, // chanel 6
		};

ULONG	RF2421RegTable[] = {
		0x94022058, 0x940c203e, 0x94000101, // chanel 6
		};

ULONG   RF2421RegTableAt500KRefClk[] = {
		0x940220b0, 0x940c407a, 0x94000101, // chanel 6
        };

ULONG   RF2421RegTableAtInvalidRefClk[] = {
        0x94022058, 0x940c2a32, 0x94000101
        };

// Ralink RF channel value, support RT2420/RT2421
ULONG	RFChannelValue[] = {
	0x940c1fda, 0x940c1fee, 0x940c2002, 0x940c2016, 0x940c202a, 0x940c203e, 0x940c2052,
	0x940c2066, 0x940c207a, 0x940c208e, 0x940c20a2, 0x940c20b6, 0x940c20ca, 0x940c20fa,
};

// RF channel value for RT2421B using 500Khz reference clock
ULONG   RFChannelValueAt500KRefClk[] = {
	0x940c3fb2, 0x940c3fda, 0x940c4002, 0x940c402a, 0x940c4052, 0x940c407a, 0x940c40a2,
	0x940c40ca, 0x940c40f2, 0x940c411a, 0x940c4142, 0x940c416a, 0x940c4192, 0x940c41f2,
};

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID AsicSwitchChannel(
    IN PRTMP_ADAPTER pAd, 
    IN ULONG Channel) 
{
    DBGPRINT(RT_DEBUG_INFO, "AsicSwitchChannel to #%d\n", Channel);
	// Initialize RF register to default value
	if (pAd->PortCfg.RfType == RFIC_2421)
    {
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTable[0]); // program RF R1
        RTMP_RF_IO_WRITE32(pAd, RFChannelValue[Channel - 1]);   // program RF R2
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTable[2]); // program RF R3
    }
    else
    {
	    RTMP_RF_IO_WRITE32(pAd, RF2420RegTable[0]); // program RF R1
        RTMP_RF_IO_WRITE32(pAd, RFChannelValue[Channel - 1]);   // program RF R2
	    RTMP_RF_IO_WRITE32(pAd, RF2420RegTable[2]); // program RF R3
    }
}

/*
    ==========================================================================
    Description:
        This function is required for 2421 only, and should not be used during
        site survey. It's only required after NIC decided to stay at a channel
        for a longer period.
        When this function is called, it's always after AsicSwitchChannel().
    ==========================================================================
 */
VOID AsicLockChannel(
    IN PRTMP_ADAPTER pAd, 
    IN ULONG Channel) 
{
	if (pAd->PortCfg.RfType == RFIC_2421)
    {
        // program RF at some invalid reference clock rate to activate the auto-tuning function
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTableAtInvalidRefClk[0]); // program RF R1
        RTMP_RF_IO_WRITE32(pAd, RF2421RegTableAtInvalidRefClk[1]); // program RF R2
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTableAtInvalidRefClk[2]); // program RF R3
	    pAd->PortCfg.RfTuningStep = 1;
	    pAd->PortCfg.RfTuningChannel = (UCHAR)Channel;
        DBGPRINT(RT_DEBUG_TRACE, "1. R1=0x%x, R2=0x%0x, R3=0x%x\n", 
            RF2421RegTableAtInvalidRefClk[0], 
            RF2421RegTableAtInvalidRefClk[1], 
            RF2421RegTableAtInvalidRefClk[2]);

        // perform next step 1-ms later
        pAd->PortCfg.RfTuningTimer.expires = jiffies + 1;	/* 1 msec */
        add_timer(&pAd->PortCfg.RfTuningTimer);
    }
}

VOID AsicRfTuningExec(
    IN	unsigned long data) 
{
    RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;

    if (pAd->PortCfg.RfTuningStep == 1)
    {
	    // re-program RF using 1Mhz reference clock
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTable[0]); // program RF R1
        RTMP_RF_IO_WRITE32(pAd, RFChannelValue[pAd->PortCfg.RfTuningChannel - 1]);   // program RF R2
	    RTMP_RF_IO_WRITE32(pAd, RF2421RegTable[2]); // program RF R3
	    pAd->PortCfg.RfTuningStep = 2;
        DBGPRINT(RT_DEBUG_TRACE, "2. R1=0x%x, R2=0x%0x, R3=0x%x\n", 
            RF2421RegTable[0], 
            RFChannelValue[pAd->PortCfg.RfTuningChannel - 1], 
            RF2421RegTable[2]);

        // set 1 ms one-shot timer to turn off RF auto tuning
        pAd->PortCfg.RfTuningTimer.expires = jiffies + 1;	/* 1 msec */
		add_timer(&pAd->PortCfg.RfTuningTimer);
    }
    else if (pAd->PortCfg.RfTuningStep == 2)
    {
        ULONG R1 = RF2421RegTable[0] & 0xfffdffff; // RF R1.bit17 "tune_en1" OFF
        ULONG R3 = RF2421RegTable[2] & 0xfffffeff; // RF R3.bit8 "tune_en2" OFF

        // disable auto-tuning to local on this channel
        RTMP_RF_IO_WRITE32(pAd, R1); 
        RTMP_RF_IO_WRITE32(pAd, R3); 
	    pAd->PortCfg.RfTuningStep = 0;
    
        DBGPRINT(RT_DEBUG_TRACE, "3. R1=0x%x, R3=0x%x\n", R1, R3);
    }
}

/*
    ==========================================================================
    Description:
        put PHY to sleep here, and set next wakeup timer
    ==========================================================================
 */
VOID AsicSleepThenAutoWakeup(
    IN PRTMP_ADAPTER pAd, 
    IN USHORT TbttNumToNextWakeUp) 
{
    CSR20_STRUC Csr20;
    PWRCSR1_STRUC Pwrcsr1;

    // we have decided to SLEEP, so at least do it for a BEACON period.
    if (TbttNumToNextWakeUp==0)
        TbttNumToNextWakeUp=1;
    
    // PWRCSR0 remains untouched
    
    // set CSR20 for next wakeup
    Csr20.word = 0;
    Csr20.field.NumBcnBeforeWakeup = TbttNumToNextWakeUp - 1;
    Csr20.field.DelayAfterBcn = (pAd->PortCfg.BeaconPeriod - 20) << 4; // 20 TU ahead of desired TBTT
    Csr20.field.AutoWake = 1;
    RTMP_IO_WRITE32(pAd, CSR20, Csr20.word);

    // set PWRCSR1 to put PHY into SLEEP state
    Pwrcsr1.word = 0;
    Pwrcsr1.field.PutToSleep = 1;
    Pwrcsr1.field.BbpDesireState = 1; // 01:SLEEP
    Pwrcsr1.field.RfDesireState = 1;  // 01:SLEEP
    RTMP_IO_WRITE32(pAd, PWRCSR1, Pwrcsr1.word);
//  pAd->PortCfg.Pss = PWR_SAVE;
}

/*
    ==========================================================================
    Description:
        AsicForceWakeup() is used whenever manual wakeup is required
        AsicForceSleep() should only be used when Massoc==FALSE. When
        Massoc==TRUE, we should use AsicSleepThenAutoWakeup() instead.
    ==========================================================================
 */
VOID AsicForceSleep(
    IN PRTMP_ADAPTER pAd)
{
    PWRCSR1_STRUC Pwrcsr1;

    DBGPRINT(RT_DEBUG_TRACE, ">>>AsicForceSleep<<<\n");
    Pwrcsr1.word = 0;
    Pwrcsr1.field.RfDesireState = 1; // 01:SLEEP state
    Pwrcsr1.field.BbpDesireState = 1; // 01:SLEEP state
    Pwrcsr1.field.SetState = 1;
    RTMP_IO_WRITE32(pAd, PWRCSR1, Pwrcsr1.word);
}

VOID AsicForceWakeup(
    IN PRTMP_ADAPTER pAd)
{
    PWRCSR1_STRUC Pwrcsr1;

    DBGPRINT(RT_DEBUG_TRACE, ">>>AsicForceWakeup<<<\n");
    Pwrcsr1.word = 0;
    Pwrcsr1.field.RfDesireState = 3; // 11:AWAKE state
    Pwrcsr1.field.BbpDesireState = 3; // 11:AWAKE state
    Pwrcsr1.field.SetState = 1;
    RTMP_IO_WRITE32(pAd, PWRCSR1, Pwrcsr1.word);
}

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID AsicSetBssid(
    IN PRTMP_ADAPTER pAd, 
    IN MACADDR *pBssid) 
{
    ULONG         Addr4;

    COPY_MAC_ADDR(&pAd->PortCfg.Bssid, pBssid);
    Addr4 = (ULONG)(pBssid->Octet[0]) | 
            (ULONG)(pBssid->Octet[1] << 8) | 
            (ULONG)(pBssid->Octet[2] << 16) |
            (ULONG)(pBssid->Octet[3] << 24);
    RTMP_IO_WRITE32(pAd, CSR5, Addr4);
    
    Addr4 = (ULONG)(pBssid->Octet[4]) | (ULONG)(pBssid->Octet[5] << 8);
    RTMP_IO_WRITE32(pAd, CSR6, Addr4);
}

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID AsicDisableSync(
    IN PRTMP_ADAPTER pAd) 
{
    DBGPRINT(RT_DEBUG_TRACE, "--->Disable TSF synchronization\n");
    RTMP_IO_WRITE32(pAd, CSR14, 0x00000000);
}

/*
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID AsicEnableBssSync(
    IN PRTMP_ADAPTER pAd) 
{
    CSR12_STRUC Csr12;
    CSR13_STRUC Csr13;
    CSR14_STRUC Csr14;
    BCNCSR1_STRUC Bcncsr1;
    BOOLEAN IsApPc;

    DBGPRINT(RT_DEBUG_TRACE, "--->AsicEnableBssSync(INFRA mode)\n");
    
    RTMP_IO_WRITE32(pAd, CSR14, 0x00000000);
    
    Csr12.word = 0;
    Csr12.field.BeaconInterval = pAd->PortCfg.BeaconPeriod << 4; // ASIC register in units of 1/16 TU
    Csr12.field.CfpMaxDuration = pAd->PortCfg.CfpMaxDuration << 4; // ASIC register in units of 1/16 TU
    RTMP_IO_WRITE32(pAd, CSR12, Csr12.word);
    
    Csr13.word = 0;
    Csr13.field.CfpPeriod = pAd->PortCfg.CfpDurRemain << 4; // ASIC register in units of 1/16 TU
    RTMP_IO_WRITE32(pAd, CSR13, Csr13.word);

    Bcncsr1.word = 0;
    Bcncsr1.field.Preload = TBTT_PRELOAD_TIME; // we guess TBTT is 2 TU ahead of BEACON-RxEnd time
    RTMP_IO_WRITE32(pAd, BCNCSR1, Bcncsr1.word);

    IsApPc = (CAP_IS_CF_POLLABLE_ON(pAd->PortCfg.CapabilityInfo) && 
              CAP_IS_CF_POLL_REQ_ON(pAd->PortCfg.CapabilityInfo));
    IsApPc = FALSE; // TODO: not support so far
    
    Csr14.word = 0;
    Csr14.field.TsfCount = 1;
    Csr14.field.TsfSync = 2; // sync TSF in AD HOC mode??
    if (IsApPc) 
    {
        Csr14.field.CfpCntPreload = pAd->PortCfg.CfpCount;
        Csr14.field.Tcfp = 1;
    }
    Csr14.field.BeaconGen = 1;
//  Csr14.field.TbcnPreload = (pAd->PortCfg.BeaconPeriod - 30) << 4; // TODO: ???? 1 TU ???
    Csr14.field.Tbcn = 1;
    RTMP_IO_WRITE32(pAd, CSR14, Csr14.word);
    
}

VOID AsicLedPeriodicExec(
    IN	unsigned long data) 
{
    RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;
    ULONG LedCsr = 0x0000461E; // 0x0000461E;
    
    pAd->PortCfg.LedCntl.fOdd = ! pAd->PortCfg.LedCntl.fOdd;

    LedCsr |= 0x00010000; // enable hardwired TX activity LED
    if (pAd->PortCfg.LedCntl.fOdd && pAd->PortCfg.LedCntl.fRxActivity) 
        LedCsr |= 0x00020000; // turn on software-based RX activity LED
    pAd->PortCfg.LedCntl.fRxActivity = FALSE;

    if (LedCsr != pAd->PortCfg.LedCntl.LastLedCsr)
    {
//        DBGPRINT(RT_DEBUG_TRACE, ("AsicLedPeriodicExec(%8x)\n",LedCsr));
        pAd->PortCfg.LedCntl.LastLedCsr = LedCsr;
        RTMP_IO_WRITE32(pAd, LEDCSR, LedCsr);
    }

	add_timer(&pAd->PortCfg.LedCntl.BlinkTimer);
}

