/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rtmp.h

    Abstract:
    Miniport generic portion header file

    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
    Paul Lin    08-01-2002    created

*/
#ifndef __RTMP_H__
#define __RTMP_H__

#include "mlme.h"
#include "oid.h"

//
//  Queue structure and macros
//
typedef struct  _QUEUE_ENTRY    {
    struct _QUEUE_ENTRY     *Next;
}   QUEUE_ENTRY, *PQUEUE_ENTRY;

// Queue structure
typedef struct  _QUEUE_HEADER   {
    PQUEUE_ENTRY    Head;
    PQUEUE_ENTRY    Tail;
    ULONG			Number;
}   QUEUE_HEADER, *PQUEUE_HEADER;

#define InitializeQueueHeader(QueueHeader)              \
{                                                       \
    (QueueHeader)->Head = (QueueHeader)->Tail = NULL;   \
    (QueueHeader)->Number = 0;							\
}

#define RemoveHeadQueue(QueueHeader)                \
(QueueHeader)->Head;                                \
{                                                   \
    PQUEUE_ENTRY pNext;                             \
    pNext = (QueueHeader)->Head->Next;              \
    (QueueHeader)->Head = pNext;                    \
    if (pNext == NULL)                              \
        (QueueHeader)->Tail = NULL;                 \
    (QueueHeader)->Number--;						\
}

#define InsertHeadQueue(QueueHeader, QueueEntry)            \
{                                                           \
    ((PQUEUE_ENTRY)QueueEntry)->Next = (QueueHeader)->Head; \
    (QueueHeader)->Head = (PQUEUE_ENTRY)(QueueEntry);       \
    if ((QueueHeader)->Tail == NULL)                        \
        (QueueHeader)->Tail = (PQUEUE_ENTRY)(QueueEntry);   \
    (QueueHeader)->Number++;								\
}

#define InsertTailQueue(QueueHeader, QueueEntry)                \
{                                                               \
    ((PQUEUE_ENTRY)QueueEntry)->Next = NULL;                    \
    if ((QueueHeader)->Tail)                                    \
        (QueueHeader)->Tail->Next = (PQUEUE_ENTRY)(QueueEntry); \
    else                                                        \
        (QueueHeader)->Head = (PQUEUE_ENTRY)(QueueEntry);       \
    (QueueHeader)->Tail = (PQUEUE_ENTRY)(QueueEntry);           \
    (QueueHeader)->Number++;					\
}

//
//  MACRO for debugging information
//
#ifdef RT2400_DBG
extern ULONG    RTDebugLevel;

#define DBGPRINT(Level, fmt, args...) 					\
{									\
	if (Level <= RTDebugLevel)      				\
	{								\
		printk(NIC_DBG_STRING);					\
		printk(KERN_DEBUG fmt, ## args);			\
	}								\
}
#else
#define DBGPRINT(Level, fmt, args...)
#endif

//
//	spin_lock enhanced for Nested spin lock
//
#define RTMPFreeSpinLock(pAd, lock)						\
{										\
	pAd->SpinLockCount--;							\
	if(pAd->SpinLockCount < 0)						\
	{									\
		DBGPRINT(RT_DEBUG_ERROR, "SpinLockCount error!!!\n");		\
	}									\
	spin_unlock(lock);							\
	if(pAd->SpinLockCount == 0)						\
	{									\
		local_irq_enable();						\
	}									\
}

#define RTMPAcquireSpinLock(pAd, lock)			\
{							\
	local_irq_disable();				\
	spin_lock(lock);				\
	pAd->SpinLockCount++;				\
}

#define RTMPFreeSkbBuffer(skb)				\
{							\
	if (in_interrupt())				\
		dev_kfree_skb_irq(skb);			\
	else						\
		dev_kfree_skb(skb);			\
}

//  Assert MACRO to make sure program running
//
#undef  ASSERT
#define ASSERT(x)                                                                       \
{                                                                                       \
    if (!(x))                                                                           \
    {                                                                                   \
        printk(KERN_WARNING __FILE__ ":%d assert " #x "failed\n", __LINE__);		\
    }                                                                                   \
}

//
//  Macros for flag and ref count operations
//
#define RTMP_SET_FLAG(_M, _F)       ((_M)->Flags |= (_F))
#define RTMP_CLEAR_FLAG(_M, _F)     ((_M)->Flags &= ~(_F))
#define RTMP_CLEAR_FLAGS(_M)        ((_M)->Flags = 0)
#define RTMP_TEST_FLAG(_M, _F)      (((_M)->Flags & (_F)) != 0)
#define RTMP_TEST_FLAGS(_M, _F)     (((_M)->Flags & (_F)) == (_F))

#define RTMP_INC_RCV_REF(_A)        ((_A)->RcvRefCount++)
#define RTMP_DEC_RCV_REF(_A)        ((_A)->RcvRefCount--)
#define RTMP_GET_RCV_REF(_A)        ((_A)->RcvRefCount)

#define RTMP_INC_SEND_REF(_A)       ((_A)->SendRefCount++)
#define RTMP_DEC_SEND_REF(_A)       ((_A)->SendRefCount--)
#define RTMP_GET_SEND_REF(_A)       ((_A)->SendRefCount)

#define	NdisEqualMemory(Source1, Source2, Length)	RTMPEqualMemory(Source1, Source2, Length)

//
// MACRO for 32-bit PCI register read / write
//
// Usage : RTMP_IO_READ32(
//              PRTMP_ADAPTER pAdapter,
//              ULONG Register_Offset,
//              PULONG  pValue)
//
//         RTMP_IO_WRITE32(
//              PRTMP_ADAPTER pAdapter,
//              ULONG Register_Offset,
//              ULONG Value)
//
#define RTMP_IO_READ32(_A, _R, _pV)	(*_pV = readl(_A->CSRBaseAddress+_R))
#define RTMP_IO_WRITE32(_A, _R, _V)	(writel(_V, _A->CSRBaseAddress+_R))
//
// BBP & RF are using indirect access. Before write any value into it.
// We have to make sure there is no outstanding command pending via checking busy bit.
//
#define	MAX_BUSY_COUNT	10			// Nunber of retry before failing access BBP & RF indirect register
//
#define	RTMP_BBP_IO_WRITE32(_A, _V)				\
{												\
	BBPCSR_STRUC	Value;						\
	ULONG			BusyCnt = 0;				\
	do {										\
		RTMP_IO_READ32(_A, BBPCSR, &Value.word);	\
		if (Value.field.Busy == IDLE)			\
			break;								\
		BusyCnt++;								\
	}	while (BusyCnt < MAX_BUSY_COUNT);		\
	RTMP_IO_WRITE32(_A, BBPCSR, _V);				\
}
//
#define	RTMP_RF_IO_WRITE32(_A, _V)				\
{												\
	RFCSR_STRUC		Value;						\
	ULONG			BusyCnt = 0;				\
	do {										\
		RTMP_IO_READ32(_A, RFCSR, &(Value.word));	\
		if (Value.field.Busy == IDLE)			\
			break;								\
		BusyCnt++;								\
	}	while (BusyCnt < MAX_BUSY_COUNT);		\
	RTMP_IO_WRITE32(_A, RFCSR, _V);				\
}
//
#define	RTMP_BBP_IO_READ32(_A, _pV)				\
{												\
	BBPCSR_STRUC	Value;						\
	ULONG			BusyCnt = 0;				\
	RTMP_BBP_IO_WRITE32(_A, *(_pV));			\
	do {										\
		RTMP_IO_READ32(_A, BBPCSR, &Value.word);\
		if (Value.field.Busy == IDLE)			\
			break;								\
		BusyCnt++;								\
	}	while (BusyCnt < MAX_BUSY_COUNT);		\
	if (BusyCnt == MAX_BUSY_COUNT)				\
		*(_pV) = 0xff;							\
	else										\
	        *(_pV) = Value.field.Value;					\
}
// Read BBP register by register's ID
#define	RTMP_BBP_IO_READ32_BY_REG_ID(_A, _I, _pV)	\
{												\
	BBPCSR_STRUC	BbpCsr;						\
	BbpCsr.word = 0;								\
	BbpCsr.field.WriteControl = 0;				\
	BbpCsr.field.Busy = 1;						\
	BbpCsr.field.RegNum = _I;					\
	RTMP_BBP_IO_READ32(_A, &BbpCsr.word);		\
	*(_pV) = (UCHAR) BbpCsr.field.Value;			\
}
// Write BBP register by register's ID & value
#define	RTMP_BBP_IO_WRITE32_BY_REG_ID(_A, _I, _V)	\
{												\
	BBPCSR_STRUC	BbpCsr;						\
	BbpCsr.word = 0;								\
    BbpCsr.field.WriteControl = 1;				\
    BbpCsr.field.Busy = 1;						\
    BbpCsr.field.Value = _V;					\
    BbpCsr.field.RegNum = _I;					\
    RTMP_BBP_IO_WRITE32(_A, BbpCsr.word);		\
}


//
//  Some utility macros
//
#ifndef min
#define min(_a, _b)     (((_a) < (_b)) ? (_a) : (_b))
#endif

#ifndef max
#define max(_a, _b)     (((_a) > (_b)) ? (_a) : (_b))
#endif

#define	INC_COUNTER(Val)	(Val++)

//#define	INFRA_ON(_p)		        (((_p)->PortCfg.Massoc) == TRUE)	// Check Massoc
//#define ADHOC_ON(_p)                (((_p)->PortCfg.Mibss) == TRUE)    // check Mibss

// This control buffer owned by you as long as you have the skb queue.
#define	RTMP_SET_PACKET_FRAGMENTS(_p, number)	((_p)->cb[10] = number)
#define RTMP_GET_PACKET_FRAGMENTS(_p)			((_p)->cb[10])
#define	RTMP_SET_PACKET_RTS(_p, number)			((_p)->cb[11] = number)
#define RTMP_GET_PACKET_RTS(_p)					((_p)->cb[11])
#define RTMP_SET_PACKET_SOURCE(_p, pktsrc)      ((_p)->cb[12] = pktsrc)
#define RTMP_GET_PACKET_SOURCE(_p)              ((_p)->cb[12])
#ifdef WDS
#define RTMP_SET_PACKET_NET_DEVICE(_p, idx)    	((_p)->cb[13] = (idx + 0x10))
#define RTMP_GET_PACKET_NET_DEVICE(_p)			((_p)->cb[13])
#endif

#define	MAKE_802_3_HEADER(_p, _pMac1, _pMac2, _pType)								\
{																					\
	NdisMoveMemory(_p, _pMac1, ETH_LENGTH_OF_ADDRESS);								\
	NdisMoveMemory((_p + ETH_LENGTH_OF_ADDRESS), _pMac2, ETH_LENGTH_OF_ADDRESS);	\
	NdisMoveMemory((_p + ETH_LENGTH_OF_ADDRESS * 2), _pType, LENGTH_802_3_TYPE);	\
}

//
// Register set pair for initialzation register set definition
//
typedef	struct	_RTMP_REG_PAIR
{
	ULONG	Register;
	ULONG	Value;
}	RTMP_REG_PAIR, *PRTMP_REG_PAIR;

//
//  Statistic counter structure
//
typedef struct _COUNTER_802_3
{
    // General Stats
	ULONG		GoodTransmits;
	ULONG		GoodReceives;
	ULONG		TxErrors;
	ULONG		RxErrors;
	ULONG		RxNoBuffer;

    // Ethernet Stats
	ULONG		RcvAlignmentErrors;
	ULONG		OneCollision;
	ULONG		MoreCollisions;

}   COUNTER_802_3, *PCOUNTER_802_3;

typedef struct _COUNTER_802_11 {
    ULONG           Length;
    LARGE_INTEGER   TransmittedFragmentCount;
    LARGE_INTEGER   MulticastTransmittedFrameCount;
    LARGE_INTEGER   FailedCount;
    LARGE_INTEGER   RetryCount;
    LARGE_INTEGER   MultipleRetryCount;
    LARGE_INTEGER   RTSSuccessCount;
    LARGE_INTEGER   RTSFailureCount;
    LARGE_INTEGER   ACKFailureCount;
    LARGE_INTEGER   FrameDuplicateCount;
    LARGE_INTEGER   ReceivedFragmentCount;
    LARGE_INTEGER   MulticastReceivedFrameCount;
    LARGE_INTEGER   FCSErrorCount;
} COUNTER_802_11, *PCOUNTER_802_11;

typedef struct _COUNTER_RALINK {
    ULONG           TransmittedByteCount;   // both successful and failure, used to calculate TX throughput
    ULONG           ReceivedByteCount;      // both CRC okay and CRC error, used to calculate RX throughput
    ULONG           BeenDisassociatedCount;
    ULONG           BadCQIAutoRecoveryCount;
    ULONG           PoorCQIRoamingCount;
} COUNTER_RALINK, *PCOUNTER_RALINK;

typedef enum _NDIS_MEDIA_STATE
{
	NdisMediaStateConnected,
	NdisMediaStateDisconnected
} NDIS_MEDIA_STATE, *PNDIS_MEDIA_STATE;

typedef enum {PHY_11A, PHY_11B, PHY_11G, PHY_11_TYPE_MAX} PHYMODE;

//
//  Arcfour Structure Added by PaulWu
//
typedef struct  _ARCFOUR
{
	UINT			X;
	UINT			Y;
	UCHAR			STATE[256];
}   ARCFOURCONTEXT, *PARCFOURCONTEXT;

// Shared key data structure
typedef	struct	_WEP_KEY {
	UCHAR	KeyLen;						// Key length for each key, 0: entry is invalid
	UCHAR	Key[MAX_LEN_OF_KEY];		// right now we implement 4 keys, 128 bits max
}	WEP_KEY, *PWEP_KEY;

// Shared key data structure
typedef	struct	_WPA_KEY {
	UCHAR	KeyLen;				// Key length for each key, 0: entry is invalid
	UCHAR	Key[16];			// right now we implement 4 keys, 128 bits max
	UCHAR	RxMic[8];
	UCHAR	TxMic[8];
	NDIS_802_11_MAC_ADDRESS	BssId;	// For pairwise key only
	UCHAR	TxTsc[6];			// 48bit TSC value
	UCHAR	RxTsc[6];			// 48bit TSC value
}	WPA_KEY, *PWPA_KEY;

// configuration to be used when this STA starts a new ADHOC network
typedef struct _IBSS_CONFIG {
    USHORT    BeaconPeriod;
    USHORT    AtimWin;
    UCHAR     Channel;
    UCHAR     SupportedRates[MAX_LEN_OF_SUPPORTED_RATES];    // Supported rates
    UCHAR     SupportedRatesLen;
}   IBSS_CONFIG, *PIBSS_CONFIG;

typedef struct _LED_CONTROL {
    BOOLEAN				fOdd;
    BOOLEAN				fRxActivity;
    struct timer_list	BlinkTimer; // 50 ms periodic timer
    ULONG				LastLedCsr;
}   LED_CONTROL;

// structure to define WPA Group Key Rekey Interval
typedef struct _RT_802_11_WPA_REKEY {
    ULONG ReKeyMethod;          // mechanism for rekeying: 0:disable, 1: time-based, 2: packet-based
    ULONG ReKeyInterval;        // time-based: seconds, packet-based: kilo-packets
} RT_WPA_REKEY,*PRT_WPA_REKEY,RT_802_11_WPA_REKEY, *PRT_802_11_WPA_REKEY;

#ifdef WDS
typedef struct _RT_802_11_WDS_ENTRY {
    UCHAR	   				WdsAddr[ETH_LENGTH_OF_ADDRESS];
    struct net_device		*dev;
    UCHAR           		MaxSupportedRate;
    UCHAR           		CurrTxRate;
    USHORT          		TxQuality[MAX_LEN_OF_SUPPORTED_RATES];
    USHORT          		OneSecTxOkCount;
    USHORT          		OneSecTxRetryOkCount;
    USHORT          		OneSecTxFailCount;
    ULONG                   CurrTxRateStableTime; // # of second in current TX rate
    UCHAR                   TxRateUpPenalty;      // extra # of second penalty due to last unstable condition
} RT_802_11_WDS_ENTRY, PRT_802_11_WDS_ENTRY;

typedef struct _RT_802_11_WDS_TABLE {
	ULONG					Num;
	WPA_KEY					Wpa_key;
    RT_802_11_WDS_ENTRY	   	WdsEntry[MAX_WDS_ENTRY];
} RT_802_11_WDS_TABLE, PRT_802_11_WDS_TABLE;

typedef struct _WDS_ADAPTER
{
	struct net_device		*RtmpDev;
	struct net_device		*WdsDev;
} WDS_ADAPTER, PWDS_ADAPTER;
#endif
// structure to tune BBP R13 "RX AGC VGC init"
typedef struct _RT_802_11_RX_AGC_VGC_TUNING {
    UCHAR   Duration;           // unit: sec
    UCHAR   FlaseCcaThreshold;  // delta of flase CCA happen in the latest "Duration"
    UCHAR   RxAgcVgcDelta;      // R13 += RxAgcVgcDelta whenever flase CCA >= threshold
    UCHAR   MaxRxAgcVgc;        // max value of R13
} RT_802_11_RX_AGC_VGC_TUNING, *PRT_802_11_RX_AGC_VGC_TUNING;

typedef struct _RT_802_11_EVENT_LOG {
    ULONG			SystemTime;  // timestammp via NdisGetCurrentSystemTime()
    UCHAR           Addr[ETH_LENGTH_OF_ADDRESS];
    USHORT          Event;       // EVENT_xxx
} RT_802_11_EVENT_LOG, *PRT_802_11_EVENT_LOG;

typedef struct _RT_802_11_EVENT_TABLE {
    ULONG       Num;
    ULONG       Rsv;     // to align Log[] at LARGE_INEGER boundary
    RT_802_11_EVENT_LOG   Log[MAX_NUM_OF_EVENT];
} RT_802_11_EVENT_TABLE, PRT_802_11_EVENT_TABLE;

typedef struct _RT_802_11_FILTER_ENTRY {
    ULONG       				Mode;
    UCHAR				   	Addr[MAX_NUM_OF_FILTER_ENTRY][ETH_LENGTH_OF_ADDRESS];
} RT_802_11_FILTER_ENTRY, PRT_802_11_FILTER_ENTRY;

//  configuration initial parametes.  
typedef struct _RTMP_PROFILE_SETTING {
	UCHAR		Profile[32+1];
	UCHAR		SSID[32+1];
	UINT		SsidLen;
	UINT		Channel;
	UINT		Authentication;
	UINT		Encryption;
	UINT		NetworkType;
	UINT		KeyIndex;
	UINT		KeyLength;
	UINT		KeyType;
	UCHAR		Key1[26+1];
	UCHAR		Key2[26+1];
	UCHAR		Key3[26+1];
	UCHAR		Key4[26+1];
	UINT		TransRate;
	UINT		TransPower;
	UINT		RTS; 
	UINT		RTSCheck;
	UINT     	FragmentCheck;
	UINT		PSmode;
	UINT		PreamType;
	UINT		AntennaRx;
	UINT		AntennaTx;
	UINT		Fragment;
	RT_802_11_FILTER_ENTRY	FilterEntry;
	UCHAR		CountryRegion;
	struct  _RT_PROFILE_SETTING		*Next;
} RTMP_PROFILE_SETTING, *PRTMP_PROFILE_SETTING;

#define MAC_TABLE_FULL(Tab) ((Tab).size == MAX_LEN_OF_MAC_TABLE)

typedef struct _MAC_TABLE_ENTRY {
    BOOLEAN         Valid;
    //jan for wpa
    BOOLEAN         RetryTimerRunning;  
    // record which entry revoke MIC Failure , if it leaves the BSS itself, AP won't update aMICFailTime MIB
    UCHAR           CMTimerRunning;
    UCHAR           RSNIE_Len;
    UCHAR           RSN_IE[MAX_LEN_OF_RSNIE];
    UCHAR           ANonce[32];
    UCHAR           Flag;   
    UCHAR           R_Counter[LEN_KEY_DESC_REPLAY];
    UCHAR           PTK[64];
    UCHAR           ReTryCounter;   
    struct timer_list                 RetryTimer;
    NDIS_802_11_AUTHENTICATION_MODE     AuthMode;   // This should match to whatever microsoft defined
    NDIS_802_11_WEP_STATUS              WepStatus;
    WPA_STATE       WpaState;
    GTK_STATE       GTKState;
    USHORT          PortSecured;
    NDIS_802_11_PRIVACY_FILTER  PrivacyFilter;      // PrivacyFilter enum for 802.1X
    WPA_KEY         PairwiseKey[PAIRWISE_KEY_NO];
    
    MACADDR         Addr;
    UCHAR           PsMode;
    SST             Sst;
    AUTH_STATE      AuthState; // for SHARED KEY authentication state machine used only
    USHORT          Aid;
    USHORT          CapabilityInfo;
    UCHAR           MaxSupportedRate;
    UCHAR           CurrTxRate;
    UCHAR           Rate;
    ULONG           NoDataIdleCount;
    ULONG           PsQIdleCount;
    QUEUE_HEADER    PsQueue;
    struct _MAC_TABLE_ENTRY *pNext;
} MAC_TABLE_ENTRY, *PMAC_TABLE_ENTRY;

typedef struct _MAC_TABLE {
    UCHAR           Size;
    MAC_TABLE_ENTRY *Hash[HASH_TABLE_SIZE];
    MAC_TABLE_ENTRY Content[MAX_LEN_OF_MAC_TABLE];
    QUEUE_HEADER    McastPsQueue;
    ULONG           PsQIdleCount;
    BOOLEAN         fAnyStationInPsm;   // TRUE, then outgoing BCAST/MCAST frame should 
                                        // be put into McastPsQueue until DtimCount=0
} MAC_TABLE, *PMAC_TABLE;


// PortConfig
typedef struct _PORT_CONFIG {
    // MIB:ieee802dot11.dot11smt(1).dot11StationConfigTable(1)
    ULONG     MediumOccupancyLimit; // in units of TU (1 TU = 1024 usec)
    USHORT    CapabilityInfo;

    ULONG     AuthRspTimeout;       // in units of TU
//    USHORT    Psm;                  // power management mode   (PWR_ACTIVE|PWR_SAVE)
    USHORT    BeaconPeriod;         // in units of TU
	NDIS_802_11_AUTHENTICATION_MODE		AuthMode;	// This should match to whatever microsoft defined
	NDIS_802_11_WEP_STATUS				WepStatus;
    ULONG        ApdPid;
//  USHORT    DtimPeriod;           // number of BEACONS in between DTIMs
    ULONG     AssocRspTimeout;      // in units of TU

    USHORT    CfpMaxDuration;
    USHORT    CfpDurRemain;
    USHORT    CfpCount;
    USHORT    CfpPeriod;

//    USHORT    DisassocReason;
//    MACADDR   DisassocSta;
//    USHORT    DeauthReason;
//    MACADDR   DeauthSta;
//    USHORT    AuthFailReason;
//    MACADDR   AuthFailSta;
    
    // MIB:ieee802dot11.dot11smt(1).dot11WEPDefaultKeysTable(3)
	ULONG		ThisFieldIsUseless; // to align each key[i] at word-boundary

    WEP_KEY	  SharedKey[SHARE_KEY_NO];
    WPA_KEY         GroupKey[GROUP_KEY_NO];
    
    // for wpa
    UCHAR           R_Counter[LEN_KEY_DESC_REPLAY];
    UCHAR           GMK[32];
    UCHAR           PMK[32];
    UCHAR           GNonce[32];
    UCHAR           PortSecured;  // WPA 802.1x port control, WPA_802_1X_PORT_SECURED, WPA_802_1X_PORT_NOT_SECURED
    WPA_GTK_STATE   WpaGTKState;
    struct timer_list         CounterMeasureTimer;
    ULONG           MICErrorCounter;
    NDIS_802_11_PRIVACY_FILTER  PrivacyFilter;      // PrivacyFilter enum for 802.1X
    UCHAR           GKeyDoneStations; // number of entry left to update group key
    UCHAR           WPAGKeyID;
    // MIB:ieee802dot11.dot11smt(1).dot11WEPKeyMappingsTable(4)
    // not implemented yet

    UCHAR           CMTimerRunning;
    UCHAR           BANClass3Data;
    LARGE_INTEGER   aMICFailTime;
    LARGE_INTEGER   PrevaMICFailTime;
    ULONG           MICFailureCounter;
    // for Group Rekey 
    RT_WPA_REKEY    WPAREKEY;
    ULONG           REKEYCOUNTER;
    struct timer_list     REKEYTimer;
    UCHAR           REKEYTimerRunning;
        
    // MIB:ieee802dot11.dot11smt(1).dot11PrivacyTable(5)
    // 2 counters dot11WEPICVErrorCount and dot11WEPExcludedCount put in g_portsts
    UCHAR     DefaultKeyId;

    // MIB:ieee802dot11.dot11mac(2).dot11OperationTable(1)
    //MACADDR   StaAddr;            // my MAC address
    USHORT    RtsThreshold;       // in units of BYTE
    UCHAR     ShortRetryLimit;    // STA short retry count
    UCHAR     LongRetryLimit;     // STA long retry count
    USHORT    FragmentThreshold;
	BOOLEAN   bFragmentZeroDisable;		// Microsoft use 0 as disable 
    UINT      MaxTxLifeTime;      // in units of TU
    UINT      MaxRxLifeTime;      // in units of TU
    CHAR      ManufactureId[MAX_LEN_OF_MANUFACTURE_ID]; // NOT NULL terminated
    UCHAR     ManufactureIdLen;
    CHAR      ProductId[MAX_LEN_OF_PRODUCT_ID];     // NOT NULL terminated
    UCHAR     ProductIdLen;
    
    // MIB:ieee802dot11.dot11mac(2).dot11CountersTable(2)
    // all statistics counters put in g_portsts
    
    // MIB:ieee802dot11.dot11mac(2).dot11GroupAddressesTable(3)
    // not implemented yet
    
    // MIB:ieee802dot11.dot11res(3).dot11resAttribute(1)
    // not implemented yet

    // MIB:ieee802dot11.dot11phy(4).dot11PhyOperationTable(1)
    UCHAR     RadioType;          // always PHYTYPE_DSSS(2)
    UINT      CurrentRegDomain;
    UCHAR     TempType;          // TEMPTYPE_0_40(1) or TEMPTYPE_n30_70(2)
    
    // MIB:ieee802dot11.dot11phy(4).dot11PhyAntennaTable(2)
    UCHAR     CurrentTxAntenna;
    UCHAR     CurrentRxAntenna;
    UCHAR     NumberOfAntenna;

    // MIB:ieee802dot11.dot11phy(4).dot11PhyTxPowerTable(3)
    USHORT    TxPowerLevel[MAX_NUM_OF_POWER_LEVEL];
    UCHAR     NumOfTxPowerLevel;
    UCHAR     CurrentTxPowerLevelIndex;
    UCHAR     TxPower;
    UCHAR     TxRate;       // RATE_1, RATE_2, RATE_5_5, RATE_11
    
    // MIB:ieee802dot11.dot11phy(4).dot11PhyFHSSTable(4)
    // we don't support FHSS
    
    // MIB:ieee802dot11.dot11phy(4).dot11PhyDSSSTable(5)
    UCHAR     Channel;        // current (I)BSS channel used in the station
    UCHAR     CountryRegion;	// Enum of country region, 0: USA, 1: Europe, 2: Japan
    UCHAR     CcaModeSupported;
    UCHAR     CurrentCcaMode;
    UINT      EdThreshold;       // energy detection threshold
    
    // MIB:ieee802dot11.dot11phy(4).dot11PhyIRTable(6)
    // we don't support InfraRed PHY
    
    // MIB:ieee802dot11.dot11phy(4).dot11RegisteredDoaminSupportedTable(7)
    //UCHAR     DomainSupported[MAX_NUM_OF_DOMAIN];
    //UCHAR     NumOfSupportedDomain;
    
    // MIB:ieee802dot11.dot11phy(4).dot11AntennasListTable(8)
    BOOLEAN AntennaSupportTx;
    BOOLEAN AntennaSupportRx;
    BOOLEAN AntennaSupportDiversityRx;
    
	// Use user changed MAC
    BOOLEAN bLocalAdminMAC;                           
    
    // MIB:ieee802dot11.dot11phy(4).dot11SupportedDataRatesTxTable(9)
    // MIB:ieee802dot11.dot11phy(4).dot11SupportedDataRatesRxTable(10)
    UCHAR     SupportedRates[MAX_LEN_OF_SUPPORTED_RATES];    // Supported rates
    UCHAR     SupportedRatesLen;

    //
    // other parameters not defined in standard MIB
    //
    UCHAR     DesiredRates[MAX_LEN_OF_SUPPORTED_RATES];      // OID_802_11_DESIRED_RATES
    USHORT    RecvDtim;
    MACADDR   Bssid;
    MACADDR   Broadcast;            // FF:FF:FF:FF:FF:FF
    UCHAR     RssiTrigger;
    UCHAR     RssiTriggerMode;      // RSSI_TRIGGERED_UPON_BELOW_THRESHOLD or RSSI_TRIGGERED_UPON_EXCCEED_THRESHOLD
    UCHAR     LastRssi;             // last received BEACON's RSSI
//  UCHAR     AvgRssi;              // average of last 2 BEACON's RSSI
    USHORT    AtimWin;              // in kusec; IBSS parameter set element
    USHORT    Aid;                  // association ID
    UCHAR     MaxBasicRate;         // RATE_1, RATE_2, RATE_5_5, RATE_11
    UCHAR     MaxTxRate;            // RATE_1, RATE_2, RATE_5_5, RATE_11
    USHORT    DefaultListenCount;   // default listen count;
    USHORT    CurrListenCount;      // in units of Beacon interval
    UINT      MinChnlTime;          // minimum channel time (in uSEC)
    UINT      MaxChnlTime;          // maximum channel time (in uSEC)

    UCHAR     SsidLen;               // the actual ssid length in used
    CHAR      Ssid[MAX_LEN_OF_SSID]; // NOT NULL-terminated

    BSS_TABLE BssTab;     // BSS Table

    // global variables mXXXX used in MAC protocol state machines
    BOOLEAN   Mibss;
    BOOLEAN   Massoc;

    // PHY specification
    PHYMODE   PhyMode;    // PHY_11A, PHY_11G, PHY_11B
    USHORT    Drxtx;      // RF RX to TX turn around time, in units of usec
    USHORT    Dswchnl;    // in units of usec, time to switch to a new channel
    USHORT    Dpreamble;  // in units of usec
    USHORT    Dplcp;      // in units of usec
    USHORT    Dslot;      // in units of usec
    USHORT    Dsifs;      // in units of usec
    USHORT    Dpifs;      // in units of usec
    USHORT    Ddifs;      // in units of usec
    USHORT    Deifs;      // in units of usec, change when BasicRateSet changes
    USHORT    TxPreamble; // Rt802_11PreambleLong, Rt802_11PreambleShort

    ULONG     WindowsPowerMode;			// Power mode for AC power
    ULONG     WindowsBatteryPowerMode;	// Power mode for battery if exists
    BOOLEAN   WindowsACCAMEnable;		// Enable CAM power mode when AC on
    ULONG     PacketFilter;				// Packet filter for receiving
    BOOLEAN   AutoReconnect;			// Set to TRUE when setting OID_802_11_SSID with no matching BSSID
#ifdef NDIS51_MINIPORT   
    ULONG     WindowsPowerProfile;	// Windows power profile, for NDIS5.1 PnP
#endif
    ULONG     WindowsTxPreamble; // Rt802_11PreambleLong, Rt802_11PreambleShort, Rt802_11PreambleAuto

    UCHAR	ChannelTxPower[MAX_LEN_OF_CHANNELS];		// Store Tx power value for all channels.

    // configuration to be used when this STA starts a new ADHOC network
//  IBSS_CONFIG IbssConfig;

    ULONG     LastBeaconRxTime; // OS's timestamp of the last BEACON RX time
    ULONG     LastScanTime;		// Record last scan time for issue BSSID_SCAN_LIST
    ULONG     IgnoredScanNumber;	// Ignored BSSID_SCAN_LIST requests

    LED_CONTROL					LedCntl;
    UCHAR						RfType;         // 0-2420, 1:2421
    UCHAR						LedMode;        // 0-default, 1 TX/RX activity mode
    struct timer_list			RfTuningTimer;
    UCHAR						RfTuningStep;
    UCHAR						RfTuningChannel;

    // the following fields are for dynamic BBP "R13: RX AGC VGC INIT" control
    RT_802_11_RX_AGC_VGC_TUNING	RxVgc;
    USHORT						BbpFalseCca;
    BOOLEAN						EnableRxAgcVgcTuning;
    UCHAR                   RSNIE_Len;
    UCHAR                   RSN_IE[MAX_LEN_OF_RSNIE];
        
    UCHAR                   Key_Counter[32];  //jan for wpa
} PORT_CONFIG, *PPORT_CONFIG;



typedef struct _MLME_STRUCT {
    STATE_MACHINE           CntlMachine, AssocMachine, AuthMachine, AuthRspMachine, SyncMachine, WpaMachine;
    STATE_MACHINE_FUNC      CntlFunc[CNTL_FUNC_SIZE], AssocFunc[ASSOC_FUNC_SIZE];
    STATE_MACHINE_FUNC      AuthFunc[AUTH_FUNC_SIZE], AuthRspFunc[AUTH_RSP_FUNC_SIZE];
    STATE_MACHINE_FUNC      SyncFunc[SYNC_FUNC_SIZE], WpaFunc[WPA_FUNC_SIZE];
    
	AUTH_RSP_AUX            AuthRspAux;
    
    COUNTER_802_11          PrevWlanCounters;
    ULONG                   RoamCqi;  // 0..100, Channel Quality Indication for Roaming
    ULONG                   DrsCqi;   // 0..100, Channel Quality Indication for Dynamic Rate Switching
    ULONG                   TxPER;    // 0..100 %
    ULONG                   RxPER;    // 0..100 %

    BOOLEAN                 Running;
    spinlock_t				TaskLock;
    MLME_QUEUE              Queue;

    UINT                    ShiftReg;
//    PSPOLL_FRAME            PsFr;
    
    struct timer_list		PeriodicTimer;
    ULONG                   PeriodicRound;
    ULONG                   PrevTxCnt;

	UCHAR       TimIELocationInBeacon;
    ULONG       TimBitmap;      // bit0 for broadcast, 1 for AID1, 2 for AID2, ...so on
    UCHAR       DtimCount;      // 0.. DtimPeriod-1
    UCHAR       DtimPeriod;     // default = 3

} MLME_STRUCT, *PMLME_STRUCT;

//
// Management ring buffer format
//
typedef	struct	_MGMT_STRUC	{
	BOOLEAN		Valid;
	PUCHAR		pBuffer;
	ULONG		Length;
}	MGMT_STRUC, *PMGMT_STRUC;

//
// P802.11 Frame control field, 16 bit
//
typedef	struct	_FRAME_CONTROL	{
	USHORT		Ver:2;				// Protocol version
	USHORT		Type:2;				// MSDU type
	USHORT		Subtype:4;			// MSDU subtype
	USHORT		ToDs:1;				// To DS indication
	USHORT		FrDs:1;				// From DS indication
	USHORT		MoreFrag:1;			// More fragment bit
	USHORT		Retry:1;			// Retry status bit
	USHORT		PwrMgt:1;			// Power management bit
	USHORT		MoreData:1;			// More data bit
	USHORT		Wep:1;				// Wep data
	USHORT		Order:1;			// Strict order expected
}	FRAME_CONTROL, *PFRAME_CONTROL;

//
// P802.11 intermediate header format
//
typedef	struct	_CONTROL_HEADER	{
	FRAME_CONTROL	Frame;				// Frame control structure
	USHORT			Duration;			// Duration value
	MACADDR			Addr1;				// Address 1 field
	MACADDR			Addr2;				// Address 2 field
}	CONTROL_HEADER, *PCONTROL_HEADER;

//
// P802.11 header format
//
typedef	struct	_HEADER_802_11	{
	CONTROL_HEADER	Controlhead;
	MACADDR			Addr3;				// Address 3 field
	USHORT			Frag:4;				// Fragment number
	USHORT			Sequence:12;		// Sequence number
}	HEADER_802_11, *PHEADER_802_11;

//
// Receive Tuple Cache Format
//
typedef	struct	_TUPLE_CACHE	{
	BOOLEAN			Valid;
	MACADDR			MAC;
	USHORT			Sequence; 
	USHORT			Frag;
}	TUPLE_CACHE, *PTUPLE_CACHE;

//
// Fragment Frame structure
//
typedef	struct	_FRAGMENT_FRAME	{
	UCHAR		Header802_3[14];
	UCHAR		Header_LLC[8];
	UCHAR		Buffer[MAX_FRAME_SIZE];
	ULONG		RxSize;
	USHORT		Sequence;
	USHORT		LastFrag;
	ULONG		Flags;			// Some extra frame information. bit 0: LLC presented
}	FRAGMENT_FRAME, *PFRAGMENT_FRAME;

//
// Tkip Key structure which RC4 key & MIC calculation
//
typedef	struct	_TKIP_KEY_INFO	{
	UINT		nBytesInM;	// # bytes in M for MICKEY
	ULONG		IV16;
	ULONG		IV32;	
	ULONG		K0;			// for MICKEY Low
	ULONG		K1;			// for MICKEY Hig
	ULONG		L;			// Current state for MICKEY
	ULONG		R;			// Current state for MICKEY
	ULONG		M;			// Message accumulator for MICKEY
	UCHAR		RC4KEY[16];
	UCHAR		MIC[8];
}	TKIP_KEY_INFO, *PTKIP_KEY_INFO;

//
// Private / Misc data, counters for driver internal use
//
typedef	struct	__PRIVATE_STRUC	{
	ULONG		SystemResetCnt;			// System reset counter
	ULONG		TxRingFullCnt;			// Tx ring full occurrance number
	ULONG		ResetCountDown;			// Count down before issue reset
	ULONG		CCAErrCnt;				// CCA error count, for debug purpose, might move to global counter
	ULONG		PhyRxErrCnt;			// PHY Rx error count, for debug purpose, might move to global counter
	ULONG		PhyTxErrCnt;			// PHY Tx error count, for debug purpose, might move to global counter
	// Variables for WEP encryption / decryption in rtmp_wep.c
	ULONG			FCSCRC32;
	ARCFOURCONTEXT	WEPCONTEXT;
	ARCFOURCONTEXT	WPACONTEXT;

	// Tkip stuff
	TKIP_KEY_INFO	Tx;
	TKIP_KEY_INFO	Rx;
}	PRIVATE_STRUC, *PPRIVATE_STRUC;

//
//	All DMA ring formats
//
struct	ring_desc	{
	// Descriptor size & dma address
	u32			size;
	void		*va_addr;
	dma_addr_t	pa_addr;
	// Dma buffer size and address for real transfer
	u32			data_size;
	void		*va_data_addr;
	dma_addr_t	pa_data_addr;
    UCHAR		FrameType;			// Type of frame in ring buffer
};

//
//  The miniport adapter structure
//
typedef struct _RTMP_ADAPTER
{
#if 0    // System Information
    UCHAR                   RevsionID;          // Read from PCI config
    USHORT                  SubVendorID;        // Read from PCI config
    USHORT                  SubSystemID;        // Read from PCI config

    // PCI information resource variable
    ULONG                   IoBaseAddress;      // Read from PCI config
    ULONG                   IoRange;            // Read from PCI config
    ULONG                   InterruptLevel;     // Read from PCI config
    NDIS_PHYSICAL_ADDRESS   MemPhysAddress;     // Read from PCI config
    ULONG                   MemRange;           // Read from PCI config
    PVOID                   PortOffset;         // mapped VA of IO Port
#endif    
    unsigned long	    	CSRBaseAddress;     // PCI MMIO Base Address, all access will use
                                                // NdisReadRegisterXx or NdisWriteRegisterXx

    // configuration
    UCHAR                   PermanentAddress[ETH_LENGTH_OF_ADDRESS];    // Factory default MAC address
    UCHAR                   CurrentAddress[ETH_LENGTH_OF_ADDRESS];      // User changed MAC address

    UCHAR					EEPROMAddressNum;		// 93c46=6	93c66=8
    USHORT					EEPROMDefaultValue[NUM_EEPROM_BBP_PARMS];

    // resource for DMA operation
    struct ring_desc		TxRing[TX_RING_SIZE];       // Tx Ring
    struct ring_desc		AtimRing[ATIM_RING_SIZE];   // Atim Ring
    struct ring_desc		PrioRing[PRIO_RING_SIZE];   // Priority Ring
    struct ring_desc		RxRing[RX_RING_SIZE];       // Rx Ring
    struct ring_desc		BeaconRing;       			// Beacon Ring, only one

    MGMT_STRUC				MgmtRing[MGMT_RING_SIZE];	// management ring size
	
    ULONG					CurRxIndex;					// Next RxD read pointer
    ULONG					CurTxIndex;					// Next TxD write pointer
    ULONG					CurAtimIndex;				// Next AtimD write pointer
    ULONG					CurPrioIndex;				// Next PrioD write pointer
    ULONG					PushMgmtIndex;				// Next SW management ring index
    ULONG					PopMgmtIndex;				// Next SW management ring index
    ULONG					MgmtQueueSize;				// Number of Mgmt request stored in MgmtRing
    ULONG					NextTxDoneIndex;
    ULONG					NextAtimDoneIndex;
    ULONG					NextPrioDoneIndex;

    // 802.3 multicast support
    ULONG					NumberOfMcAddresses;		// Number of mcast entry exists
    UCHAR					McastTable[MAX_MCAST_LIST_SIZE][ETH_LENGTH_OF_ADDRESS];		// Mcast list
   //flags
    ULONG					Flags;                      // Represent current device status

    // SEND queue list
    QUEUE_HEADER            SendTxWaitQueue;
    //QUEUE_HEADER            McastPsQueue;

	USHORT					Sequence;					// Current sequence number

	TUPLE_CACHE				TupleCache[MAX_CLIENT];		// Maximum number of tuple caches, only useful in Ad-Hoc
	FRAGMENT_FRAME			FragFrame;					// Frame storage for fragment frame
	
	// For MiniportTransferData
	PUCHAR					pRxData;					// Pointer to current RxRing offset / fragment frame offset
	
    // Counters for 802.3 & generic.
    // Add 802.11 specific counters later
    COUNTER_802_3           Counters;                   // 802.3 counters
    COUNTER_802_11          WlanCounters;				// 802.11 MIB counters
    COUNTER_RALINK          RalinkCounters;				// Ralink propriety counters

	NDIS_MEDIA_STATE        MediaState;

	PRIVATE_STRUC			PrivateInfo;				// Private information & counters

    // SpinLocks
	spinlock_t				TxRingLock;					// Tx Ring spinlock
	spinlock_t				PrioRingLock;				// Prio Ring spinlock
	spinlock_t				AtimRingLock;				// Atim Ring spinlock
	spinlock_t				RxRingLock;					// Rx Ring spinlock
	spinlock_t				SendTxWaitQueueLock;		// SendTxWaitQueue spinlock

// Boolean control for packet filter
	BOOLEAN					bAcceptDirect;
	BOOLEAN					bAcceptMulticast;
	BOOLEAN					bAcceptBroadcast;
	BOOLEAN					bAcceptAllMulticast;
	
	BOOLEAN					bRaConfig;

    PORT_CONFIG             PortCfg;
    MLME_STRUCT             Mlme;

	MAC_TABLE               MacTab;
	spinlock_t              MacTabLock;
	
    RT_802_11_EVENT_TABLE   EventTab;

    struct pci_dev			*pPci_Dev;
    struct net_device		*net_dev;

    struct timer_list		timer;	// Periodic Media monitoring timer.
    
#if WIRELESS_EXT >= 12
	struct iw_statistics iw_stats;
#endif
	struct net_device_stats stats;
	
#ifdef WDS
	RT_802_11_WDS_TABLE		WdsTab;
#endif
	RT_802_11_FILTER_ENTRY	FilterEntry;

	/* Frequently used values: keep some adjacent for cache effect. */
	int chip_id;
	BOOLEAN					bContinuousTx;
	BOOLEAN					bContinuousRx;
	volatile int	SpinLockCount;
}   RTMP_ADAPTER, *PRTMP_ADAPTER;

//
// Enable & Disable NIC interrupt via writing interrupt mask register
// Since it use ADAPTER structure, it have to be put after structure definition.
//
inline	VOID    NICDisableInterrupt(
    IN  PRTMP_ADAPTER   pAdapter);

inline	VOID    NICEnableInterrupt(
    IN  PRTMP_ADAPTER   pAdapter);

BOOLEAN	NICCheckForHang(
	IN	PRTMP_ADAPTER	pAd);


INT		RT2400_close(
	IN	struct net_device *net_dev);

VOID    RTMPIsr(
    IN	INT				irq, 
	IN	VOID			*dev_instance, 
	IN	struct pt_regs	*rgs);

VOID	RT2400_timer(
	IN	unsigned long data);

INT		RT2400_open(
	IN	struct net_device *net_dev);

INT	RTMPSendPackets(
	IN	struct sk_buff *skb, 
	IN	struct net_device *net_dev);

#ifdef WDS
INT	Wds_open(
	IN	struct net_device *dev);

INT	Wds_close(
	IN	struct net_device *dev);

INT Wds_ioctl(
	IN	struct net_device   *net_dev,
	IN	struct ifreq        *rq, 
	IN	INT                 cmd);

INT	WdsSendPackets(
	IN	struct sk_buff      *skb, 
	IN	struct net_device   *net_dev);
#endif

INT		RT2400_probe(
	IN	struct pci_dev				*pPci_Dev, 
	IN	const struct pci_device_id	*ent);

#if WIRELESS_EXT >= 12
struct iw_statistics *RT2400_get_wireless_stats(
	IN	struct net_device *net_dev);
#endif

struct net_device_stats *RT2400_get_ether_stats(
	IN	struct net_device *net_dev);

VOID	RT2400_set_rx_mode(
	IN	struct net_device *net_dev);

NDIS_STATUS RTMPAllocDMAMemory(
    IN  PRTMP_ADAPTER   pAd);

VOID	RTMPFreeDMAMemory(
	IN  PRTMP_ADAPTER   pAd);

VOID	NICReadEEPROMParameters(
    IN  PRTMP_ADAPTER       pAdapter);

VOID	NICInitAsicFromEEPROM(
	IN  PRTMP_ADAPTER       pAdapter);

VOID    NICInitTransmit(
    IN  PRTMP_ADAPTER   pAdapter);

NDIS_STATUS	NICReadAdapterInfo(
	IN	PRTMP_ADAPTER		pAdapter);

VOID	NICInitializeAdapter(
    IN  PRTMP_ADAPTER   pAdapter);

VOID	NICInitializeAsic(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	NICIssueReset(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	PortCfgInit(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	NICResetFromError(
	IN	PRTMP_ADAPTER	pAdapter);

INT	RTMPIsFindSection(
	IN	PUCHAR	ptr,
	IN	PUCHAR	buffer);

PUCHAR	RTMPFindSection(
	IN	PCHAR	buffer,
	IN	PCHAR	section);

INT RTMPGetKeyParameter(
	IN	PUCHAR	section,
	IN	PCHAR	key,
	OUT	PCHAR	dest,	
	IN	INT		destsize,
	IN	PCHAR	buffer);

VOID	RTMPReadParametersFromFile(
	IN	PRTMP_ADAPTER	pAd);
	
ULONG	RTMPEqualMemory(
	IN	PVOID	pSrc1,
	IN	PVOID	pSrc2,
	IN	ULONG	Length);

VOID	NdisZeroMemory(
	IN	PVOID	pSrc,
	IN	ULONG	Length);

VOID	NdisMoveMemory(
	OUT	PVOID	pDest,
	IN	PVOID	pSrc,
	IN	ULONG	Length);

void AtoH(char * src, UCHAR * dest, int destlen);
UCHAR BtoH(char ch);


//
// Private routines in rtmp_data.c
//
VOID	RTMPHandleRxDoneInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHandleTxRingTxDoneInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHandlePrioRingTxDoneInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHandleAtimRingTxDoneInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHandleTbcnInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHandleTwakeupInterrupt(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPHardTransmitDone(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PTXD_STRUC		pTxD,
	IN	UCHAR			FrameType);

NDIS_STATUS	RTMPSendPacket(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	struct sk_buff *skb);
	
VOID	RTMPDeQueuePacket(
    IN	PRTMP_ADAPTER	pAdapter,
    IN	PQUEUE_HEADER	pQueue);

NDIS_STATUS	RTMPHardTransmit(
	IN	PRTMP_ADAPTER	pAd,
	IN	struct sk_buff *skb,
	IN	UCHAR			NumberRequired);

NDIS_STATUS	RTMPFreeDescriptorRequest(
	IN		PRTMP_ADAPTER	pAdapter,
	IN		UCHAR			RingType,
	IN		UCHAR			NumberRequired);

VOID	MlmeHardTransmit(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuffer,
	IN	ULONG			Length);

USHORT	RTMPCalcDuration(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	UCHAR			Rate,
	IN	ULONG			Size);

VOID	RTMPWriteTxDescriptor(
	IN	PTXD_STRUC	pTxD,
	IN	BOOLEAN		Ack,
	IN	BOOLEAN		Fragment,
	IN  BOOLEAN     InsTimestamp,
	IN	UCHAR		RetryMode,
	IN	UCHAR		Ifs,
	IN	UINT		Rate,
	IN	UCHAR		Service,
	IN	UINT		Length,
	IN  USHORT      TxPreamble);

BOOLEAN	RTMPSearchTupleCache(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PHEADER_802_11	pHeader);

VOID	RTMPUpdateTupleCache(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PHEADER_802_11	pHeader);

VOID    RTMPSuspendMsduTransmission(
   	IN	PRTMP_ADAPTER	pAdapter);

VOID    RTMPResumeMsduTransmission(
   	IN	PRTMP_ADAPTER	pAdapter);

NDIS_STATUS	MiniportMMRequest(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuffer,
	IN	ULONG			Length);

NDIS_STATUS	RTMPCheckRxDescriptor(
	IN	PRXD_STRUC		pRxD);

NDIS_STATUS RTMPCheckDuplicateFrame(
   	IN	PRTMP_ADAPTER	pAdapter,
	IN  PHEADER_802_11	pHeader);

BOOLEAN RTMPCheckClass2Class3Error(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PHEADER_802_11	pHeader);

VOID RTMPHandleRxPsPoll(
   	IN	PRTMP_ADAPTER	pAdapter,
	IN  PMACADDR	    pAddr,
	IN  USHORT          Aid);

//
// Private routines in rtmp_wep.c
//
VOID	RTMPInitWepEngine(
	IN	PRTMP_ADAPTER	pAdapter,	
	IN	PUCHAR			pKey,
	IN	UCHAR			KeyId,
	IN	UCHAR			KeyLen, 
	IN	PUCHAR			pDest);

VOID	RTMPEncryptData(
	IN	PRTMP_ADAPTER	pAdapter,	
	IN	PUCHAR			pSrc,
	IN	PUCHAR			pDest,
	IN	UINT			Len);

BOOLEAN	RTMPDecryptData(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PUCHAR			pSrc,
	IN	UINT			Len);

VOID	RTMPSetICV(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PUCHAR			pDest);

VOID	ARCFOUR_INIT(
	IN	PARCFOURCONTEXT	Ctx,
	IN	PUCHAR			pKey,
	IN	UINT			KeyLen);

UCHAR	ARCFOUR_BYTE(
	IN	PARCFOURCONTEXT		Ctx);

VOID	ARCFOUR_DECRYPT(
	IN	PARCFOURCONTEXT	Ctx,
	IN	PUCHAR			pDest, 
	IN	PUCHAR			pSrc,
	IN	UINT			Len);

VOID	ARCFOUR_ENCRYPT(
	IN	PARCFOURCONTEXT	Ctx,
	IN	PUCHAR			pDest,
	IN	PUCHAR			pSrc,
	IN	UINT			Len);

VOID	WPAARCFOUR_ENCRYPT(
	IN	PARCFOURCONTEXT	Ctx,
	IN	PUCHAR			pDest,
	IN	PUCHAR			pSrc,
	IN	UINT			Len);

ULONG	RTMP_CALC_FCS32(
	IN	ULONG	Fcs,
	IN	PUCHAR	Cp,
	IN	INT		Len);

//
// MLME routines
//

// Asic/RF/BBP related functions

VOID    AsicSwitchChannel(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN	ULONG Channel);

VOID    AsicLockChannel(
    IN PRTMP_ADAPTER pAd, 
    IN ULONG Channel) ;

VOID AsicRfTuningExec(
    IN unsigned long data);

VOID    AsicSleepThenAutoWakeup(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  USHORT TbttNumToNextWakeUp);

VOID    AsicForceSleep(
    IN PRTMP_ADAPTER pAdapter);

VOID    AsicForceWakeup(
    IN PRTMP_ADAPTER pAdapter);

VOID    AsicSetBssid(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MACADDR *Bssid);

VOID    AsicDisableSync(
    IN	PRTMP_ADAPTER	pAdapter);

VOID    AsicEnableBssSync(
    IN	PRTMP_ADAPTER	pAdapter);

VOID    AsicLedPeriodicExec(
    IN unsigned long data);

VOID    MgtMacHeaderInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN OUT PMACHDR Hdr, 
    IN UCHAR Subtype, 
    IN PMACADDR pAddr1);

NDIS_STATUS  MlmeQueueInit(
    IN MLME_QUEUE *Queue);

VOID  MlmeQueueDestroy(
    IN MLME_QUEUE *Queue);

BOOLEAN MlmeEnqueue(
    IN	PRTMP_ADAPTER	pAd,
    OUT MLME_QUEUE *Queue, 
    IN ULONG Machine, 
    IN ULONG MsgType, 
    IN ULONG MsgLen, 
    IN VOID *Msg);

BOOLEAN MlmeEnqueueForRecv(
    IN	PRTMP_ADAPTER	pAdapter, 
    OUT MLME_QUEUE *Queue, 
    IN UCHAR Rssi, 
    IN ULONG MsgLen, 
    IN PVOID Msg);

BOOLEAN MlmeDequeue(
    IN	PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue, 
    OUT MLME_QUEUE_ELEM **Elem);

BOOLEAN  MlmeQueueEmpty(
    IN PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue);

BOOLEAN  MlmeQueueFull(
    IN	PRTMP_ADAPTER	pAd,
    IN MLME_QUEUE *Queue);

BOOLEAN  MsgTypeSubst(
    IN MACFRAME *Fr, 
    OUT INT *Machine, 
    OUT INT *MsgType);

VOID StateMachineInit(
    IN STATE_MACHINE *Sm, 
    IN STATE_MACHINE_FUNC Trans[], 
    IN ULONG StNr, 
    IN ULONG MsgNr, 
    IN STATE_MACHINE_FUNC DefFunc, 
    IN ULONG InitState, 
    IN ULONG Base);

VOID StateMachineSetAction(
    IN STATE_MACHINE *S, 
    IN ULONG St, 
    ULONG Msg, 
    IN STATE_MACHINE_FUNC F);

VOID StateMachinePerformAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN STATE_MACHINE *S, 
    IN MLME_QUEUE_ELEM *Elem);

VOID Drop(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN MLME_QUEUE_ELEM *Elem);

VOID StateMachineDestroy(
    IN STATE_MACHINE *Sm);

VOID  AssocStateMachineInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  STATE_MACHINE *Sm, 
    OUT STATE_MACHINE_FUNC Trans[]);

VOID  ReassocTimeout(
    IN unsigned long data);

VOID  AssocTimeout(
    IN unsigned long data);

VOID  DisassocTimeout(
    IN unsigned long data);

//----------------------------------------------
VOID  MlmeDisassocReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID  PeerAssocReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID  PeerReassocReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID  PeerDisassocReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID  Cls3errAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR pAddr);

VOID  ComposePsPoll(
    IN	PRTMP_ADAPTER	pAdapter);

VOID  AssocPostProc(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MACADDR *Addr2, 
    IN  USHORT CapabilityInfo, 
    IN  USHORT Aid, 
    IN  UCHAR Rates[], 
    IN  UCHAR RatesLen);

USHORT AssignAid(
    IN PRTMP_ADAPTER pAdapter);

USHORT BuildAssociation(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR      pAddr,
    IN USHORT        CapabilityInfo,
    IN UCHAR *RSN,
    IN UCHAR *pRSNLen, 
    OUT USHORT       *pAid);

VOID AuthStateMachineInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN PSTATE_MACHINE sm, 
    OUT STATE_MACHINE_FUNC Trans[]);

VOID MlmeAuthReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID PeerAuthRspAtSeq2Action(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID PeerAuthRspAtSeq4Action(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID Cls2errAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR pAddr);

VOID MlmeDeauthReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID InvalidStateWhenAuth(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

//=============================================

VOID AuthRspStateMachineInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PSTATE_MACHINE Sm, 
    IN  STATE_MACHINE_FUNC Trans[]);

VOID PeerAuthAtAuthRspIdleAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID PeerDeauthReqAction(
    IN PRTMP_ADAPTER Adaptor, 
    IN MLME_QUEUE_ELEM *Elem);

VOID PeerAuthSimpleRspGenAndSend(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACHDR Hdr, 
    IN  USHORT Alg, 
    IN  USHORT Seq, 
    IN  USHORT StatusCode);

BOOLEAN	CheckFilterTable(
    IN PRTMP_ADAPTER	pAd,
    IN MACADDR		*pAddr);

//========================================

VOID SyncStateMachineInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  STATE_MACHINE *Sm, 
    OUT STATE_MACHINE_FUNC Trans[]);

VOID BeaconTimeout(
    IN unsigned long data);

VOID AtimTimeout(
    IN unsigned long data);

VOID ScanTimeout(
    IN unsigned long data);

VOID MlmeScanReqAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID InvalidStateWhenScan(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID InvalidStateWhenJoin(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID InvalidStateWhenStart(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

//=========================================

VOID ScanTimeoutAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID BeaconTimeoutAtJoinAction(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID PeerProbeReqAction(
    IN  PRTMP_ADAPTER pAd, 
    IN  MLME_QUEUE_ELEM *Elem);

VOID ScanNextChannel(
    IN	PRTMP_ADAPTER	pAdapter);

ULONG MakeBssBeacon(
    IN	PRTMP_ADAPTER	pAdapter);

BOOLEAN StartParmSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT CHAR Ssid[], 
    OUT UCHAR *Ssidlen);

BOOLEAN PeerAuthSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr, 
    OUT USHORT *Alg, 
    OUT USHORT *Seq, 
    OUT USHORT *Status, 
    OUT CHAR ChlgText[]);

BOOLEAN PeerAssocReqSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr2,
    OUT USHORT *CapabilityInfo, 
    OUT USHORT *ListenInterval,
    OUT UCHAR *SsidLen,
    OUT char *Ssid,
    OUT UCHAR *RatesLen,
    OUT UCHAR Rates[],
    OUT UCHAR *RSN,
    OUT UCHAR *pRSNLen);

BOOLEAN PeerReassocReqSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr2, 
    OUT USHORT *CapabilityInfo, 
    OUT USHORT *ListenInterval, 
    OUT MACADDR *ApAddr,
    OUT UCHAR *SsidLen,
    OUT char *Ssid,
    OUT UCHAR *RatesLen,
    OUT UCHAR Rates[],
    OUT UCHAR *RSN,
    OUT UCHAR *pRSNLen);

BOOLEAN PeerDisassocReqSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr2, 
    OUT USHORT *Reason);

BOOLEAN PeerDeauthReqSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr2, 
    OUT USHORT *Reason);

BOOLEAN PeerProbeReqSanity(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  ULONG MsgLen, 
    OUT MACADDR *Addr2,
    OUT CHAR Ssid[], 
    OUT UCHAR *SsidLen, 
    OUT UCHAR Rates[], 
    OUT UCHAR *RatesLen);

ULONG MakeOutgoingFrame(
    OUT CHAR *Buffer, 
    OUT ULONG *Length, ...);

VOID  LfsrInit(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  ULONG Seed);

UCHAR RandomByte(
    IN	PRTMP_ADAPTER	pAdapter);

VOID  MlmePeriodicExec(
    IN unsigned long data);

VOID MlmeAutoScan(
    IN PRTMP_ADAPTER pAdapter);

VOID MlmeAutoRecoverNetwork(
    IN PRTMP_ADAPTER pAdapter);

VOID MlmeAutoReconnectLastSSID(
    IN PRTMP_ADAPTER pAdapter);

VOID MlmeCheckForRoaming(
    IN PRTMP_ADAPTER pAdapter,
    IN ULONG    Now32);

VOID MlmeCheckChannelQuality(
    IN PRTMP_ADAPTER pAdapter,
    IN ULONG Now);

VOID MlmeSetTxPreamble(
    IN PRTMP_ADAPTER pAdapter, 
    IN USHORT TxPreamble);

VOID MlmeUpdateTxRates(
    IN PRTMP_ADAPTER pAdapter);

NDIS_STATUS MlmeInit(
    IN	PRTMP_ADAPTER	pAdapter);

VOID MlmeHandler(
    IN	PRTMP_ADAPTER	pAdapter);

VOID MlmeHalt(
    IN	PRTMP_ADAPTER	pAdapter);

VOID MlmeConfirm(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  VOID *Msg, 
    IN  INT MsgLen);

UCHAR FirstChannel(
    IN	PRTMP_ADAPTER	pAdapter);

UCHAR NextChannel(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  UCHAR channel);

INT RT2400_ioctl(
	IN	struct net_device *net_dev,
	IN	struct ifreq *rq, 
	IN	INT cmd);

VOID RaiseClock(
    IN	PRTMP_ADAPTER	pAd,
    IN  ULONG *x);

VOID LowerClock(
    IN	PRTMP_ADAPTER	pAd,
    IN  ULONG *x);

USHORT ShiftInBits(
    IN	PRTMP_ADAPTER	pAd);

VOID ShiftOutBits(
    IN	PRTMP_ADAPTER	pAd,
    IN  USHORT data,
    IN  USHORT count);

VOID EEpromCleanup(
    IN	PRTMP_ADAPTER	pAd);

VOID EWDS(
	IN	PRTMP_ADAPTER	pAd);

VOID EWEN(
	IN	PRTMP_ADAPTER	pAd);
    
USHORT RTMP_EEPROM_READ16(
    IN	PRTMP_ADAPTER	pAd,
    IN  USHORT Offset);

VOID RTMP_EEPROM_WRITE16(
    IN	PRTMP_ADAPTER	pAd,
    IN  USHORT Offset,
    IN  USHORT Data);
    
// ------------------------------------------------------------
NDIS_STATUS ApInitialize(
    IN  PRTMP_ADAPTER   pAdapter);

VOID ApShutdown(
    IN PRTMP_ADAPTER    pAd);

VOID ApStartUp(
    IN	PRTMP_ADAPTER	pAdapter);

VOID ApStop(
    IN	PRTMP_ADAPTER	pAdapter);

BOOLEAN RTMPBridgeToWirelessSta(
   	IN	PRTMP_ADAPTER	pAdapter,
	IN  PUCHAR          pHeader,
	IN  UINT            HdrLen,
	IN  PUCHAR          pData,
	IN  UINT            DataLen);

VOID CleanupPsQueue(
    IN  PRTMP_ADAPTER   pAdpater,
    IN  PQUEUE_HEADER   pQueue);

VOID MacTableReset(
    IN  PRTMP_ADAPTER   pAd);

MAC_TABLE_ENTRY *MacTableInsertEntry(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR pAddr); 

BOOLEAN MacTableDeleteEntry(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR pAddr);

MAC_TABLE_ENTRY *MacTableLookup(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR pAddr);

MAC_TABLE_ENTRY *SsPsInquiry(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN	MACADDR *pAddr, 
    OUT SST   *Sst, 
    OUT USHORT *Aid,
    OUT UCHAR *PsMode,
    OUT UCHAR *Rate);

BOOLEAN PsIndicate(
    IN	PRTMP_ADAPTER	pAdapter, 
    IN  PMACADDR        pAddr, 
    IN  UCHAR Psm);

VOID ApLogEvent(
    IN PRTMP_ADAPTER    pAd,
    IN PMACADDR         pAddr,
    IN USHORT           Event);

INT RTMPSetInformation(
	IN	PRTMP_ADAPTER       pAdapter,
	IN	OUT	struct ifreq	*rq,
	IN	INT					cmd);

INT RTMPQueryInformation(
	IN	PRTMP_ADAPTER       pAdapter,
	IN	OUT	struct ifreq	*rq,
	IN	INT					cmd);

INT	Set_SSID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_AuthMode_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_EncrypType_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_DefaultKeyID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_Key1_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_Key2_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_Key3_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_Key4_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_WPAPSK_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_RekeyType_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);

INT	Set_RekeyPeriod_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg);
//
// Prototypes of function definition in rtmp_tkip.c
//
VOID	RTMPInitTkipEngine(
	IN	PRTMP_ADAPTER	pAdapter,	
	IN	PUCHAR			pTKey,
	IN	UCHAR			KeyId,
	IN	PUCHAR			pTA,
	IN	PUCHAR			pMICKey,
	IN	PUCHAR			pTSC,
	OUT PULONG			pIV16,
	OUT PULONG			pIV32,
	IN	PUCHAR			pDest);

VOID	RTMPTkipEncryptData(
	IN	PRTMP_ADAPTER	pAdapter,	
	IN	PUCHAR			pSrc,
	IN	PUCHAR			pDest,
	IN	UINT			Len);

BOOLEAN	RTMPTkipDecryptData(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PUCHAR			pSrc,
	IN	PUCHAR			pTA,
	IN	PUCHAR			pKey,
	IN	PUCHAR			pRxTsc,
	IN	UINT			Len);

BOOLEAN	RTMPTkipCompareMICValue(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PUCHAR			pSrc,
	IN	PUCHAR			pDA,
	IN	PUCHAR			pSA,
	IN	PUCHAR			pMICKey,
	IN	UINT			Len);

VOID	RTMPCalculateMICValue(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	struct sk_buff	*skb,
	IN	PUCHAR			pEncap,
	IN	INT				LenEncap,
	IN	PWPA_KEY		pWpaKey);

BOOLEAN	RTMPTkipCompareMICValueWithLLC(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PUCHAR			pLLC,
	IN	PUCHAR			pSrc,
	IN	PUCHAR			pDA,
	IN	PUCHAR			pSA,
	IN	PUCHAR			pMICKey,
	IN	UINT			Len);

NDIS_STATUS	RTMPWPAAddKeyProc(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuf);

NDIS_STATUS	RTMPWPARemoveKeyProc(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuf);

VOID	RTMPWPARemoveAllKeys(
	IN	PRTMP_ADAPTER	pAdapter);

VOID	RTMPInitMICEngine(
	IN	PRTMP_ADAPTER	pAdapter,	
	IN	PUCHAR			pKey,
	IN	PUCHAR			pDA,
	IN	PUCHAR			pSA,
	IN	PUCHAR			pMICKey);

VOID	RTMPTkipAppend( 
	IN	PTKIP_KEY_INFO	pTkip,	
	IN	PUCHAR			pSrc,
	IN	UINT			nBytes);

VOID	RTMPTkipGetMIC( 
	IN	PTKIP_KEY_INFO	pTkip);
#endif  // __RTMP_H__
