/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rtmp_data.c

    Abstract:
    Data path subroutines

    Revision History:
    Who         When            What
    --------    ----------      ----------------------------------------------
    Name        mm/dd/yy        Modification logs
    HK                          Initial
    Paul        10/02/02        Merge & modify
    John            02/03/03        Serve as Access Point
*/
#include "rt_config.h"

const ULONG Bit[32] = {
    0x00000001, 0x00000002, 0x00000004, 0x00000008, 0x00000010, 0x00000020, 0x00000040, 0x00000080,
    0x00000100, 0x00000200, 0x00000400, 0x00000800, 0x00001000, 0x00002000, 0x00004000, 0x00008000,
    0x00010000, 0x00020000, 0x00040000, 0x00080000, 0x00100000, 0x00200000, 0x00400000, 0x00800000,
    0x01000000, 0x02000000, 0x04000000, 0x08000000, 0x10000000, 0x20000000, 0x40000000, 0x80000000 };
        
static  UCHAR   SNAP_802_1H[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};
static  UCHAR   SNAP_BRIDGE_TUNNEL[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8};
static  UCHAR   EAPOL[] = {0x88, 0x8e};

static  UCHAR   IPX[] = {0x81, 0x37};
static  UCHAR   APPLE_TALK[] = {0x80, 0xf3};

/*
    ========================================================================
    Routine Description:
        Process RxDone interrupt, running in DPC level

    Arguments:
        pAdapter    Pointer to our adapter

    Return Value:
        None

    Note:
        This routine has to maintain Rx ring read pointer.
    ========================================================================
*/
VOID    RTMPHandleRxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PRXD_STRUC          pRxD;
    PHEADER_802_11      pHeader;
    PUCHAR              pData;
    PUCHAR              pEncap;
    PVOID               pManage;
    PUCHAR              pDestMac, pSrcMac;
    UCHAR               Header802_3[14];
    UCHAR               LLC_Len[2];
    UINT                PacketSize;
    UCHAR               Count;
    ULONG               i;
    NDIS_STATUS         Status;
    struct net_device   *net_dev = pAdapter->net_dev;
    struct sk_buff      *skb;
    NDIS_802_11_PRIVACY_FILTER   Privacy;
    PWPA_KEY        pWpaKey  = (PWPA_KEY) NULL;
    UCHAR           RxTsc[6];
    WPA_STATE       WpaState;
    MAC_TABLE_ENTRY *pEntry;
    ULONG           wdsidx = 0xFF;
#ifdef WDS
    ULONG           j;
    BOOLEAN         bWdsPacket;
#endif
    // Make sure Rx ring resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->RxRingLock);

    Count = 0;
    do
    {
        // Point to Rx indexed rx ring descriptor
        pRxD = (PRXD_STRUC) pAdapter->RxRing[pAdapter->CurRxIndex].va_addr;

        // In case of false alarm or processed at last instance
        if (pRxD->Owner != DESC_OWN_HOST)
        {
            break;
        }

        if((pAdapter->bContinuousTx == TRUE) || (pAdapter->bContinuousRx == TRUE))
        {
            pRxD->Owner = DESC_OWN_NIC;
            pAdapter->CurRxIndex++;
            if (pAdapter->CurRxIndex >= RX_RING_SIZE)
            {
                pAdapter->CurRxIndex = 0;
            }
            Count++;
            continue;
        }
        
        // Point to Rx ring buffer where stores the real data frame
        pData   = (PUCHAR) (pAdapter->RxRing[pAdapter->CurRxIndex].va_data_addr);
        // Cast to 802.11 header for flags checking
        pHeader = (PHEADER_802_11) pData;

        // Increase Total receive byte counter after real data received no mater any error or not
        pAdapter->RalinkCounters.ReceivedByteCount +=  pRxD->DataByteCnt;
        
        // Check for all RxD errors, add receive counters
        Status = RTMPCheckRxDescriptor(pRxD);
        if (Status == NDIS_STATUS_SUCCESS)
        {
            INC_COUNTER(pAdapter->WlanCounters.ReceivedFragmentCount);
            pAdapter->Counters.GoodReceives++;
        }
        else
        {
            pAdapter->Counters.RxErrors++;
        }

        // All frames to AP are directed except probe_req. IEEE 802.11/1999 - p.463
        // Do this before checking "duplicate frame".
        if (pRxD->U2M)
            ;
        else if ((pHeader->Controlhead.Frame.Type == BTYPE_MGMT) && (pHeader->Controlhead.Frame.Subtype == SUBTYPE_PROBE_REQ))
            ;
        else
            Status = NDIS_STATUS_FAILURE;

        // Gather PowerSave information from all valid frames. IEEE 802.11/1999 p.461
        if ((Status == NDIS_STATUS_SUCCESS) && (pRxD->U2M))
        {
            if (pHeader->Controlhead.Frame.PwrMgt)
                PsIndicate(pAdapter, &pHeader->Controlhead.Addr2, PWR_SAVE);
            else
                PsIndicate(pAdapter, &pHeader->Controlhead.Addr2, PWR_ACTIVE);
        }
        
        // Duplicate frame check. This check only apply to unicast data & management frames
        if ((Status == NDIS_STATUS_SUCCESS) && (pRxD->U2M) && 
            ((pHeader->Controlhead.Frame.Type == BTYPE_DATA) || (pHeader->Controlhead.Frame.Type == BTYPE_MGMT)))
            Status = RTMPCheckDuplicateFrame(pAdapter, pHeader);

        //
        // Do RxD release operation for all failure frames
        //
        if (Status == NDIS_STATUS_SUCCESS)
        {
            // Saved data pointer for management frame which will pass to MLME block
            pManage = (PVOID) pData;
                
            //
            // Start of main loop to parse receiving frames.
            // The sequence will be Type first, then subtype...
            //

            pEntry=PACInquiry(pAdapter, &pHeader->Controlhead.Addr2, &Privacy,&WpaState);
             

#ifdef	WDS
            bWdsPacket = FALSE;
        	if ((pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
        	{
        		for (i = 0; i < pAdapter->WdsTab.Num; i++)
        		{
        			if (MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, &pHeader->Controlhead.Addr2))
        			{
        				net_dev = pAdapter->WdsTab.WdsEntry[i].dev;
        				bWdsPacket = TRUE;
        				DBGPRINT(RT_DEBUG_INFO, "WDS packet from (%d), type = %d!!!\n", i,pHeader->Controlhead.Frame.Type);
            			Privacy = Ndis802_11PrivFilterAcceptAll;
        				break;
        			}
        		}
        		if (i == pAdapter->WdsTab.Num)
        		{
        			DBGPRINT(RT_DEBUG_ERROR, "WDS packet, but not our WDS group!!!\n");
                    RTMPFreeSpinLock(pAdapter, &pAdapter->RxRingLock);
        			//pRxD->Drop == 1;
        			break;
        		}
        	}
            
			if (bWdsPacket)
			{
			    PacketSize = pRxD->DataByteCnt - LENGTH_802_11_WITH_ADDR4;
			}
			else
#endif
			{
				PacketSize = pRxD->DataByteCnt - LENGTH_802_11;
			}
            if (pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled)
            {
                pWpaKey = (PWPA_KEY) NULL;
#ifdef  WDS
                if (bWdsPacket)
                {
        			pWpaKey = (PWPA_KEY) &pAdapter->WdsTab.Wpa_key;
                    DBGPRINT(RT_DEBUG_INFO, "  Rx Use WDS Key, length=%d\n",pAdapter->WdsTab.Wpa_key.KeyLen);
                }
                else
#endif
                {
                    if((pHeader->Controlhead.Addr1.Octet[0] & 0x01))
                    {
                        if (pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0)
                        {
                            pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID];
                            //DBGPRINT(DEBUG_TEMP, ("  Rx Use Group Key\n"));
                        }
                    }
                    else
                    {
                        if (pEntry && (pEntry->PairwiseKey[0].KeyLen != 0))
                        {
                            pWpaKey = (PWPA_KEY) &pEntry->PairwiseKey[0];
                            DBGPRINT(RT_DEBUG_INFO, "Rx Use Pairwise Key \n");
                        }
                        // Use default Group Key if there is no Pairwise key present
                        if ((pWpaKey == NULL) && (pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID].KeyLen != 0))
                            pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID];               
                    }
                }
            }
            switch (pHeader->Controlhead.Frame.Type)
            {
                // Frame with data type
                case BTYPE_DATA:

#ifdef WDS
                    if (pHeader->Controlhead.Frame.ToDs == 0)
                        break;
#else
                    if ((pHeader->Controlhead.Frame.FrDs==1) || (pHeader->Controlhead.Frame.ToDs==0))
                        break;
#endif
                    if(pAdapter->PortCfg.BANClass3Data == TRUE)
                        break;

                    // check if Class2 or 3 error
#ifdef WDS
                    if(!(pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
#endif
                    {
                        if (RTMPCheckClass2Class3Error(pAdapter, pHeader))
                            break;
                    }
                    
 #ifdef  WDS
                    // WDS packet's DA & SA is not the same
                    if (bWdsPacket)
                    {
                        pDestMac = (PUCHAR) &(pHeader->Addr3);
                        pSrcMac = (PUCHAR)pHeader + sizeof(HEADER_802_11);
                        DBGPRINT(RT_DEBUG_INFO,"WDS RX-DA=%02x:%02x:%02x:%02x:%02x:%02x, SA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
                        *(pDestMac+0),*(pDestMac+1),*(pDestMac+2),*(pDestMac+3),*(pDestMac+4),*(pDestMac+5),
                        *(pSrcMac+0),*(pSrcMac+1),*(pSrcMac+2),*(pSrcMac+3),*(pSrcMac+4),*(pSrcMac+5));
                    }
                    else
#endif
                    {
                        pDestMac = (PUCHAR) &(pHeader->Addr3);
                        pSrcMac = (PUCHAR) &(pHeader->Controlhead.Addr2);
                    }
                    switch(Privacy)
                    {
                        case Ndis802_11PrivFilter8021xWEP:
                            if (pEntry == NULL)   // this should not happen
                                break;
                            if (!RTMPCheckWPAframe(pAdapter, (UCHAR *)pHeader, pRxD->DataByteCnt))
                                break;
                            
                            if ((pEntry->Flag&DISASSOC_4_WAY_TMOUT) && (pAdapter->PortCfg.AuthMode >= Ndis802_11AuthModeWPA))
                            {
                                DBGPRINT(RT_DEBUG_TRACE, "HandleRx : GO to DisAssocAction : REASON_4_WAY_TIMEOUT \n");
                                if (pEntry->Flag&DIS_REASON_RSNIE_DIFF)
                                    DisAssocAction( pAdapter, pEntry,SUBTYPE_DISASSOC, REASON_IE_DIFFERENT);
                                else
                                    DisAssocAction( pAdapter, pEntry, SUBTYPE_DISASSOC,REASON_4_WAY_TIMEOUT);
                                break;
                            }

                            if ((pEntry->WpaState < AS_INITPMK) && (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPA))
                            {
                                NdisMoveMemory(LLC_Len, EAPOL, 2);
                                MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                if ((skb = dev_alloc_skb(PacketSize + LENGTH_802_3 + 2)) != NULL)
                                {                                
                                	skb->dev = net_dev;
                                	skb_reserve(skb, 2);	// 16 byte align the IP header
                                	memcpy(skb_put(skb, LENGTH_802_3), Header802_3, LENGTH_802_3);                                        
                                	memcpy(skb_put(skb, PacketSize), pData+LENGTH_802_11, PacketSize);
                                	skb->protocol = eth_type_trans(skb, net_dev);
                                	netif_rx(skb);
                                	net_dev->last_rx = jiffies;
                                	pAdapter->stats.rx_packets++;
                                }
                            }
                            else
                            {
                                MlmeEnqueueForRecv(
                                    pAdapter,
                                    &pAdapter->Mlme.Queue,  
                                    (UCHAR)pRxD->BBR1, 
                                    pRxD->DataByteCnt, 
                                    pManage);
                            }
                            break;
                            
                        case Ndis802_11PrivFilterAcceptAll:
                            // Drop NULL, CF-ACK(no data), CF-POLL(no data), and CF-ACK+CF-POLL(no data) data frame
                            if (pHeader->Controlhead.Frame.Subtype & 0x04) // bit 2 : no DATA
                                break;

                            // pData : Pointer skip the first 24 bytes, 802.11 HEADER
                            // The total available payload should exclude 24-byte 802.11 Header
#ifdef  WDS
            				if (bWdsPacket)
            				{
            					pData += LENGTH_802_11_WITH_ADDR4;
                                DBGPRINT(RT_DEBUG_INFO, "Ndis802_11PrivFilterAcceptAll : GO to   :  \n");
            				}
            				else
#endif
            				{
            					pData += LENGTH_802_11;
            				}
                            // group key timeout checking
                            if( pEntry&& pEntry->Flag&DISASSOC_GRP_UDATE_TMOUT && pAdapter->PortCfg.AuthMode>=Ndis802_11AuthModeWPA)
                            {         
                                DBGPRINT(RT_DEBUG_TRACE, "HandleRx: GO to DisAssocAction : REASON_GROUP_KEY_UPDATE_TIMEOUT\n");
                                DisAssocAction(pAdapter, pEntry, SUBTYPE_DISASSOC, REASON_GROUP_KEY_UPDATE_TIMEOUT);
                            }

                            // check/perform if WEP decryption is required
                            //
                            if (pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled)
                            {
                                if (pHeader->Controlhead.Frame.Wep)
                                {
                                    if (RTMPDecryptData(pAdapter, pData, PacketSize) == FALSE)
                                    {
                                        DBGPRINT(RT_DEBUG_TRACE, "!!! wrong WEP key, decrypt fail !!!\n");
                                        break;
                                    }
                                    PacketSize -= 8; // Minus WEP size
                                }
                            }
                            else if ( (bWdsPacket == FALSE) &&pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled)
                            {
                                // TKIP algorithm
                                if (pHeader->Controlhead.Frame.Wep)
                                {
                                    int i = 0;
                                
                                    if (pWpaKey == NULL)
                                    {
                                        DBGPRINT(RT_DEBUG_ERROR,"  No WPA Key exists for TKIP \n");
                                        Status = NDIS_STATUS_FAILURE;
                                        break;
                                    }

                                    // Start processing TKIP decryption

                                    if (RTMPTkipDecryptData(pAdapter, pData, pHeader->Controlhead.Addr2.Octet, pWpaKey->Key, RxTsc, PacketSize) == FALSE)
                                    {
                                        Status = NDIS_STATUS_FAILURE;
                                        DBGPRINT(RT_DEBUG_ERROR,"  RTMPTkipDecryptData ERROR \n");
                                        break;
                                    }

                                    // Minus IV EIV ICV                             
                                    PacketSize -= 12;
                                    // Second, increase RxTsc value for next transmission
                                    while (++pWpaKey->RxTsc[i] == 0x0)
                                    {
                                        i++;
                                        if (i == 6)
                                            break;
                                    }
                                }
                            }
                            else if ( (bWdsPacket == TRUE) &&pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled)
                            {
                                // TKIP algorithm
                                if (pHeader->Controlhead.Frame.Wep)
                                {
                                    int i = 0;
                                
                                    if (pWpaKey == NULL)
                                    {
                                        DBGPRINT(RT_DEBUG_ERROR,"  No WPA Key exists for TKIP \n");
                                        Status = NDIS_STATUS_FAILURE;
                                        break;
                                    }

                                    // Start processing TKIP decryption

                                    if (RTMPTkipDecryptData(pAdapter, pData, pHeader->Controlhead.Addr2.Octet, pWpaKey->Key, RxTsc, PacketSize) == FALSE)
                                     {   
                                        DBGPRINT(RT_DEBUG_ERROR,"  WDS  RTMPTkipDecryptData ERROR \n");

                                     }
                                    else
                                        DBGPRINT(RT_DEBUG_ERROR,"  WDS  RTMPTkipDecryptData CORRECT \n");
                                    // Minus IV EIV ICV                             
                                    PacketSize -= 12;
                                    // Second, increase RxTsc value for next transmission
                                    while (++pWpaKey->RxTsc[i] == 0x0)
                                    {
                                        i++;
                                        if (i == 6)
                                            break;
                                    }
                                }

                            }
                            else if (pHeader->Controlhead.Frame.Wep)
                            {
                                DBGPRINT(RT_DEBUG_TRACE, "!!! drop WEP frame !!!\n");
                                // Drop WEP frame when PrivacyInvoked is FALSE
                                break;
                            }
    
                            //
                            // re-assembly, if required
                            //
                            if (pHeader->Frag == 0)     // First or Only fragment
                            {
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {
                                    // Check MIC Value right before remove LLC header & pass to upper layer
                                    if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && 
                                            (pHeader->Controlhead.Frame.Wep) )
                                    {
                                        // Minus MIC value, This is plain text length
                                        PacketSize -= 8;
                                        if (bWdsPacket == FALSE){
                                        if (RTMPTkipCompareMICValue(pAdapter, pData, pDestMac, pSrcMac, pWpaKey->RxMic, PacketSize) == FALSE)
                                        {
                                            DBGPRINT(RT_DEBUG_ERROR,"  Rx MIC Value error \n");
                                            HandleCounterMeasure(pAdapter, pEntry);
                                            Status = NDIS_STATUS_FAILURE;
                                            break;
                                        }
                                        }

                                    }
                                }
                                pAdapter->FragFrame.Flags &= 0xFFFFFFFE;

                                // Save encapaturation starting pointer
                                pEncap = pData;
                        
                                // Check for encapsulataion other than RFC1042 & Bridge tunnel
                                if ((!RTMPEqualMemory(SNAP_802_1H, pEncap, 6)) && (!RTMPEqualMemory(SNAP_BRIDGE_TUNNEL, pEncap, 6)))
                                {
                                    LLC_Len[0] = PacketSize / 256;
                                    LLC_Len[1] = PacketSize % 256;
                                    MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                }
                                else
                                {
                                    // Remove 802.11 H header & reconstruct 802.3 header
                                    pData += (LENGTH_802_1_H - LENGTH_802_3_TYPE);
                                    // Patch for WHQl only, which did not turn on Netbios but use IPX within its payload
                                                
                                    if (RTMPEqualMemory(EAPOL, pData, 2)) 
                                    {
                                        PacketSize += LENGTH_802_11;
                                        DBGPRINT(RT_DEBUG_ERROR,"EAPOL packet EthType = %x,%x,packetSize = %d\n",pData[0],pData[1],PacketSize);                                                        
                                        if (pEntry)
                                            MlmeEnqueueForRecv(pAdapter, &pAdapter->Mlme.Queue, (UCHAR)pRxD->BBR1, PacketSize, pManage);
                                        
                                        break;
                                    }

                                    if ((RTMPEqualMemory(IPX, pData, 2) || RTMPEqualMemory(APPLE_TALK, pData, 2)) && RTMPEqualMemory(SNAP_802_1H, pEncap, 6))
                                    {
                                        LLC_Len[0] = PacketSize / 256;
                                        LLC_Len[1] = PacketSize % 256;
                                        pData = pData - LENGTH_802_1_H;
                                        MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                    }
                                    else
                                    {
                                        MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, pData);
                                        // The total available payload should exclude 24-byte 802.11 Header
                                        // and 8-byte 802.2 LLC
                                        PacketSize -= LENGTH_802_1_H;
                                        NdisMoveMemory(pAdapter->FragFrame.Header_LLC, pEncap, 8);
                                        pAdapter->FragFrame.Flags |= 0x01;
                                    }
                                    
                                    // Point to read 802.3 payload
                                    pData += LENGTH_802_3_TYPE;
                                }
                            
                                // One & The only fragment
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {                               
                                    // For miniportTransferData
                                    pAdapter->pRxData = pData;
                                
                                    pAdapter->PortCfg.LedCntl.fRxActivity = TRUE; // for RX ACTIVITY LED
                    
                                    // check if DA is another associted WSTA reachable via wireless bridging,
                                    // if it is, then no need to indicate to LLC
                                    if (RTMPBridgeToWirelessSta(pAdapter, Header802_3, LENGTH_802_3, pData, PacketSize))
                                        break;
                            
                                    // Acknolwdge upper layer the received frame
                                    if ((skb = dev_alloc_skb(PacketSize + LENGTH_802_3 + 2)) != NULL) 
                                    {
                                        skb->dev = net_dev;
                                        skb_reserve(skb, 2);    /* 16 byte align the IP header */
                                        memcpy(skb_put(skb, LENGTH_802_3), Header802_3, LENGTH_802_3);
                                        memcpy(skb_put(skb, PacketSize), pData, PacketSize);
                                        skb->protocol = eth_type_trans(skb, net_dev);
                                        netif_rx(skb);
                                        net_dev->last_rx = jiffies;
                                        pAdapter->stats.rx_packets++;
                                    }
                                    NdisZeroMemory(Header802_3, LENGTH_802_3);
                                }
                                // First fragment of fragmented frames
                                else
                                {
                                    NdisMoveMemory(pAdapter->FragFrame.Buffer,  pData, PacketSize);
                                    NdisMoveMemory(pAdapter->FragFrame.Header802_3, Header802_3, LENGTH_802_3);
                                    NdisZeroMemory(Header802_3, LENGTH_802_3);
                                    pAdapter->FragFrame.RxSize   = PacketSize;
                                    pAdapter->FragFrame.Sequence = pHeader->Sequence;
                                    pAdapter->FragFrame.LastFrag = pHeader->Frag;       // Should be 0
                                }
                            }
                            // Middle & End of fragment burst fragments
                            else
                            {
                                // No LLC-SNAP header in except the first fragment frame
                                    
                                if ((pHeader->Sequence != pAdapter->FragFrame.Sequence) ||
                                    (pHeader->Frag != (pAdapter->FragFrame.LastFrag + 1)))
                                {
                                    // Fragment is not the same sequence or out of fragment number order
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    break;
                                }
                                else if ((pAdapter->FragFrame.RxSize + PacketSize) > MAX_FRAME_SIZE)
                                {
                                    // Fragment frame is too large, it exeeds the maximum frame size.
                                    // We have to drop it.
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    break;
                                }
                                NdisMoveMemory(&pAdapter->FragFrame.Buffer[pAdapter->FragFrame.RxSize], pData, PacketSize);
                                pAdapter->FragFrame.RxSize  += PacketSize;
                                pAdapter->FragFrame.LastFrag = pHeader->Frag;       // Update fragment number
                            
                                // Last fragment
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {
                                    // For miniportTransferData
                                    pAdapter->pRxData = pAdapter->FragFrame.Buffer;
                                
                                    pAdapter->PortCfg.LedCntl.fRxActivity = TRUE; // for RX ACTIVITY LED
                    
                                    // check if DA is another associted WSTA reachable via wireless bridging,
                                    // if it is, then no need to indicate to LLC
                                    if (RTMPBridgeToWirelessSta(pAdapter, pAdapter->FragFrame.Header802_3, LENGTH_802_3, pAdapter->FragFrame.Buffer, pAdapter->FragFrame.RxSize))
                                        break;
                            
                                    // Check MIC Value right before indicate frame to upper layer
                                    if (pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled && (pHeader->Controlhead.Frame.Wep) )
                                    {
                                        // Minus MIC value, This is plain text length
                                        pAdapter->FragFrame.RxSize -= 8;
                    
                                        if ( (pAdapter->FragFrame.Flags & 0x00000001) && (bWdsPacket == FALSE))
                                        {
                                            if (RTMPTkipCompareMICValueWithLLC(
                                                pAdapter,
                                                pAdapter->FragFrame.Header_LLC,
                                                pAdapter->FragFrame.Buffer,
                                                pDestMac,
                                                pSrcMac,
                                                pWpaKey->RxMic,
                                                pAdapter->FragFrame.RxSize) == FALSE)
                                            {
                                                DBGPRINT(RT_DEBUG_ERROR,"Rx MIC Value With LLC error in last fragment (multi fragmets). \n");                          
                                                Status = NDIS_STATUS_FAILURE;
                                                break;
                                            }
                                        }
                                        else if (bWdsPacket == FALSE)
                                        {
                                            if (RTMPTkipCompareMICValue(
                                                pAdapter,
                                                pAdapter->FragFrame.Buffer,
                                                pDestMac,
                                                pSrcMac,
                                                pWpaKey->RxMic,
                                                pAdapter->FragFrame.RxSize) == FALSE)
                                            {
                                                DBGPRINT(RT_DEBUG_ERROR,"Rx MIC Value error in last frag (multi frag) , \n");
                                                Status = NDIS_STATUS_FAILURE;
                                                break;
                                            }
                                        }
                                    }
                                                
                                    // Acknolwdge upper layer the received frame
                                    if ((skb = dev_alloc_skb(pAdapter->FragFrame.RxSize + LENGTH_802_3 + 2)) != NULL)
                                    {
                                        skb->dev = net_dev;
                                        skb_reserve(skb, 2);    /* 16 byte align the IP header */
                                        memcpy(skb_put(skb, LENGTH_802_3), pAdapter->FragFrame.Header802_3, LENGTH_802_3);
                                        memcpy(skb_put(skb, pAdapter->FragFrame.RxSize), pAdapter->FragFrame.Buffer, pAdapter->FragFrame.RxSize);
                                        skb->protocol = eth_type_trans(skb, net_dev);
                                        netif_rx(skb);
                                        net_dev->last_rx = jiffies;
                                        pAdapter->stats.rx_packets++;
                                    }
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    DBGPRINT(RT_DEBUG_INFO, "!!! Frame with Fragment Indicated !!!\n");
                                }
                            }
                            break;
                    }
                    break;

                case BTYPE_MGMT:
                    if(pAdapter->PortCfg.BANClass3Data == TRUE)
                    {
                        DBGPRINT(RT_DEBUG_TRACE, "Disallow new Assciation \n ");
                        // disallow new association
                        MACFRAME    *Fr = (MACFRAME *)pManage;
                        if ((Fr->Hdr.SubType == SUBTYPE_ASSOC_REQ) || (Fr->Hdr.SubType == SUBTYPE_AUTH))
                            break;
                    }
                   
                    // Enqueue this frame to MLME engine
                    if (pRxD->U2M || (pRxD->Bcast && pHeader->Controlhead.Frame.Subtype == SUBTYPE_PROBE_REQ))
                    {
                        if (pHeader->Controlhead.Frame.Wep)
                        {
                            UCHAR *pPayload = (UCHAR *)pManage + LENGTH_802_11;

                            DBGPRINT(RT_DEBUG_TRACE, "One Packet received WEP bit on<<<\n");

                            // must be SHARED_KEY AUTH seq#3, decryption starts after MAC header
                            if (RTMPDecryptData(pAdapter, pPayload, pRxD->DataByteCnt - LENGTH_802_11) == FALSE)
                            {
                                DBGPRINT(RT_DEBUG_TRACE, "One Packet drop due to Decrypt fail\n");
                                break;
                            }
                            // Minus WEP size
                            pRxD->DataByteCnt -= 8;
                        }
                        
                        MlmeEnqueueForRecv(
                            pAdapter,
                            &pAdapter->Mlme.Queue,  
                            (UCHAR)pRxD->BBR1, 
                            pRxD->DataByteCnt, 
                            pManage);
                    }               
                    break;

                case BTYPE_CNTL:
                    // handle PS-POLL here
                    if ((pRxD->U2M) && (pHeader->Controlhead.Frame.Subtype == SUBTYPE_PS_POLL))
                    {
                        USHORT Aid = pHeader->Controlhead.Duration & 0x3fff;
                        PMACADDR pAddr = &pHeader->Controlhead.Addr2;
                        
                        DBGPRINT(RT_DEBUG_TRACE,"rcv PS-POLL (AID=%d) from %02x:%02x:%02x:%02x:%02x:%02x\n", 
                            Aid, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5]);
                        RTMPHandleRxPsPoll(pAdapter, pAddr, Aid);
                    }
                    break;

                default :
                    break;
            }
        }   

        // release this RX Ring cell to ASIC
        pRxD->Owner = DESC_OWN_NIC;
        pAdapter->CurRxIndex++;
        if (pAdapter->CurRxIndex >= RX_RING_SIZE)
        {
            pAdapter->CurRxIndex = 0;
        }
        Count++;
        
    }   while (Count < MAX_RX_PROCESS);
    
    // Make sure to release Rx ring resource
    RTMPFreeSpinLock(pAdapter, &pAdapter->RxRingLock);
}

/*
    ========================================================================
    Routine Description:
        Process TxRing TxDone interrupt, running in DPC level
    ========================================================================
*/
VOID    RTMPHandleTxRingTxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PTXD_STRUC      pTxD;
    UCHAR           Count;
    
    // Make sure Tx ring resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->TxRingLock);
    
    Count = 0;
    do
    {
        pTxD = (PTXD_STRUC) (pAdapter->TxRing[pAdapter->NextTxDoneIndex].va_addr);
        // Check for the descriptor ownership
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->Valid == FALSE))
        {
            break;
        }
        
        if(pAdapter->bContinuousTx == TRUE)
        {
            pTxD->Owner = DESC_OWN_NIC;
            pTxD->Valid = TRUE;
        }
        else
        {
                RTMPHardTransmitDone(
                    pAdapter, 
                    pTxD, 
                    pAdapter->TxRing[pAdapter->NextTxDoneIndex].FrameType);
        
        pTxD->Valid = FALSE;        
        }
        
        // Increase Total transmit byte counter after real data sent out
        pAdapter->RalinkCounters.TransmittedByteCount +=  pTxD->DataByteCnt;
        
        pAdapter->NextTxDoneIndex++;
        if (pAdapter->NextTxDoneIndex >= TX_RING_SIZE)
        {
            pAdapter->NextTxDoneIndex = 0;
            // Group Rekey counter based on packets number tramsmitted
            if(   (pAdapter->PortCfg.WPAREKEY.ReKeyMethod==PKT_REKEY) \
                && (pAdapter->PortCfg.REKEYTimerRunning==TRUE ) )
            {   
                pAdapter->PortCfg.REKEYCOUNTER++;
            }
        }
    }   while (++Count < MAX_TX_PROCESS);

    // Make sure to release Tx ring resource
    RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
    
    // Some Tx ring resource freed, check for pending send frame for hard transmit
    if ((pAdapter->bContinuousTx == FALSE) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
    {
        RTMPDeQueuePacket(pAdapter, &pAdapter->SendTxWaitQueue);
    }
}

/*
    ========================================================================
    Routine Description:
        Process Priority ring TxDone interrupt, running in DPC level
    ========================================================================
*/
VOID    RTMPHandlePrioRingTxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PTXD_STRUC      pTxD;
    UCHAR           Count;
    PMGMT_STRUC     pMgmt;
    
    // Make sure Prio ring resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->PrioRingLock); 
    
    Count = 0;
    do
    {
        pTxD = (PTXD_STRUC) (pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].va_addr);
        // Check for the descriptor ownership
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->Valid == FALSE))
        {
            break;
        }
        
        // No need to put in reply for MLME
        RTMPHardTransmitDone(
            pAdapter, 
            pTxD, 
            pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].FrameType);

        // It might happend with no Ndis packet to indicate back to upper layer
        pTxD->Valid = FALSE;        
        
        // Increase Total transmit byte counter after real data sent out
        pAdapter->RalinkCounters.TransmittedByteCount +=  pTxD->DataByteCnt;

        pAdapter->NextPrioDoneIndex++;
        if (pAdapter->NextPrioDoneIndex >= PRIO_RING_SIZE)
        {
            pAdapter->NextPrioDoneIndex = 0;
        }
    }   while (++Count < MAX_TX_PROCESS);

    // Make sure to release Prio ring resource
    RTMPFreeSpinLock(pAdapter, &pAdapter->PrioRingLock);    
    
    if (pAdapter->PushMgmtIndex != pAdapter->PopMgmtIndex)
    {
        if (RTMPFreeDescriptorRequest(pAdapter, PRIO_RING, 1) == NDIS_STATUS_SUCCESS)
        {
            pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PopMgmtIndex];
            if (pMgmt->Valid == TRUE)
            {
                MlmeHardTransmit(pAdapter, pMgmt->pBuffer, pMgmt->Length);
                MlmeConfirm((PVOID)pAdapter, pMgmt->pBuffer, 0);
                pMgmt->Valid = FALSE;
                pAdapter->PopMgmtIndex++;
                pAdapter->MgmtQueueSize--;
                if (pAdapter->PopMgmtIndex >= MGMT_RING_SIZE)
                {
                    pAdapter->PopMgmtIndex = 0;
                }
                     }
        }
        else
        {
            DBGPRINT(RT_DEBUG_INFO, "RTMPHandlePrioRingTxDoneInterrupt --> TxRing not enough\n");
            RTMP_IO_WRITE32(pAdapter,TXCSR0,0x4);
        }
    }
}

/*
    ========================================================================
    Routine Description:
    Arguments:
        Adapter     Pointer to our adapter
    ========================================================================
*/
void    RTMPHandleTbcnInterrupt(IN PRTMP_ADAPTER pAdapter)
{
    PTXD_STRUC pTxD = (PTXD_STRUC)pAdapter->BeaconRing.va_addr;
    MACHDR *pBcnHdr = (MACHDR *)pAdapter->BeaconRing.va_data_addr;
    UCHAR  *pTimIE = (UCHAR *)pBcnHdr + pAdapter->Mlme.TimIELocationInBeacon;
    UCHAR byte0 = (UCHAR)(pAdapter->Mlme.TimBitmap & 0x000000fe);  // skip AID#0
    UCHAR byte1 = (UCHAR)((pAdapter->Mlme.TimBitmap & 0x0000ff00) >> 8);
    UCHAR byte2 = (UCHAR)((pAdapter->Mlme.TimBitmap & 0x00ff0000) >> 16);
    UCHAR byte3 = (UCHAR)((pAdapter->Mlme.TimBitmap & 0xff000000) >> 24);

        // update BEACON frame's sequence number
    pAdapter->Sequence = ((pAdapter->Sequence) + 1) & (MAX_SEQ_NUMBER);
    pBcnHdr->Seq = pAdapter->Sequence;

    // update TIM IE
    if (pAdapter->Mlme.DtimCount == 0)
        pAdapter->Mlme.DtimCount = pAdapter->Mlme.DtimPeriod - 1;
    else
        pAdapter->Mlme.DtimCount -= 1;
    *pTimIE = IE_TIM;
    *(pTimIE + 2) = pAdapter->Mlme.DtimCount;
    *(pTimIE + 3) = pAdapter->Mlme.DtimPeriod;

    if (byte0 || byte1) // there's some backlogged frame for AID 1-15
    {
        *(pTimIE + 4) = 0;      // Virtual TIM bitmap stars from AID #0
        *(pTimIE + 5) = byte0;
        *(pTimIE + 6) = byte1;
        *(pTimIE + 7) = byte2;
        *(pTimIE + 8) = byte3;
        if (byte3)      *(pTimIE + 1) = 7; // IE length
        else if (byte2) *(pTimIE + 1) = 6; // IE length
        else if (byte1) *(pTimIE + 1) = 5; // IE length
        else            *(pTimIE + 1) = 4; // IE length
    }
    else if (byte2 || byte3) // there's some backlogged frame for AID 16-31
    {
        *(pTimIE + 4) = 2;      // Virtual TIM bitmap starts from AID #16
        *(pTimIE + 5) = byte2;
        *(pTimIE + 6) = byte3;
        if (byte3) *(pTimIE + 1) = 5; // IE length
        else       *(pTimIE + 1) = 4; // IE length
    }
    else // no backlogged frames for AID 1-31
    {
        *(pTimIE + 1) = 4; // IE length
        *(pTimIE + 4) = 0;
        *(pTimIE + 5) = 0;
    }
    
    // bit0 means backlogged mcast/bcast
    *(pTimIE + 4) |= (pAdapter->Mlme.TimBitmap & 0x01); 
    
    // adjust BEACON length according to the new TIM
    pTxD->DataByteCnt = pAdapter->Mlme.TimIELocationInBeacon + 2 + *(pTimIE+1); 
    
    // move backlogged bcast/mcast frames from PSQ to TXQ whenever DtimCount==0
    if (pAdapter->Mlme.DtimCount == 0)
    {
        PQUEUE_ENTRY    pEntry;
#if 1
        RTMPAcquireSpinLock(pAdapter, &pAdapter->MacTabLock);
        RTMPAcquireSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
        while (pAdapter->MacTab.McastPsQueue.Head)
        {
            pEntry = RemoveHeadQueue(&pAdapter->MacTab.McastPsQueue);
            InsertTailQueue(&pAdapter->SendTxWaitQueue, pEntry);
            DBGPRINT(RT_DEBUG_TRACE, "DTIM=%d/%d, tx mcast/bcast out...\n",pAdapter->Mlme.DtimCount,pAdapter->Mlme.DtimPeriod);
        }
        RTMPFreeSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
        RTMPFreeSpinLock(pAdapter, &pAdapter->MacTabLock);
        pAdapter->Mlme.TimBitmap &= ~Bit[0]; 
        pAdapter->MacTab.PsQIdleCount = 0;

        // Dequeue one frame from SendTxWait queue and process it
        // There are two place calling dequeue for TX ring.
        // 1. Here, right after queueing the frame.
        // 2. At the end of TxRingTxDone service routine.
        // 3. Within RTMPBridgeToWirelessSta() service routine
        if ((pAdapter->bContinuousTx == FALSE) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
        {
            RTMPDeQueuePacket(pAdapter, &pAdapter->SendTxWaitQueue);
        }   
#endif
    }
}

/*
    ========================================================================
    Routine Description:
        Process all transmit ring Tx Done interrupt, running    in DPC level

    Arguments:
        Adapter     Pointer to our adapter
    ========================================================================
*/
VOID    RTMPHardTransmitDone(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PTXD_STRUC      pTxD,
    IN  UCHAR           FrameType)
{
    switch (pTxD->TxResult)
    {
        case SUCCESS_WITHOUT_RETRY:         // Success without any retry
            // Return send complete status
            DBGPRINT(RT_DEBUG_INFO, "One Packet Complete Success without retry<<<\n");
            if (pTxD->RTS)
            {
                // Increase 802.11 counters
                INC_COUNTER(pAdapter->WlanCounters.RTSSuccessCount);
                pTxD->RTS = 0;
            }
            
            // Increase general counters
            pAdapter->Counters.GoodTransmits++;
            INC_COUNTER(pAdapter->WlanCounters.TransmittedFragmentCount);
            break;
              
        case SUCCESS_WITH_RETRY:            // Success with some retry
            // Return send complete status
            DBGPRINT(RT_DEBUG_INFO, "One Packet Complete Success with retry<<<\n");
            
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.RetryCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            INC_COUNTER(pAdapter->WlanCounters.TransmittedFragmentCount);
            
            // Increase general counters
            pAdapter->Counters.GoodTransmits++;
            
            if (pTxD->RetryCount > 1)
            {
                // Increase 802.11 counters
                INC_COUNTER(pAdapter->WlanCounters.MultipleRetryCount);
                
                // Increase general counters
                pAdapter->Counters.MoreCollisions++;
            }
            else
            {
                // Increase general counters
                pAdapter->Counters.OneCollision++;
            }
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSSuccessCount);
                pTxD->RTS = 0;
            }
            break;

        case FAIL_RETRY_LIMIT:              // Fail on hitting retry count limit
            DBGPRINT(RT_DEBUG_WARN, "One Packet Complete Failed <<<\n");
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FailedCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }
            break;
            
        case FAIL_INVALID:
            DBGPRINT(RT_DEBUG_WARN, "One Packet Complete Failed with invalid bit ???<<<\n");
            
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }
            break;          
            
        case FAIL_OTHER:
        default:
            DBGPRINT(RT_DEBUG_ERROR, "One Packet Complete Failed other ???<<<\n");
            
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FailedCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }
            break;          
    }
}

/*
    ========================================================================
    Routine Description:
        API for MLME to transmit management frame to AP (BSS Mode)
    or station (IBSS Mode)
    
    Arguments:
        pAdapter    Pointer to our adapter
        Buffer      Pointer to  memory of outgoing frame
        Length      Size of outgoing management frame
        
    Return Value:
        NDIS_STATUS_FAILURE
        NDIS_STATUS_PENDING
        NDIS_STATUS_SUCCESS
    ========================================================================
*/
NDIS_STATUS MiniportMMRequest(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PVOID           pBuffer,
    IN  ULONG           Length)
{
    PMGMT_STRUC     pMgmt;
    PHEADER_802_11  pHeader;
    NDIS_STATUS     Status = NDIS_STATUS_SUCCESS;    

    DBGPRINT(RT_DEBUG_INFO, "---> MiniportMMRequest\n");
    // Check management ring free avaliability
    pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PushMgmtIndex];
    
    // This management cell has been occupied
    if (pMgmt->Valid == TRUE)   
    {
        // No Management ring buffer avaliable
        if (pBuffer)
        {
            MlmeConfirm((PVOID)pAdapter, pBuffer, 0);
        }
        Status = NDIS_STATUS_FAILURE; 
        DBGPRINT(RT_DEBUG_WARN, "<--- MiniportMMRequest (error:: MgmtRing full)\n");
        return (Status);
    }
    
    // Insert this request into software managemnet ring
    if (pBuffer)
    {
        pMgmt->pBuffer = pBuffer;       
        pMgmt->Length  = Length;
        pMgmt->Valid   = TRUE;
        pAdapter->PushMgmtIndex++;
        pAdapter->MgmtQueueSize++;
        if (pAdapter->PushMgmtIndex >= MGMT_RING_SIZE)
        {
            pAdapter->PushMgmtIndex = 0;
        }
    }   
    else
    {
        // Null pBuffer, no need to free memory buffer.
        // This should not happen
        DBGPRINT(RT_DEBUG_WARN, "<--- MiniportMMRequest (error:: NULL msg)\n");
        Status = NDIS_STATUS_FAILURE; 
        return (Status);
    }
    
    // Check Free priority queue
    if (RTMPFreeDescriptorRequest(pAdapter, PRIO_RING, 1) == NDIS_STATUS_SUCCESS)
    {
        pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PopMgmtIndex];
        if (pMgmt->Valid == TRUE)
        {
            MlmeHardTransmit(pAdapter, pMgmt->pBuffer, pMgmt->Length);
            MlmeConfirm((PVOID)pAdapter, pMgmt->pBuffer, 0);
            pMgmt->Valid = FALSE;
            pAdapter->PopMgmtIndex++;
            pAdapter->MgmtQueueSize--;
            if (pAdapter->PopMgmtIndex >= MGMT_RING_SIZE)
            {
                pAdapter->PopMgmtIndex = 0;
            }
        }
    }
    else
    {
        DBGPRINT(RT_DEBUG_INFO, "not enough space in PrioRing\n");
        RTMP_IO_WRITE32(pAdapter,TXCSR0,0x4);
    }

    DBGPRINT(RT_DEBUG_INFO, "<--- MiniportMMRequest\n");
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        Copy frame from waiting queue into relative ring buffer and set 
    appropriate ASIC register to kick hardware transmit function
    
    Arguments:
        pAdapter    Pointer to our adapter
        pBuffer     Pointer to  memory of outgoing frame
        Length      Size of outgoing management frame
    ========================================================================
*/
VOID    MlmeHardTransmit(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PVOID           pBuffer,
    IN  ULONG           Length)
{
    PTXD_STRUC      pTxD;
    PUCHAR          pDest;  
    PHEADER_802_11  pHeader_802_11;
    BOOLEAN         AckRequired, InsertTimestamp;
    
    DBGPRINT(RT_DEBUG_INFO, "MlmeHardTransmit\n");
    
    // Make sure Prio ring resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->PrioRingLock);
        
    pDest = (PUCHAR) pAdapter->PrioRing[pAdapter->CurPrioIndex].va_data_addr;              
    pTxD  = (PTXD_STRUC) pAdapter->PrioRing[pAdapter->CurPrioIndex].va_addr;
        
    if (pTxD->Owner == DESC_OWN_NIC)
    {
        // Descriptor owned by NIC. No descriptor avaliable
        // This should not happen since caller guaranteed.
        // Make sure to release Prio ring resource
        RTMPFreeSpinLock(pAdapter, &pAdapter->PrioRingLock);
        return;
    }
    if (pTxD->Valid == TRUE)
    {
        // Ndis packet of last round did not cleared.
        // This should not happen since caller guaranteed.
        // Make sure to release Prio ring resource
        RTMPFreeSpinLock(pAdapter, &pAdapter->PrioRingLock);
        return;
    }
    
    pHeader_802_11           = (PHEADER_802_11) pBuffer;
    InsertTimestamp = FALSE;
    if (pHeader_802_11->Controlhead.Frame.Type == BTYPE_CNTL) // must be PS-POLL
    {
        AckRequired = FALSE;
        pAdapter->PrioRing[pAdapter->CurPrioIndex].FrameType = BTYPE_CNTL;
    }
    else // BTYPE_MGMT or BMGMT_DATA(must be NULL frame)
    {
        pAdapter->PrioRing[pAdapter->CurPrioIndex].FrameType = BTYPE_MGMT;
        pAdapter->Sequence       = ((pAdapter->Sequence) + 1) & (MAX_SEQ_NUMBER);
        pHeader_802_11->Sequence = pAdapter->Sequence;

        if (pHeader_802_11->Controlhead.Addr1.Octet[0] & 0x01) // MULTICAST, BROADCAST
        {
            AckRequired = FALSE;
            pHeader_802_11->Controlhead.Duration = 0;
        }
        else
        {
            AckRequired = TRUE;
            pHeader_802_11->Controlhead.Duration = RTMPCalcDuration(pAdapter, RATE_2, 14);
            if (pHeader_802_11->Controlhead.Frame.Subtype == SUBTYPE_PROBE_RSP)
            {
                InsertTimestamp = TRUE;
            }
        }
    }
    
    NdisMoveMemory(pDest, pBuffer, Length);
   
    // Initialize Priority Descriptor
    // For inter-frame gap, the number is for this frame and next frame
    // For MLME rate, we will fix as 2Mb to match other vendor's implement
    RTMPWriteTxDescriptor(pTxD, AckRequired, FALSE, InsertTimestamp, SHORT_RETRY, IFS_BACKOFF, RATE_2, 4, Length, pAdapter->PortCfg.TxPreamble);

    // Increase & maintain Tx Ring Index
    pAdapter->CurPrioIndex++;
    if (pAdapter->CurPrioIndex >= PRIO_RING_SIZE)
    {
        pAdapter->CurPrioIndex = 0;
    }
        
    // Kick priority ring transmit
    RTMP_IO_WRITE32(pAdapter,TXCSR0,0x4);
    
    // Make sure to release Prio ring resource
    RTMPFreeSpinLock(pAdapter, &pAdapter->PrioRingLock);
}   

/*
    ========================================================================
    Routine Description:
        This routine is used to en-queue outgoing packets when
        there is no enough shread memory
        
    Arguments:
        pAdapter    Pointer to our adapter
        pPacket     Pointer to send packet
        
    Return Value:
        None
    ========================================================================
*/
NDIS_STATUS RTMPSendPacket(
    IN	PRTMP_ADAPTER	pAdapter,
    IN  struct sk_buff *skb)
{
    PVOID           pVirtualAddress;
    UINT            Length;
    UINT            AllowFragSize;
    UCHAR           NumberOfFrag;
    UCHAR           RTSRequired;
    NDIS_STATUS     Status = NDIS_STATUS_FAILURE;

    SST             Sst;
    UCHAR           PsMode, Rate;
    USHORT          Aid;
    MAC_TABLE_ENTRY *pMacEntry;

    DBGPRINT(RT_DEBUG_INFO, "<==== RTMPSendPacket\n");

    pVirtualAddress = skb->data;

    // Check for virtual address allocation, it might fail !!!
    if (pVirtualAddress == NULL)
    {
        // Resourece is low, system did not allocation virtual address
        // return NDIS_STATUS_FAILURE directly to upper layer
        return (Status);
    }

    //
    // Check for multicast or broadcast (First byte of DA)
    //
    if ((*((PUCHAR) pVirtualAddress) & 0x01) != 0)
    {
        // For multicast & broadcast, there is no fragment allowed
        NumberOfFrag = 1;
    }
    else
    {
        // Check for payload allowed for each fragment
        AllowFragSize = (pAdapter->PortCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;

        // Calculate fragments required
        NumberOfFrag = ((skb->len - LENGTH_802_3 + LENGTH_802_1_H) / AllowFragSize) + 1;
        // Minus 1 if the size just match to allowable fragment size
        if (((skb->len - LENGTH_802_3 + LENGTH_802_1_H) % AllowFragSize) == 0)
        {
            NumberOfFrag--;
        }
    }

    // Check for requirement of RTS 
    if (NumberOfFrag > 1)
    {
        // If multiple fragment required, RTS is required only for the first fragment
        // if the fragment size large than RTS threshold
        RTSRequired = (pAdapter->PortCfg.FragmentThreshold > pAdapter->PortCfg.RtsThreshold) ? 1 : 0;
    }
    else
    {
        RTSRequired = (skb->len > pAdapter->PortCfg.RtsThreshold) ? 1 : 0;
    }
    DBGPRINT(RT_DEBUG_INFO, "Number of fragments include RTS :%d\n", NumberOfFrag + RTSRequired);

    // Save framnet number to Ndis packet reserved field
    RTMP_SET_PACKET_FRAGMENTS(skb, NumberOfFrag);

    // Save RTS requirement to Ndis packet reserved field
    RTMP_SET_PACKET_RTS(skb, RTSRequired);

#if 1
    // check if target STA is in power-saving mode. care only those associated STAs
    pMacEntry = SsPsInquiry(pAdapter, pVirtualAddress, &Sst, &Aid, &PsMode, &Rate);

    // 1. multicast/broadcast frames are put to PSQ as long as there's any associated STA in power-save mode
    // 2. if the associted STA in power-save mode, frame also goes to PSQ
    if ((*(CHAR *)pVirtualAddress & 0x01) && pAdapter->MacTab.fAnyStationInPsm)
    {
        // 1. move outgoing frame to power saving queue 
        RTMPAcquireSpinLock(pAdapter, &pAdapter->MacTabLock);
        InsertTailQueue(&pAdapter->MacTab.McastPsQueue, skb);
        RTMPFreeSpinLock(pAdapter, &pAdapter->MacTabLock);

        // 2. mark corresponding TIM bit in outgoing BEACON frame
        pAdapter->Mlme.TimBitmap |= Bit[0];
        DBGPRINT(RT_DEBUG_TRACE, "at least 1 STA in psm, move M/BCAST to PSQ, TIM bitmap=%08x\n", pAdapter->Mlme.TimBitmap);
    }
    else if (pMacEntry && (Sst == SST_ASSOC) && (PsMode == PWR_SAVE))
    {
        // 1. move outgoing frame to power saving queue 
        RTMPAcquireSpinLock(pAdapter, &pAdapter->MacTabLock);
        InsertTailQueue(&pMacEntry->PsQueue, skb);
        RTMPFreeSpinLock(pAdapter, &pAdapter->MacTabLock);

        // 2. mark corresponding TIM bit in outgoing BEACON frame
        pAdapter->Mlme.TimBitmap |= Bit[Aid];
        DBGPRINT(RT_DEBUG_TRACE, "target (AID=%d) in psm, move to PSQ, TIM bitmap=%08x\n", Aid, pAdapter->Mlme.TimBitmap);
    }
    // 3. otherwise, transmit the frame
    else // (PsMode == PWR_ACTIVE) || (PsMode == PWR_UNKNOWN)
#endif  
    {
        // Enqueue Ndis packet to end of Tx wait queue
        RTMPAcquireSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
        InsertTailQueue(&pAdapter->SendTxWaitQueue, skb);
        RTMPFreeSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
    }

    Status = NDIS_STATUS_SUCCESS;
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        To do the enqueue operation and extract the first item of waiting 
        list. If a number of available shared memory segments could meet 
        the request of extracted item, the extracted item will be fragmented
        into shared memory segments.
        
    Arguments:
        pAdapter    Pointer to our adapter
        pQueue      Pointer to Waiting Queue
    ========================================================================
*/
VOID    RTMPDeQueuePacket(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PQUEUE_HEADER   pQueue)
{
    struct sk_buff  *skb;
    UCHAR           FragmentRequired;
    NDIS_STATUS     Status;
    UCHAR           Count = 0;
    
    // Make sure SendTxWait queue resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
    
    // Check queue before dequeue
    while ((pQueue->Head != NULL) && (Count < MAX_TX_PROCESS)) 
    {
        // Reset is in progress, stop immediately
        if (RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS))
        {
            break;
        }
            
        // Dequeue the first entry from head of queue list
        skb = (struct sk_buff *) RemoveHeadQueue(pQueue);

        // Total fragment required = number of fragment + RST if required
        FragmentRequired = RTMP_GET_PACKET_FRAGMENTS(skb) + RTMP_GET_PACKET_RTS(skb);
        
        if (RTMPFreeDescriptorRequest(pAdapter, TX_RING, FragmentRequired) == NDIS_STATUS_SUCCESS)
        {
            // Avaliable ring descriptors are enough for this frame
            // Call hard transmit
            Status = RTMPHardTransmit(pAdapter, skb, FragmentRequired);
            if (Status == NDIS_STATUS_FAILURE)
            {
                // Packet failed due to various Ndis Packet error
                RTMPFreeSkbBuffer(skb);
                break;
            }
            else if (Status == NDIS_STATUS_RESOURCES)
            {
                // Not enough free tx ring, it might happen due to free descriptor inquery might be not correct
                // It also might change to NDIS_STATUS_FAILURE to simply drop the frame
                // Put the frame back into head of queue
                InsertHeadQueue(pQueue, skb);
                break;
            }           
            Count++;
        }   
        else
        {
            InsertHeadQueue(pQueue, skb);
            pAdapter->PrivateInfo.TxRingFullCnt++;
            RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x1);     // rekick to ensure TxPointer not miss
            DBGPRINT(RT_DEBUG_INFO,"RTMPDequeuePacket --> Not enough free Tx Ring descriptor !!!\n");
            break;
        }
    }

    // Release SendTxWaitQueue resources
    RTMPFreeSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
}    

/*
    ========================================================================
    Routine Description:
        This subroutine will scan through releative ring descriptor to find
        out avaliable free ring descriptor and compare with request size.
        
    Arguments:
        pAdapter    Pointer to our adapter
        RingType    Selected Ring
        
    Return Value:
        NDIS_STATUS_FAILURE     Not enough free descriptor
        NDIS_STATUS_SUCCESS     Enough free descriptor
    ========================================================================
*/
NDIS_STATUS RTMPFreeDescriptorRequest(
    IN      PRTMP_ADAPTER   pAdapter,
    IN      UCHAR           RingType,
    IN      UCHAR           NumberRequired)
{
    UCHAR           FreeNumber = 0;
    UINT            Index;
    PTXD_STRUC      pTxD;
    NDIS_STATUS     Status = NDIS_STATUS_FAILURE;

    switch (RingType)
    {
        case TX_RING:
            RTMPAcquireSpinLock(pAdapter, &pAdapter->TxRingLock);
            Index = pAdapter->CurTxIndex;
            do
            {
                pTxD  = (PTXD_STRUC) pAdapter->TxRing[Index].va_addr;
                
                // While Owner bit is NIC, obviously ASIC still need it.
                // If valid bit is TRUE, indicate that TxDone has not process yet
                // We should not use it until TxDone finish cleanup job
                if ((pTxD->Owner == DESC_OWN_HOST) && (pTxD->Valid == FALSE))
                {
                    // This one is free
                    FreeNumber++;
                }
                else
                {
                    break;
                }
                    
                Index++;
                if (Index >= TX_RING_SIZE)              // Wrap around issue
                {
                    Index = 0;
                }
                
            }   while (FreeNumber < NumberRequired);    // Quit here ! Free number is enough !
            
            if (FreeNumber >= NumberRequired)
            {
                Status = NDIS_STATUS_SUCCESS;
            }

            // Make sure to release Tx ring resource
            RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
            break;
            
        case PRIO_RING:
            RTMPAcquireSpinLock(pAdapter, &pAdapter->PrioRingLock);
            Index = pAdapter->CurPrioIndex;
            do
            {
                pTxD  = (PTXD_STRUC) pAdapter->PrioRing[Index].va_addr;
                
                // While Owner bit is NIC, obviously ASIC still need it.
                // If valid bit is TRUE, indicate that TxDone has not process yet
                // We should not use it until TxDone finish cleanup job
                if ((pTxD->Owner == DESC_OWN_HOST) && (pTxD->Valid == FALSE))
                {
                    // This one is free
                    FreeNumber++;
                }
                else
                {
                    break;
                }
                    
                Index++;
                if (Index >= PRIO_RING_SIZE)            // Wrap around issue
                {
                    Index = 0;
                }
                
            }   while (FreeNumber < NumberRequired);    // Quit here ! Free number is enough !
            
            if (FreeNumber >= NumberRequired)
            {
                Status = NDIS_STATUS_SUCCESS;
            }

            // Make sure to release Prio ring resource
            RTMPFreeSpinLock(pAdapter, &pAdapter->PrioRingLock);
            break;

        default:
            break;
    }
    
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        Copy frame from waiting queue into relative ring buffer and set 
    appropriate ASIC register to kick hardware transmit function
        
    Arguments:
        pAdapter        Pointer to our adapter
        PNDIS_PACKET    Pointer to outgoing Ndis frame
        NumberOfFrag    Number of fragment required
        
    Return Value:
        None
    ========================================================================
*/
NDIS_STATUS RTMPHardTransmit(
    IN  PRTMP_ADAPTER       pAdapter,
    IN  struct sk_buff      *skb,
    IN  UCHAR               NumberRequired)
{
    PVOID           pVirtualAddress;
    UCHAR           ThisFrameKeyID = 0;
    UINT            Length;
    UINT            BytesCopied;
    UINT            TxSize;
    UINT            FreeFragSize;
    UINT            RemainSize;
    USHORT          Protocol;
    UCHAR           FragNumber;
    UCHAR           FrameGap;
    HEADER_802_11   Header_802_11;
    PUCHAR          pDest;
    PUCHAR          pSrc;
    ULONG           Iv16;
    ULONG           Iv32;
    PUCHAR          pEncap;
    PTXD_STRUC      pTxD;
    BOOLEAN         StartOfFrame;
    BOOLEAN         EAPOLFrame;
    BOOLEAN         Encapped;
    BOOLEAN         MICFrag;
    PWPA_KEY        pWpaKey  = (PWPA_KEY) NULL;
    MAC_TABLE_ENTRY *pEntry = (MAC_TABLE_ENTRY*)NULL;
#ifdef	WDS	
	BOOLEAN			bWdsPacket = FALSE;
#endif

    // Make sure Tx ring resource won't be used by other threads
    RTMPAcquireSpinLock(pAdapter, &pAdapter->TxRingLock);
	if (skb->data == NULL)
	{
		DBGPRINT(RT_DEBUG_ERROR, "Error, Null skb data buffer!!!\n");

		RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
		return (NDIS_STATUS_FAILURE);
	}
	

#ifdef	WDS	
	if (RTMP_GET_PACKET_NET_DEVICE(skb) >= 0x10)
	{
		DBGPRINT(RT_DEBUG_INFO, "RTMPHardTransmit  :: WDS !!!\n");
		bWdsPacket = TRUE;
	}
	else
	{
		DBGPRINT(RT_DEBUG_INFO, "RTMPHardTransmit  :: not WDS !!!\n");
		bWdsPacket = FALSE;
	}
#endif
    if (pAdapter->PortCfg.BANClass3Data == TRUE)
    {
		RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
        RTMPFreeSkbBuffer(skb);
        return NDIS_STATUS_FAILURE;
    }
    pSrc = (UCHAR*) skb->data;
    Length = skb->len;
    pVirtualAddress = skb->data;
    
    FrameGap = IFS_BACKOFF;     // Default frame gap mode
    // Sequence Number is identical for all fragments belonged to the same frame
    // Sequence is 0 - 4095
    pAdapter->Sequence = ((pAdapter->Sequence) + 1) & (MAX_SEQ_NUMBER);
    pAdapter->TxRing[pAdapter->CurTxIndex].FrameType = BTYPE_DATA;

    // Initialize 802.11 header for each frame
    NdisZeroMemory(&Header_802_11, sizeof(HEADER_802_11));
    
    if ((*((PUCHAR) pVirtualAddress) & 0x01) != 0)  // Multicast or Broadcast
    {
        INC_COUNTER(pAdapter->WlanCounters.MulticastTransmittedFrameCount);
    }

    if (Length < 14)
    {
        DBGPRINT(RT_DEBUG_ERROR,"RTMPHardTransmit --> Ndis Packet buffer error !!!\n");
        // Make sure to release Tx ring resource
        RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
        return (NDIS_STATUS_FAILURE);
    }       
    //
    // Start making 802.11 frame header
    //
#ifdef	WDS
    if (bWdsPacket)
    {
//		char *pSrc = ((char *)pVirtualAddress) + 6; // SA
        // Addr1: RA, Addr2: TA, Addr3: DA, Addr4: SA
        Header_802_11.Controlhead.Frame.FrDs = 1;
        Header_802_11.Controlhead.Frame.ToDs = 1;
        NdisMoveMemory(&Header_802_11.Controlhead.Addr1, &pAdapter->WdsTab.WdsEntry[(RTMP_GET_PACKET_NET_DEVICE(skb) - 0x10)].WdsAddr, ETH_LENGTH_OF_ADDRESS);// RA
        NdisMoveMemory(&Header_802_11.Controlhead.Addr2, &pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);// TA
        NdisMoveMemory(&Header_802_11.Addr3, pVirtualAddress, ETH_LENGTH_OF_ADDRESS);// DA
        if (*((PUCHAR)pVirtualAddress+0) !=0xff && (*((PUCHAR)pVirtualAddress+0) !=0xc2))
		DBGPRINT(RT_DEBUG_INFO,"WDS-DA=%02x:%02x:%02x:%02x:%02x:%02x, SA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
			*((PUCHAR)pVirtualAddress+0),*((PUCHAR)pVirtualAddress+1),*((PUCHAR)pVirtualAddress+2),*((PUCHAR)pVirtualAddress+3),*((PUCHAR)pVirtualAddress+4),*((PUCHAR)pVirtualAddress+5),
			*((PUCHAR)pVirtualAddress+6),*((PUCHAR)pVirtualAddress+7),*((PUCHAR)pVirtualAddress+8),*((PUCHAR)pVirtualAddress+9),*((PUCHAR)pVirtualAddress+10),*((PUCHAR)pVirtualAddress+11));
    }
    else
#endif
    {
        char *pSrc = ((char *)pVirtualAddress) + 6; // SA
        // Addr1: DA, Addr2: BSSID, Addr3: SA
        Header_802_11.Controlhead.Frame.FrDs = 1;
        Header_802_11.Controlhead.Frame.ToDs = 0;
        NdisMoveMemory(&Header_802_11.Controlhead.Addr1, pVirtualAddress, ETH_LENGTH_OF_ADDRESS);
        NdisMoveMemory(&Header_802_11.Controlhead.Addr2, &pAdapter->PortCfg.Bssid, ETH_LENGTH_OF_ADDRESS);
        NdisMoveMemory(&Header_802_11.Addr3, pSrc, ETH_LENGTH_OF_ADDRESS);
    }

    if (pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled)
        pEntry = MacTableLookup(pAdapter, &Header_802_11.Controlhead.Addr1);

    Header_802_11.Sequence = pAdapter->Sequence;        // Sequence number
    Header_802_11.Controlhead.Frame.Type = BTYPE_DATA;  // Frame type
    Header_802_11.Controlhead.Frame.PwrMgt = 0;

    if (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01)
    {
        // No ACK expected for multicast frame      
        Header_802_11.Controlhead.Duration = 0;
    }
    else
    {
        // ACK size is 14 include CRC, and its rate is 2Mb
        Header_802_11.Controlhead.Duration = pAdapter->PortCfg.Dsifs 
            + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxBasicRate, 14);
    }

    // For the purpose to calculate duration for the second last fragment
    RemainSize = skb->len - LENGTH_802_3 + LENGTH_CRC;

    // Add 802.11x protocol check.
    // For non-WPA network, 802.11x message should not encrypt even
    // the privacy is on.
    if (RTMPEqualMemory(EAPOL, ((PUCHAR) pVirtualAddress) + 12, 2))
    {
        EAPOLFrame = TRUE;
    }
    else
    {
        EAPOLFrame = FALSE;
    }
    
    MICFrag = FALSE;
    Encapped = FALSE;
    pEncap = NULL;
    
    pSrc = (PUCHAR) pVirtualAddress;
    Protocol = *(pSrc + 12) * 256 + *(pSrc + 13);
    if (Protocol > 1500)    // CHeck for LLC encaped
    {
        pEncap = SNAP_802_1H;
        Encapped = TRUE;
        if (RTMPEqualMemory(IPX, pSrc + 12, 2))
        {
            pEncap = SNAP_BRIDGE_TUNNEL;
        }
        
        if (RTMPEqualMemory(APPLE_TALK, pSrc + 12, 2))
        {
            pEncap = SNAP_BRIDGE_TUNNEL;
        }
    }

    //
    // Make RTS frame if required
    //
    if (RTMP_GET_PACKET_RTS(skb))
    {
        PCONTROL_HEADER     pControlHeader;
        UINT                RTSFrameSize;
        
        DBGPRINT(RT_DEBUG_TRACE, "Making RTS Frame\n");
        
        pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurTxIndex].va_data_addr;              
        pTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurTxIndex].va_addr;
                
        if (pTxD->Owner == DESC_OWN_NIC)
        {
            // Descriptor owned by NIC. No descriptor avaliable
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        if (pTxD->Valid == TRUE)
        {
            // This might happen when HardTransmit process faster than TxDone interrupt.
            // However, Since we did call free descriptor number check before Tx HardTransmit.
            // This case should not happened. We should return to higher caller and requeue this
            // acket until next Tx hardtransmit. Hopefully this situation is corrected then.
            // Ndis packet of last round did not cleared.
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            pTxD->Valid = FALSE;

            RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        
        pAdapter->TxRing[pAdapter->CurTxIndex].FrameType = BTYPE_CNTL;
        pControlHeader = (PCONTROL_HEADER) pDest;
        NdisZeroMemory(pControlHeader, sizeof(CONTROL_HEADER));

        //FrameControl
        pControlHeader->Frame.Type    = BTYPE_CNTL;
        pControlHeader->Frame.Subtype = SUBTYPE_RTS;        

        // Calculate duration
        // 2 SIFS + CTS + Data Frame size
        if (RTMP_GET_PACKET_FRAGMENTS(skb) > 1)
        {
            // If fragment required, size is maximum fragment size
            RTSFrameSize = pAdapter->PortCfg.FragmentThreshold;
        }
        else
        {
            // Size should be frame with 802.11 header & CRC
#ifdef  WDS
            if (bWdsPacket)
            {
                RTSFrameSize = skb->len + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_1_H + LENGTH_CRC - LENGTH_802_3;
            }
            else
#endif
            {
                RTSFrameSize = skb->len + LENGTH_802_11 + LENGTH_802_1_H + LENGTH_CRC - LENGTH_802_3;
            }
            
                        // Decrease the LLC header size if it's already encapped from protocol layer
            if (Encapped == FALSE)
            {
                RTSFrameSize -= LENGTH_802_11;
            }

            // Add for WPA support
            if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) && 
                (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
            {
                // WEP encryption
                RTSFrameSize += 8;
            }
            else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && 
                (pAdapter->PortCfg.WpaGTKState == SETKEYS_DONE))
            {
                // TKIP encryption
                RTSFrameSize += 12;     // IV + EIV + ICV, MIC already added to TotalPacketLength
            }
        }
        
        pControlHeader->Duration = 2 * (pAdapter->PortCfg.Dsifs)
            + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.TxRate, RTSFrameSize)
            + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxBasicRate, sizeof(CONTROL_HEADER)); 

        // RA
#ifdef  WDS
        if (bWdsPacket)
        {
            NdisMoveMemory(&pControlHeader->Addr1, &pAdapter->WdsTab.WdsEntry[(RTMP_GET_PACKET_NET_DEVICE(skb) - 0x10)].WdsAddr, ETH_LENGTH_OF_ADDRESS);
        }
        else
#endif
        {
            NdisMoveMemory(&pControlHeader->Addr1, (PUCHAR) pVirtualAddress, ETH_LENGTH_OF_ADDRESS);
        }
        
        // TA   
        NdisMoveMemory(&pControlHeader->Addr2, pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);

        // Write Tx descriptor
        // Don't kick tx start until all frames are prepared
        // RTS has to set more fragment bit for fragment burst
        RTMPWriteTxDescriptor(pTxD, TRUE, TRUE, FALSE, SHORT_RETRY, FrameGap, pAdapter->PortCfg.MaxBasicRate, 4, sizeof(CONTROL_HEADER), pAdapter->PortCfg.TxPreamble);
        pTxD->RTS = 1;
        
        FrameGap = IFS_SIFS;        // Init frame gap for coming data after RTS
        NumberRequired--;
        
        // Increase & maintain Tx Ring Index
        pAdapter->CurTxIndex++;
        if (pAdapter->CurTxIndex >= TX_RING_SIZE)
        {
            pAdapter->CurTxIndex = 0;
        }
    }

    StartOfFrame = TRUE;

    // Find the WPA key, either Group or Pairwise Key
    if (pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled)
    {
        pWpaKey = (PWPA_KEY) NULL;

        if( (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01))
        {
            if (pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID].KeyLen != 0)
            {
                pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID];
                ThisFrameKeyID = pAdapter->PortCfg.WPAGKeyID;
            }
        }
        else
        { 
            if (pEntry && (pEntry->PairwiseKey[0].KeyLen != 0))
            {
                pWpaKey = (PWPA_KEY) &pEntry->PairwiseKey[0];                                    
                ThisFrameKeyID = 0;
            }

            // Use default Group Key if there is no Pairwise key present
            if ((pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID].KeyLen != 0) && (pWpaKey == NULL))
            {
                pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.GroupKey[pAdapter->PortCfg.WPAGKeyID];
                ThisFrameKeyID = pAdapter->PortCfg.WPAGKeyID;
            }
        }

#ifdef  WDS
        if (bWdsPacket)
        {
            pWpaKey = (PWPA_KEY) &pAdapter->WdsTab.Wpa_key;
        }
#endif
    }

    // Start Copy Packet into Ring buffer.
    // For frame required more than one ring buffer (fragment), all ring buffers
    // have to be filled before kicking start tx bit.
    do
    {
        // Get the Tx Ring descriptor & Dma Buffer address
        pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurTxIndex].va_data_addr;              
        pTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurTxIndex].va_addr;
        // Maximum allowable payload with one ring buffer, bound by fragment size
        FreeFragSize = pAdapter->PortCfg.FragmentThreshold - LENGTH_CRC;
        
        if (pTxD->Owner == DESC_OWN_NIC)
        {
            // Descriptor owned by NIC. No descriptor avaliable
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        if (pTxD->Valid == TRUE)
        {
            // Ndis packet of last round did not cleared.
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            pTxD->Valid = FALSE;
            
            RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }

        // Make fragment number & more fragment bit of 802.11 header
        if (StartOfFrame == TRUE)
        {
            Header_802_11.Frag = 0;         // Start of fragment burst / Single Frame
        }
        else
        {
            Header_802_11.Frag++;           // Rest of fragmented frames.
        }
        
        // Turn on with no frames after this one
        if (NumberRequired > 1)
        {
            ULONG FragSize;
                    
            Header_802_11.Controlhead.Frame.MoreFrag = 1;
            // Duration should include next fragment
            if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) &&
            (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
                FragSize = pAdapter->PortCfg.FragmentThreshold + 8;
            else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                FragSize = pAdapter->PortCfg.FragmentThreshold + 12;
            else
                FragSize = pAdapter->PortCfg.FragmentThreshold;

            Header_802_11.Controlhead.Duration = 3 * pAdapter->PortCfg.Dsifs
                    + 2 * RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxBasicRate, 14)      
                    + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxTxRate, FragSize);
        }
        else
        {
            Header_802_11.Controlhead.Frame.MoreFrag = 0;
            // ACK size is 14 include CRC, and its rate is 2Mb
            Header_802_11.Controlhead.Duration = pAdapter->PortCfg.Dsifs 
                    + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxBasicRate, 14);
        }

        if (NumberRequired == 2)
        {
            ULONG FragSize;
            
#ifdef  WDS
            if(bWdsPacket)
            {
                RemainSize = RemainSize - pAdapter->PortCfg.FragmentThreshold + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_11 + LENGTH_CRC;// the length of addr4
            }
            else
#endif
            {
                RemainSize = RemainSize - pAdapter->PortCfg.FragmentThreshold + LENGTH_802_11 + LENGTH_802_11 + LENGTH_CRC;
            }
            if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) && (EAPOLFrame == FALSE) &&
            (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
                FragSize = RemainSize + 8;
            else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) &&
            (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))                
                FragSize = RemainSize + 12;
            else
                FragSize = RemainSize;
            
            Header_802_11.Controlhead.Duration = 3 * pAdapter->PortCfg.Dsifs
                + 2 * RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxBasicRate, 14)        
                + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.MaxTxRate, FragSize);

        }

        // Check for WEP enable bit and prepare for software WEP
        if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) &&
            (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
        {
            Header_802_11.Controlhead.Frame.Wep = 1;
        }
        else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
        ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
        {
            Header_802_11.Controlhead.Frame.Wep = 1;
        }
        
        // Copy 802.11 header to Tx ring buffer
#ifdef  WDS
        if(bWdsPacket)
        {
            NdisMoveMemory(pDest, &Header_802_11, sizeof(Header_802_11));
            NdisMoveMemory(pDest + sizeof(Header_802_11), (PUCHAR)pVirtualAddress + ETH_LENGTH_OF_ADDRESS, ETH_LENGTH_OF_ADDRESS);// SA
            pDest        += LENGTH_802_11_WITH_ADDR4;
            FreeFragSize -= (sizeof(Header_802_11) + ETH_LENGTH_OF_ADDRESS);
        }
        else
#endif
        {
            NdisMoveMemory(pDest, &Header_802_11, sizeof(Header_802_11));
            pDest        += sizeof(Header_802_11);
            FreeFragSize -= sizeof(Header_802_11);
        }

        if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) &&
            (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
        {
            RTMPInitWepEngine(
                pAdapter,
                pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].Key,
                pAdapter->PortCfg.DefaultKeyId,
                pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen,
                pDest);
            // Added WEP header
            pDest += 4;
        }
        //
        // Only the first fragment required LLC-SNAP header !!!
        //
        if ((StartOfFrame == TRUE) && (Encapped == TRUE))
        {
            if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) && (EAPOLFrame == FALSE) &&
                (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
            {
                // Copy LLC header
                RTMPEncryptData(pAdapter, pEncap, pDest, 6);
                pDest += 6;
                // Copy protocol type
                pSrc = (PUCHAR) pVirtualAddress;
                RTMPEncryptData(pAdapter, pSrc + 12, pDest, 2);
                pDest += 2;
                // Exclude 802.3 header size, we will recalculate the size at
                // the end of fragment preparation.
                Length -= 14;
                pSrc += 14;
                FreeFragSize -= LENGTH_802_1_H;
            }
            else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
            {
                int i = 0;

                RTMPCalculateMICValue(pAdapter, skb, pEncap, 6, pWpaKey);
                RTMPInitTkipEngine(
                    pAdapter,
                    pWpaKey->Key,
                    ThisFrameKeyID,
                    Header_802_11.Controlhead.Addr2.Octet,
                    pWpaKey->TxMic,
                    pWpaKey->TxTsc,
                    &Iv16,
                    &Iv32,
                    pDest);
                       
                // Added IV & EIV header
                pDest += 8;
                
                // Second, increase TxTsc value for next transmission
                while (++pWpaKey->TxTsc[i] == 0x0)
                {
                    i++;
                    if (i == 6)
                        break;
                }
                if (i == 6)
                {
                    // TODO: TSC has done one full cycle, do re-keying stuff follow specs
                    // Should send a special event microsoft defined to request re-key
                }
                // Copy LLC header
                RTMPTkipEncryptData(pAdapter, pEncap, pDest, 6);
                pDest += 6;
                // Copy protocol type
                pSrc = (PUCHAR) pVirtualAddress;
                RTMPTkipEncryptData(pAdapter, pSrc + 12, pDest, 2);
                pDest += 2;
                // Exclude 802.3 header size, we will recalculate the size at
                // the end of fragment preparation.
                Length -= 14;
                pSrc += 14;
                FreeFragSize -= LENGTH_802_1_H;
            }
            else
            {
                // Copy LLC header
                NdisMoveMemory(pDest, pEncap, 6);
                pDest += 6;
                // Copy protocol type
                pSrc = (PUCHAR) pVirtualAddress;
                NdisMoveMemory(pDest, pSrc + 12, 2);
                pDest += 2;
                // Exclude 802.3 header size, we will recalculate the size at
                // the end of fragment preparation.
                Length -= 14;
                pSrc += 14;
                FreeFragSize -= LENGTH_802_1_H;
            }
        }
        else if ((StartOfFrame == TRUE) && (Encapped == FALSE))
        {
            if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
            {
                int i = 0;

                // Calculate MSDU MIC Value
                RTMPCalculateMICValue(pAdapter, skb, pEncap, 0, pWpaKey);
                    RTMPInitTkipEngine(
                    pAdapter,
                    pWpaKey->Key,
                    ThisFrameKeyID,
                    Header_802_11.Controlhead.Addr2.Octet,
                    pWpaKey->TxMic,
                    pWpaKey->TxTsc,
                    &Iv16,
                    &Iv32,
                    pDest);
                
                // Added IV & EIV header
                pDest += 8;

                // Second, increase TxTsc value for next transmission
                while (++pWpaKey->TxTsc[i] == 0x0)
                {
                    i++;
                    if (i == 6)
                        break;
                }
                if (i == 6)
                {
                    // TODO: TSC has done one full cycle, do re-keying stuff follow specs
                    // Should send a special event microsoft defined to request re-key
                }
            }

            pSrc = (PUCHAR) pVirtualAddress + 14;
            Length -= 14;
        }
        // Middle to Last frame & using TKIP, we have to init TKIP right here
        else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
        ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
        {
            int i = 0;

                RTMPInitTkipEngine(
                    pAdapter,
                    pWpaKey->Key,
                    ThisFrameKeyID,
                    Header_802_11.Controlhead.Addr2.Octet,
                    pWpaKey->TxMic,
                    pWpaKey->TxTsc,
                    &Iv16,
                    &Iv32,
                    pDest);
            // Added IV & EIV header
            pDest += 8;
            // Second, increase TxTsc value for next transmission
            while (++pWpaKey->TxTsc[i] == 0x00)
            {
                i++;
                if (i == 6)
                    break;
            }
            if (i == 6)
            {
                // TODO: TSC has done one full cycle, do re-keying stuff follow specs
                // Should send a special event microsoft defined to request re-key
            }
        }

        // Start copying payload
        BytesCopied = 0;
        do
        {
            if (Length >= FreeFragSize)
            {
                // Copy only the free fragment size, and save the pointer
                // of current buffer descriptor for next fragment buffer.
                if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) && 
                    (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
                {
                    RTMPEncryptData(pAdapter, pSrc, pDest, FreeFragSize);
                }
                else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                {
                    RTMPTkipEncryptData(pAdapter, pSrc, pDest, FreeFragSize);
                }
                else
                {
                    NdisMoveMemory(pDest, pSrc, FreeFragSize);
                }
                BytesCopied += FreeFragSize;
                pSrc        += FreeFragSize;
                pDest       += FreeFragSize;
                Length      -= FreeFragSize;
                break;
            }
            else
            {
                // Copy the rest of this buffer descriptor pointed data
                // into ring buffer.
                if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) && 
                    (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
                {
                    RTMPEncryptData(pAdapter, pSrc, pDest, Length);
                }
                else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                {
                    RTMPTkipEncryptData(pAdapter, pSrc, pDest, Length);
                }
                else
                {
                    NdisMoveMemory(pDest, pSrc, Length);
                }
                BytesCopied  += Length;
                pDest        += Length;
                FreeFragSize -= Length;
            }
            // append TKIP MIC
			if (Header_802_11.Controlhead.Frame.Wep == 1 &&\
				(pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && 
				(MICFrag == FALSE) && (pWpaKey != NULL))
			{
				INT	i;
				
				DBGPRINT(RT_DEBUG_INFO, "Calculated TX MIC value =");  //MIC error.
				for (i = 0; i < 8; i++)
				{
					DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PrivateInfo.Tx.MIC[i]);  //MIC error.
				}
				DBGPRINT(RT_DEBUG_INFO, "\n");  //MIC error.
				if (FreeFragSize >= 8)
				{
					RTMPTkipEncryptData(pAdapter, pAdapter->PrivateInfo.Tx.MIC, pDest, 8);
					BytesCopied  += 8;
					pDest		 += 8;
					FreeFragSize -= 8;
				}
				else
				{
					RTMPTkipEncryptData(pAdapter, pAdapter->PrivateInfo.Tx.MIC, pDest, FreeFragSize);
					BytesCopied  += FreeFragSize;
					pSrc		  = pAdapter->PrivateInfo.Tx.MIC + FreeFragSize;
					pDest		 += FreeFragSize;
					Length       -= FreeFragSize;
					MICFrag       = TRUE;
					break;
				}
			}
        }   while (FALSE);      // End of copying payload
                
        // Real packet size, No 802.1H header for fragments except the first one.
        if ((StartOfFrame == TRUE) && (Encapped == TRUE))
        {
#ifdef  WDS
            if (bWdsPacket)
            {
                TxSize = BytesCopied + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_1_H; // the length of addr4
            }
            else
#endif
            {
                TxSize = BytesCopied + LENGTH_802_11 + LENGTH_802_1_H;
            }
        }
        else
        {
#ifdef  WDS
            if (bWdsPacket)
            {
                TxSize = BytesCopied + LENGTH_802_11_WITH_ADDR4; // contain the length of addr4
            }
            else
#endif
            {
                TxSize = BytesCopied + LENGTH_802_11;
            }
        }
        
        RemainSize = RemainSize - BytesCopied;
        
        if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption1Enabled) &&
            (pAdapter->PortCfg.SharedKey[pAdapter->PortCfg.DefaultKeyId].KeyLen != 0))
        {
            // Append ICV
            RTMPSetICV(pAdapter, pDest);
            TxSize += 8;        // WEP header & trailer
        }
        else if ((pAdapter->PortCfg.WepStatus == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
        ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
        {
            // Append ICV
            RTMPSetICV(pAdapter, pDest);
            TxSize += 12;       // WEP header & trailer
        }
            
        // Prepare Tx descriptors before kicking tx.
        // The BBP register index in Tx descriptor has to be configured too.
        if (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01)
        {
            // Multicast, retry bit is off
            RTMPWriteTxDescriptor(pTxD, FALSE, FALSE, FALSE, SHORT_RETRY, FrameGap, pAdapter->PortCfg.TxRate, 4, TxSize, pAdapter->PortCfg.TxPreamble);
        }
        else
        {
            RTMPWriteTxDescriptor(pTxD, TRUE, FALSE, FALSE, SHORT_RETRY, FrameGap, pAdapter->PortCfg.TxRate, 4, TxSize, pAdapter->PortCfg.TxPreamble);
        }

        // Set frame gap for the rest of fragment burst.
        // It won't matter if there is only one fragment (single fragment frame).
        StartOfFrame = FALSE;
        FrameGap     = IFS_SIFS;
        NumberRequired--;
        
        // Increase & maintain Tx Ring Index
        pAdapter->CurTxIndex++;
        if (pAdapter->CurTxIndex >= TX_RING_SIZE)
        {
            pAdapter->CurTxIndex = 0;
        }
        
    }   while (NumberRequired > 0);

    // Acknowledge protocol send complete of pending packet.
    RTMPFreeSkbBuffer(skb);
    
    // Kick Tx Control Register at the end of all ring buffer preparation
    RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x1);

    // Make sure to release Tx ring resource
    RTMPFreeSpinLock(pAdapter, &pAdapter->TxRingLock);

    return (NDIS_STATUS_SUCCESS);
}

/*
    ========================================================================
    Routine Description:
        Calculates the duration which is required to transmit out frames 
    with given size and specified rate.
        
    Arguments:
        pAdapter        Pointer to our adapter
        Rate            Transmit rate
        Size            Frame size in units of byte
        
    Return Value:
        Duration number in units of usec
    ========================================================================
*/
USHORT  RTMPCalcDuration(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  UCHAR           Rate,
    IN  ULONG           Size)
{
    ULONG   Duration = 0;

    Duration = pAdapter->PortCfg.Dplcp + pAdapter->PortCfg.Dpreamble;
    switch (Rate)
    {
        case RATE_1:
            Duration += (Size * 8);
            break;
            
        case RATE_2:
            Duration += (Size * 4);
            break;
            
        case RATE_5_5:
            Duration += (USHORT) (Size * 16 / 11);
            if (((Size * 16) % 11) != 0)
            {
                Duration++;
            }
            break;
            
        case RATE_11:
            Duration += (USHORT) (Size * 8 / 11);
            if (((Size * 8) % 11) != 0)
            {
                Duration++;
            }
            break;
    }
    return ((USHORT)Duration);
}

/*
    ========================================================================
    Routine Description:
        Calculates the duration which is required to transmit out frames 
    with given size and specified rate.
        
    Arguments:
        pTxD        Pointer to transmit descriptor
        Ack         Setting for Ack requirement bit
        Fragment    Setting for Fragment bit
        RetryMode   Setting for retry mode
        Ifs         Setting for IFS gap
        Rate        Setting for transmit rate
        Service     Setting for service
        Length      Frame length
    ========================================================================
*/
VOID    RTMPWriteTxDescriptor(
    IN  PTXD_STRUC  pTxD,
    IN  BOOLEAN     Ack,
    IN  BOOLEAN     Fragment,
    IN  BOOLEAN     InsTimestamp,
    IN  UCHAR       RetryMode,
    IN  UCHAR       Ifs,
    IN  UINT        Rate,
    IN  UCHAR       Service,
    IN  UINT        Length,
    IN  USHORT      TxPreamble)
{
    UINT    Residual;

    pTxD->Owner       = DESC_OWN_NIC;
    pTxD->Valid       = TRUE;
    pTxD->MoreFrag    = Fragment;
    pTxD->ACK         = Ack;
    pTxD->Timestamp   = InsTimestamp;
    pTxD->RetryMd     = RetryMode;
    pTxD->IFS         = Ifs;
    pTxD->DataByteCnt = Length;

    pTxD->BBCR0 = 0x8500 + (Rate);
    if ((Rate != RATE_1) && (TxPreamble == Rt802_11PreambleShort)) // no short preamble for RATE_1
    {
        pTxD->BBCR0 |= 0x0008;
    }
    pTxD->BBCR1 = 0x8600 + (Service);
    Length += 4;

    if ((Rate == RATE_1) || ( Rate == RATE_2))
    {
        Length = Length * 8 / (Rate + 1);
    }
    else
    {
        Residual = ((Length * 16) % (11 * (1 + Rate - RATE_5_5)));
        Length = Length * 16 / (11 * (1 + Rate - RATE_5_5));
        if (Residual != 0)
        {
            Length++;
        }
        if ((Residual <= (3 * (1 + Rate - RATE_5_5))) && (Residual != 0))
        {
            pTxD->BBCR1 = 0x8684;
    }
    }
    pTxD->BBCR2 = 0x8700 + (Length / 256);
    pTxD->BBCR3 = 0x8800 + (Length % 256);
}

/*
    ========================================================================
    Routine Description:
        Search tuple cache for receive duplicate frame from unicast frames.
        
    Arguments:
        pAdapter        Pointer to our adapter
        pHeader         802.11 header of receiving frame
        
    Return Value:
        TRUE            found matched tuple cache
        FALSE           no matched found
    ========================================================================
*/
BOOLEAN RTMPSearchTupleCache(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    INT Index;

    for (Index = 0; Index < MAX_CLIENT; Index++)
    {
        // Tuple cache filled sequencially. If not valid, the rest following this entry
        // are invalid too.
        if (pAdapter->TupleCache[Index].Valid == FALSE)
        {
            return (FALSE);
        }
        
        if (RTMPEqualMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6) &&
            (pAdapter->TupleCache[Index].Sequence == pHeader->Sequence) &&
            (pAdapter->TupleCache[Index].Frag == pHeader->Frag))
        {
            return (TRUE);
    }
    }
    return (FALSE);
}

/*
    ========================================================================
    Routine Description:
        Update tuple cache for new received unicast frames.
        
    Arguments:
        pAdapter        Pointer to our adapter
        pHeader         802.11 header of receiving frame
    ========================================================================
*/
VOID    RTMPUpdateTupleCache(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    INT Index;

    for (Index = 0; Index < MAX_CLIENT; Index++)
    {
        if (pAdapter->TupleCache[Index].Valid == FALSE)
        {
            // Add new entry
            NdisMoveMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6);
            pAdapter->TupleCache[Index].Sequence = pHeader->Sequence;
            pAdapter->TupleCache[Index].Frag     = pHeader->Frag;
            pAdapter->TupleCache[Index].Valid    = TRUE;
            break;
        }
        if (RTMPEqualMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6))
        {
            // Update old entry
            pAdapter->TupleCache[Index].Sequence = pHeader->Sequence;
            pAdapter->TupleCache[Index].Frag     = pHeader->Frag;
        }
    }
}

/*
    ========================================================================
    Routine Description:
        Suspend MSDU transmission
    ========================================================================
*/
VOID    RTMPSuspendMsduTransmission(
    IN  PRTMP_ADAPTER   pAdapter)
{
    DBGPRINT(RT_DEBUG_TRACE,"SCANNING, suspend MSDU transmission ...\n");
    RTMP_SET_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS);
}

/*
    ========================================================================
    Routine Description:
        Resume MSDU transmission
    ========================================================================
*/
VOID    RTMPResumeMsduTransmission(
    IN  PRTMP_ADAPTER   pAdapter)
{
    DBGPRINT(RT_DEBUG_TRACE,"SCAN done, resume MSDU transmission ...\n");
    RTMP_CLEAR_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS);
   
    // Dequeue Tx queue if Reset is not in progress
    if (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS))
    {
        RTMPDeQueuePacket(pAdapter, &pAdapter->SendTxWaitQueue);
    }
}

/*
    ========================================================================
    Routine Description:
        Check Rx descriptor, return NDIS_STATUS_FAILURE if any error found
    ========================================================================
*/
NDIS_STATUS RTMPCheckRxDescriptor(
    IN  PRXD_STRUC  pRxD)
{
    if (pRxD->PhyErr)
        return NDIS_STATUS_FAILURE;
    else if (pRxD->Crc)
        return NDIS_STATUS_FAILURE;
    else
        return NDIS_STATUS_SUCCESS;
}

/*
  ========================================================================
  Return:
    NDIS_STATUS_SUCCESS - This received unicast frame is NOT a duplicate
    NDIS_STATUS_FAILURE - received frame is duplicated and should be dropped
  ========================================================================
*/
NDIS_STATUS RTMPCheckDuplicateFrame(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    if (pHeader->Controlhead.Frame.Retry)
    {
        if (RTMPSearchTupleCache(pAdapter, pHeader) == TRUE)
        {
            // Found retry frame in tuple cache, Discard this frame / fragment
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FrameDuplicateCount);
            return NDIS_STATUS_FAILURE;
        }
    }
    RTMPUpdateTupleCache(pAdapter, pHeader);
    return NDIS_STATUS_SUCCESS;
}

/*
  ========================================================================
  Description:
    This routine checks if a received frame causes class 2 or class 3
    error, and perform error action (DEAUTH or DISASSOC) accordingly
  ========================================================================
*/
BOOLEAN RTMPCheckClass2Class3Error(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    SST Sst;
    UCHAR PsMode, Rate;
    USHORT Aid;

    SsPsInquiry(pAdapter, &pHeader->Controlhead.Addr2, &Sst, &Aid, &PsMode, &Rate);
    if (Sst == SST_ASSOC)
        ; // okay to receive this DATA frame
    else if (Sst == SST_AUTH)
    {
        Cls3errAction(pAdapter, &pHeader->Controlhead.Addr2);
        return TRUE; 
    }
    else
    {
        Cls2errAction(pAdapter, &pHeader->Controlhead.Addr2);
        return TRUE; 
    }
    return FALSE;
}

/*
  ========================================================================
  Description:
    This routine frees all packets in PSQ that's destined to a specific DA.
    BCAST/MCAST in DTIMCount=0 case is also handled here, just like a PS-POLL 
    is received from a WSTA which has MAC address FF:FF:FF:FF:FF:FF
  ========================================================================
*/
VOID RTMPHandleRxPsPoll(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PMACADDR        pAddr,
    IN  USHORT          Aid)
{
    PQUEUE_ENTRY      pEntry;
    PMAC_TABLE_ENTRY  pMacEntry;
    
    //DBGPRINT(RT_DEBUG_TRACE,("rcv PS-POLL (AID=%d) from %02x:%02x:%02x:%02x:%02x:%02x\n", 
    //    Aid, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5]));

    pMacEntry = MacTableLookup(pAdapter, pAddr);
    if (pMacEntry)
    {
        Aid = pMacEntry->Aid;
        RTMPAcquireSpinLock(pAdapter, &pAdapter->MacTabLock);
        RTMPAcquireSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
        // cleanup all backlogged frames in PSQ
        while (pMacEntry->PsQueue.Head)
        {
            pEntry = RemoveHeadQueue(&pMacEntry->PsQueue);
            InsertTailQueue(&pAdapter->SendTxWaitQueue, pEntry);
            DBGPRINT(RT_DEBUG_TRACE, "rx PSPOLL, tx frame out...\n");
        }
        RTMPFreeSpinLock(pAdapter, &pAdapter->SendTxWaitQueueLock);
        RTMPFreeSpinLock(pAdapter, &pAdapter->MacTabLock);
        if ((Aid > 0) && (Aid < MAX_LEN_OF_MAC_TABLE))
            pAdapter->Mlme.TimBitmap &= (~Bit[Aid]);  
        pMacEntry->PsQIdleCount = 0;

        // Dequeue one frame from SendTxWait queue and process it
        // There are two place calling dequeue for TX ring.
        // 1. Here, right after queueing the frame.
        // 2. At the end of TxRingTxDone service routine.
        // 3. Within RTMPBridgeToWirelessSta() service routine
        if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
        {
            RTMPDeQueuePacket(pAdapter, &pAdapter->SendTxWaitQueue);
        }   
    }
}

/*
    ========================================================================
    Routine Description:
        This function checks if the received frames should be sent out to the 
        wireless media based on DA look-up in the wirelss MAC table.
        
    Arguments:
        pRxD        Pointer to the Rx descriptor
        
    Return Value:
        TRUE -  the input frame is destined to a known wireless STA 
                and had being sent out to the wireless media. It's not 
                necessary to make another copy up to LLC nor Portal function.
        FALSE - the input frame still need to deliver to the upper LLC
    ========================================================================
*/
BOOLEAN RTMPBridgeToWirelessSta(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PUCHAR          pHeader802_3,
    IN  UINT            HdrLen,
    IN  PUCHAR          pData,
    IN  UINT            DataLen)
{
    SST                 Sst;
    USHORT              Aid;
    UCHAR               PsMode, Rate;
    MAC_TABLE_ENTRY     *pEntry;
    BOOLEAN             result = FALSE;
    struct sk_buff      *skb;

    // decide the return value. TRUE if no need to indicate to LLC, FALSE otherwise
    pEntry = SsPsInquiry(pAdapter, (MACADDR *)pHeader802_3, &Sst, &Aid, &PsMode, &Rate);
    if (pHeader802_3[0] & 0x01)
        result = FALSE;
    else if (pEntry && (Sst == SST_ASSOC)) 
        result = TRUE;
    else
        result = FALSE;
    
    if ((pHeader802_3[0] & 0x01) || (pEntry && (Sst == SST_ASSOC)))
    {
        NDIS_STATUS     Status;

        if ((skb = dev_alloc_skb(HdrLen + DataLen + 2)) != NULL)
        {
            skb->len = HdrLen + DataLen;
            memcpy((skb->data), pHeader802_3, HdrLen);
            memcpy((skb->data + HdrLen), pData, DataLen);
        
            // send out the packet
            DBGPRINT(RT_DEBUG_INFO,"WBRG - len=%d, DA=%02x:%02x:%02x:%02x:%02x:%02x, SA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
                HdrLen+DataLen,pHeader802_3[0],pHeader802_3[1],pHeader802_3[2],pHeader802_3[3],pHeader802_3[4],pHeader802_3[5],
                pHeader802_3[6],pHeader802_3[7],pHeader802_3[8],pHeader802_3[9],pHeader802_3[10],pHeader802_3[11]);
            Status = RTMPSendPacket(pAdapter, skb);
            if (Status == NDIS_STATUS_SUCCESS)
            {
                // Dequeue one frame from SendTxWait queue and process it
                // There are three place calling dequeue for TX ring.
                // 1. Here, right after queueing the frame.
                // 2. At the end of TxRingTxDone service routine.
                // 3. Upon NDIS call RTMPSendPackets
                if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
                {
                    RTMPDeQueuePacket(pAdapter, &pAdapter->SendTxWaitQueue);
                }   
            }
            else // free this packet space
            {
                RTMPFreeSkbBuffer(skb);
            }
        }
    }

    return result;
}

