/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

	Module Name:
	header.h

	Abstract:
	Miniport generic portion header file

	Revision History:
	Who         When          What
	--------    ----------    ----------------------------------------------
	Paul Lin    08-01-2002    created

*/
#include	"rt_config.h"

//
// BBP register initialization set
//
ULONG	BBPRegTable[] = {
	0x00018100, 		// R1 
	0x00018327,			// R3 
	0x00018408,         // R4 
	0x00018a0f,			// R10
	0x00018d08,			// R13
	0x00018f72,			// R15
	0x00019074,			// R16
	0x00019120,			// R17
	0x00019272,			// R18
	0x0001930b,			// R19
	0x00019400,			// R20
	0x00019c11,			// R28
	0x00019d04,         // R29
	0x00019e00,			// R30
	0x00019f00,			// R31
};
//
// ASIC register initialization sets
//
RTMP_REG_PAIR	MACRegTable[] =	{
	{PSCSR0,    0x00020002},			// 0xc8
	{PSCSR1,    0x00000002},			// 0xcc
	{PSCSR2,    0x00023f20},			// 0xd0
	{PSCSR3,    0x00000002},			// 0xd4
	{TIMECSR,   0x00003f20},     // 1/8 BEACON interval
	{CSR9,      0x00000800},			// 0x24
	{CSR18,     0x001e000a},			// kevin change PIFS,SIFS
	{CSR11,     0x070414a5},			// 0x2C, retry count	
	{CNT3,      0x3f080000},		    // Backoff_CCA_Th, RX_&_TX_CCA_Th

	{TXCSR1,    0x01614562},			// 0x64, ACK as 1Mb time
	{ARCSR0,    0x87888685},			// 0x98, ACK/CTS BBP register ID
	{ARCSR1,    0x0000000f},			// 0x9c
    {ARCSR2,    0x00700400},            // 0x13c, ACK/CTS PLCP at 1 Mbps
    {ARCSR3,    0x00380401},            // 0x140, ACK/CTS PLCP at 2 Mbps
    {ARCSR4,    0x00150402},            // 0x144, ACK/CTS PLCP at 5.5 Mbps
    {ARCSR5,    0x000b8403},            // 0x148, ACK/CTS PLCP at 11 Mbps
	{RXCSR0,    0xffffffff},			// 0x80	, accept ToDs=1 frame
	{RXCSR3,    0x0000a4a0},			// RT2430 BBP R36:RSSI, R32:SIGNAL
	{PWRCSR0,   0x3f3b3100},			// 0xC4
	{PSCSR2,    0x00023f20},			// 0xd4
	{PSCSR3,    0x00000002},			// 0xd4
	{PWRCSR1,   0x000001ff},			// 0xd8		
	{MACCSR0,   0x00217223},			// 0xE0, Enable Tx dribble mode			
	{MACCSR1,   0x00235518},			// 0xE4, Disable Rx Reset, tx dribble count, 2x30x16 = 960n,
	{MACCSR2, 	0x00000040},			// 0x0134, 64*33ns = 2us
 	{RALINKCSR, 0x9a009a11},			// 0xE8	
	{CSR7,      0xffffffff},				// 0x1C, Clear all pending interrupt source
    {LEDCSR,    0x00001E46},             // default both LEDs off
};
		
#define	NUM_BBP_REG_PARMS	(sizeof(BBPRegTable) / sizeof(ULONG))
#define	NUM_MAC_REG_PARMS	(sizeof(MACRegTable) / sizeof(RTMP_REG_PAIR))


/*
	========================================================================
	
	Routine Description:
		Allocate all DMA releated resources

	Arguments:
		Adapter			Pointer to our adapter

	Return Value:
		None

	Note:

	========================================================================
*/
NDIS_STATUS RTMPAllocDMAMemory(
	IN  PRTMP_ADAPTER   pAd)
{
	INT				index;
	VOID			*ring;			// VA of ring descriptor
	VOID			*ring_data;		// VA of DMA data buffer
	dma_addr_t		ring_dma;		// PA of ring descriptor
	dma_addr_t		ring_data_dma;	// PA of DMA data buffer
	PTXD_STRUC		pTxD;			// Tx type ring descriptor
	PRXD_STRUC		pRxD;			// Rx type ring descriptor

	DBGPRINT(RT_DEBUG_INFO, "--> RTMPAllocDMAMemory\n");
	
	// Allocate Ring descriptors DMA block
	ring = pci_alloc_consistent(pAd->pPci_Dev, DESCRIPTOR_REQUIRED * RING_DESCRIPTOR_SIZE, &ring_dma);
	if (!ring) {
		DBGPRINT(RT_DEBUG_ERROR, "Could not allocate DMA ring descriptor memory.\n");
		return -ENOMEM;
	}

	// Zero init ring descriptors
	memset(ring, 0, (DESCRIPTOR_REQUIRED * RING_DESCRIPTOR_SIZE));
	
	// Allocate Ring data DMA blocks
	ring_data = pci_alloc_consistent(pAd->pPci_Dev,
		(TX_RING_SIZE * TX_BUFFER_SIZE) + (PRIO_RING_SIZE * PRIO_BUFFER_SIZE) +
		(ATIM_RING_SIZE * ATIM_BUFFER_SIZE) + (RX_RING_SIZE * RX_BUFFER_SIZE) + 
		(BEACON_RING_SIZE * BEACON_BUFFER_SIZE), &ring_data_dma);
	
	// If failed, release ring descriptors DMA block & exit
	if (!ring_data) {
		pci_free_consistent(pAd->pPci_Dev, DESCRIPTOR_REQUIRED * RING_DESCRIPTOR_SIZE, ring, ring_dma);
		DBGPRINT(RT_DEBUG_ERROR, "Could not allocate DMA ring buffer memory.\n");		
		return -ENOMEM;
	}
	
	// All DMA blocks acquired, begin assign to correct place.

	// Start with Tx ring & DMA buffer
	for (index = 0; index < TX_RING_SIZE; index++)
	{
		// Init Tx Ring Size, Va, Pa variables
		pAd->TxRing[index].size = RING_DESCRIPTOR_SIZE;
		pAd->TxRing[index].va_addr = ring;
		pAd->TxRing[index].pa_addr = ring_dma;
		ring     += RING_DESCRIPTOR_SIZE;
		ring_dma += RING_DESCRIPTOR_SIZE;
		
		// Init Tx DMA buffer
		pAd->TxRing[index].data_size = TX_BUFFER_SIZE;
		pAd->TxRing[index].va_data_addr = ring_data;
		pAd->TxRing[index].pa_data_addr = ring_data_dma;
		ring_data     += TX_BUFFER_SIZE;
		ring_data_dma += TX_BUFFER_SIZE;
		
		// Write TxD buffer address & allocated buffer length
		pTxD = (PTXD_STRUC) pAd->TxRing[index].va_addr;
		pTxD->BufferAddressPa = pAd->TxRing[index].pa_data_addr;
		pTxD->BufferLen       = pAd->TxRing[index].data_size;

		DBGPRINT(RT_DEBUG_INFO, "TxRing[%d] va = 0x%x, pa = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->TxRing[index].va_addr, (UINT)pAd->TxRing[index].pa_addr, pAd->TxRing[index].size);
		DBGPRINT(RT_DEBUG_INFO, "TxRing[%d] va_data = 0x%x, pa_data = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->TxRing[index].va_data_addr, (UINT)pAd->TxRing[index].pa_data_addr, pAd->TxRing[index].data_size);
	}

	// Second with Prio ring & DMA buffer
	for (index = 0; index < PRIO_RING_SIZE; index++)
	{
		// Init Prio Ring Size, Va, Pa variables
		pAd->PrioRing[index].size = RING_DESCRIPTOR_SIZE;
		pAd->PrioRing[index].va_addr = ring;
		pAd->PrioRing[index].pa_addr = ring_dma;
		ring     += RING_DESCRIPTOR_SIZE;
		ring_dma += RING_DESCRIPTOR_SIZE;
		
		// Init Prio DMA buffer
		pAd->PrioRing[index].data_size = PRIO_BUFFER_SIZE;
		pAd->PrioRing[index].va_data_addr = ring_data;
		pAd->PrioRing[index].pa_data_addr = ring_data_dma;
		ring_data     += PRIO_BUFFER_SIZE;
		ring_data_dma += PRIO_BUFFER_SIZE;
		
		// Write TxD buffer address & allocated buffer length for priority ring
		pTxD = (PTXD_STRUC) pAd->PrioRing[index].va_addr;
		pTxD->BufferAddressPa = pAd->PrioRing[index].pa_data_addr;
		pTxD->BufferLen       = pAd->PrioRing[index].data_size;
		
		DBGPRINT(RT_DEBUG_INFO, "PrioRing[%d] va = 0x%x, pa = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->PrioRing[index].va_addr, (UINT)pAd->PrioRing[index].pa_addr, pAd->PrioRing[index].size);
		DBGPRINT(RT_DEBUG_INFO, "PrioRing[%d] va_data = 0x%x, pa_data = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->PrioRing[index].va_data_addr, (UINT)pAd->PrioRing[index].pa_data_addr, pAd->PrioRing[index].data_size);
	}
	
	// Atim ring & DMA buffer
	for (index = 0; index < ATIM_RING_SIZE; index++)
	{
		// Init Atim Ring Size, Va, Pa variables
		pAd->AtimRing[index].size = RING_DESCRIPTOR_SIZE;
		pAd->AtimRing[index].va_addr = ring;
		pAd->AtimRing[index].pa_addr = ring_dma;
		ring     += RING_DESCRIPTOR_SIZE;
		ring_dma += RING_DESCRIPTOR_SIZE;
		
		// Init Atim DMA buffer
		pAd->AtimRing[index].data_size = ATIM_BUFFER_SIZE;
		pAd->AtimRing[index].va_data_addr = ring_data;
		pAd->AtimRing[index].pa_data_addr = ring_data_dma;
		ring_data     += ATIM_BUFFER_SIZE;
		ring_data_dma += ATIM_BUFFER_SIZE;
		
		// Write TxD buffer address & allocated buffer length
		pTxD = (PTXD_STRUC) pAd->AtimRing[index].va_addr;
		pTxD->BufferAddressPa = pAd->AtimRing[index].pa_data_addr;
		pTxD->BufferLen       = pAd->AtimRing[index].data_size;

		DBGPRINT(RT_DEBUG_INFO, "AtimRing[%d] va = 0x%x, pa = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->AtimRing[index].va_addr, (UINT)pAd->AtimRing[index].pa_addr, pAd->AtimRing[index].size);
		DBGPRINT(RT_DEBUG_INFO, "AtimRing[%d] va_data = 0x%x, pa_data = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->AtimRing[index].va_data_addr, (UINT)pAd->AtimRing[index].pa_data_addr, pAd->AtimRing[index].data_size);
	}
	
	// Rx ring & DMA buffer
	for (index = 0; index < RX_RING_SIZE; index++)
	{
		// Init Rx Ring Size, Va, Pa variables
		pAd->RxRing[index].size = RING_DESCRIPTOR_SIZE;
		pAd->RxRing[index].va_addr = ring;
		pAd->RxRing[index].pa_addr = ring_dma;
		ring     += RING_DESCRIPTOR_SIZE;
		ring_dma += RING_DESCRIPTOR_SIZE;
		
		// Init Rx DMA buffer
		pAd->RxRing[index].data_size = RX_BUFFER_SIZE;
		pAd->RxRing[index].va_data_addr = ring_data;
		pAd->RxRing[index].pa_data_addr = ring_data_dma;
		ring_data     += RX_BUFFER_SIZE;
		ring_data_dma += RX_BUFFER_SIZE;
		
		// Write RxD buffer address & allocated buffer length
		pRxD = (PRXD_STRUC) pAd->RxRing[index].va_addr;
		pRxD->BufferAddressPa = pAd->RxRing[index].pa_data_addr;
		pRxD->BufferLen       = pAd->RxRing[index].data_size;
		// Rx owner bit assign to NIC immediately
		pRxD->Owner = DESC_OWN_NIC;

		DBGPRINT(RT_DEBUG_INFO, "RxRing[%d] va = 0x%x, pa = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->RxRing[index].va_addr, (UINT)pAd->RxRing[index].pa_addr, pAd->RxRing[index].size);
		DBGPRINT(RT_DEBUG_INFO, "RxRing[%d] va_data = 0x%x, pa_data = 0x%x, size = 0x%x\n",
			index, (UINT)pAd->RxRing[index].va_data_addr, (UINT)pAd->RxRing[index].pa_data_addr, pAd->RxRing[index].data_size);
	}

	// Now for beacon ring
	// Init Rx Ring Size, Va, Pa variables
	pAd->BeaconRing.size = RING_DESCRIPTOR_SIZE;
	pAd->BeaconRing.va_addr = ring;
	pAd->BeaconRing.pa_addr = ring_dma;
		
	// Init Rx DMA buffer
	pAd->BeaconRing.data_size = BEACON_BUFFER_SIZE;
	pAd->BeaconRing.va_data_addr = ring_data;
	pAd->BeaconRing.pa_data_addr = ring_data_dma;
		
	// Write RxD buffer address & allocated buffer length
	pTxD = (PTXD_STRUC) pAd->BeaconRing.va_addr;
	pTxD->BufferAddressPa = pAd->BeaconRing.pa_data_addr;
	pTxD->BufferLen       = pAd->BeaconRing.data_size;

	DBGPRINT(RT_DEBUG_INFO, "<-- RTMPAllocDMAMemory\n");
	return 0;
}

/*
	========================================================================
	
	Routine Description:
		Free all DMA memory.

	Arguments:
		Adapter			Pointer to our adapter

	Return Value:
		None

	Note:

	========================================================================
*/
VOID	RTMPFreeDMAMemory(
	IN  PRTMP_ADAPTER   pAd)
{
	DBGPRINT(RT_DEBUG_INFO, "--> RTMPFreeDMAMemory\n");
	
	// Free data DMA blocks first
	// the start address is the same as TxRing first DMA data block
	pci_free_consistent(pAd->pPci_Dev, (TX_RING_SIZE * TX_BUFFER_SIZE) + 
		(PRIO_RING_SIZE * PRIO_BUFFER_SIZE) + (ATIM_RING_SIZE * ATIM_BUFFER_SIZE) + 
		(RX_RING_SIZE * RX_BUFFER_SIZE) + (BEACON_RING_SIZE * BEACON_BUFFER_SIZE), 
		pAd->TxRing[0].va_data_addr, pAd->TxRing[0].pa_data_addr);

	// Free ring descriptor second
	// the start address is the same as TxRing first elment
	pci_free_consistent(pAd->pPci_Dev, DESCRIPTOR_REQUIRED * RING_DESCRIPTOR_SIZE,
		pAd->TxRing[0].va_addr, pAd->TxRing[0].pa_addr);
}

/*
	========================================================================
	
	Routine Description:
		Initialize transmit data structures

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		Initialize all transmit releated private buffer, include those define
		in RTMP_ADAPTER structure and all private data structures.
		
	========================================================================
*/
VOID	NICInitTransmit(
	IN	PRTMP_ADAPTER	pAd)
{
	DBGPRINT(RT_DEBUG_TRACE, "--> NICInitTransmit\n");

	// Initialize all Transmit releated queues
	InitializeQueueHeader(&pAd->SendTxWaitQueue);
	
	// Init Ring index pointer
	pAd->CurRxIndex          = 0;
	pAd->CurTxIndex          = 0;
	pAd->CurAtimIndex        = 0;
	pAd->CurPrioIndex        = 0;
	pAd->NextTxDoneIndex     = 0;
	pAd->NextAtimDoneIndex   = 0;
	pAd->NextPrioDoneIndex   = 0;
	pAd->PushMgmtIndex       = 0;
	pAd->PopMgmtIndex        = 0;
	pAd->MgmtQueueSize       = 0;
	
	pAd->PrivateInfo.TxRingFullCnt       = 0;
	
	DBGPRINT(RT_DEBUG_TRACE, "<-- NICInitTransmit\n");
}

inline	VOID    NICDisableInterrupt(
    IN  PRTMP_ADAPTER   pAd)
{
    RTMP_IO_WRITE32(pAd, CSR8, 0xFFFF);
    RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_ACTIVE);
}

inline	VOID    NICEnableInterrupt(
    IN  PRTMP_ADAPTER   pAd)
{
	// 0xFF37 : Txdone & Rxdone, 0xFF07: Txdonw, Rxdone, PrioDone, AtimDone,
    RTMP_IO_WRITE32(pAd, CSR8, 0xFF04);		
    RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_ACTIVE);
}

/*
	========================================================================
	
	Routine Description:
		Read additional information from EEPROM, such as MAC address

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		NDIS_STATUS_SUCCESS
		NDIS_STATUS_FAILURE

	Note:
	
	========================================================================
*/
NDIS_STATUS	NICReadAdapterInfo(
	IN	PRTMP_ADAPTER		pAd)
{
	CSR3_STRUC		StaMacReg0;
	CSR4_STRUC		StaMacReg1;
	NDIS_STATUS		Status = NDIS_STATUS_SUCCESS;

	// 
	// Read MAC address from CSR3 & CSR4, these CSRs reflects real value
	// stored with EEPROM.
	//
	RTMP_IO_READ32(pAd, CSR3, &StaMacReg0.word);
	RTMP_IO_READ32(pAd, CSR4, &StaMacReg1.word);
	pAd->PermanentAddress[0] = StaMacReg0.field.Byte0;
	pAd->PermanentAddress[1] = StaMacReg0.field.Byte1;
	pAd->PermanentAddress[2] = StaMacReg0.field.Byte2;
	pAd->PermanentAddress[3] = StaMacReg0.field.Byte3;
	pAd->PermanentAddress[4] = StaMacReg1.field.Byte4;
	pAd->PermanentAddress[5] = StaMacReg1.field.Byte5;
	// Set Current address to permanet address.
	// TODO: put current address check within registery read out
	pAd->CurrentAddress[0] = pAd->PermanentAddress[0];
	pAd->CurrentAddress[1] = pAd->PermanentAddress[1];
	pAd->CurrentAddress[2] = pAd->PermanentAddress[2];
	pAd->CurrentAddress[3] = pAd->PermanentAddress[3];
	pAd->CurrentAddress[4] = pAd->PermanentAddress[4];
	pAd->CurrentAddress[5] = pAd->PermanentAddress[5];

	return Status;
}

/*
	========================================================================
	
	Routine Description:
		Read initial parameters from EEPROM
		
	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		
	========================================================================
*/
VOID	NICReadEEPROMParameters(
	IN	PRTMP_ADAPTER	pAd)
{
	ULONG			data;
	USHORT			i, value;
	EEPROM_TX_PWR_STRUC	Power;

	DBGPRINT(RT_DEBUG_TRACE, "--> NICReadEEPROMParameters\n");

	// Init EEPROM Address Number, before access EEPROM; if 93c46, EEPROMAddressNum=6, else if 93c66, EEPROMAddressNum=8
	RTMP_IO_READ32(pAd, CSR21, &data);

	if(data & 0x20)
		pAd->EEPROMAddressNum = 6;		
	else
		pAd->EEPROMAddressNum = 8;

	// Read BBP default value from EEPROM and store to array(EEPROMDefaultValue) in pAdapter
	for(i = 0; i < NUM_EEPROM_BBP_PARMS; i++)
	{
		value = RTMP_EEPROM_READ16(pAd, EEPROM_BBP_BASE_OFFSET + i*2);
		
		pAd->EEPROMDefaultValue[i] = value;
	}

	// Read Tx power value for all 14 channels
	// Value from 1 - 0x7f. Default value is 0x27.
	for (i = 0; i < NUM_EEPROM_TX_PARMS; i++)
	{
		Power.word = RTMP_EEPROM_READ16(pAd, EEPROM_TX_PWR_OFFSET + i*2);
		pAd->PortCfg.ChannelTxPower[i * 2]     = ((Power.field.Byte0 > 0x7f) ? 0x27 : Power.field.Byte0);
		pAd->PortCfg.ChannelTxPower[i * 2 + 1] = ((Power.field.Byte1 > 0x7f) ? 0x27 : Power.field.Byte1);
	}

	//CountryRegion byte offset = 0x35
	value = RTMP_EEPROM_READ16(pAd, 0x34) >> 8; 
	if (value <= 7)
	{
	     pAd->PortCfg.CountryRegion = (UCHAR) value;
	}
	DBGPRINT(RT_DEBUG_TRACE, "<-- NICReadEEPROMParameters\n");
}

/*
	========================================================================
	
	Routine Description:
		Set default value from EEPROM
		
	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		
	========================================================================
*/
VOID	NICInitAsicFromEEPROM(
	IN	PRTMP_ADAPTER	pAd)
{
	ULONG					data;
	USHORT					i, value;
	UCHAR					TxValue,RxValue;
	EEPROM_ANTENNA_STRUC	Antenna;

	DBGPRINT(RT_DEBUG_TRACE, "--> NICInitAsicFromEEPROM\n");
	
	for(i = 1; i < NUM_EEPROM_BBP_PARMS; i++)
	{
		value = pAd->EEPROMDefaultValue[i];
		
		if((value != 0xFFFF) && (value != 0))
		{
			data = value | 0x18000;
			RTMP_BBP_IO_WRITE32(pAd, data);
		}
	}

	Antenna.word = pAd->EEPROMDefaultValue[0];

	if((Antenna.word != 0xFFFF) && ((Antenna.field.NumOfAntenna == 1) || (Antenna.field.NumOfAntenna == 2)) 
		&& (Antenna.field.TxDefaultAntenna <= 2) && (Antenna.field.RxDefaultAntenna <= 2))
	{
		pAd->PortCfg.NumberOfAntenna = (UCHAR)Antenna.field.NumOfAntenna;
		pAd->PortCfg.CurrentTxAntenna = (UCHAR)Antenna.field.TxDefaultAntenna;
		pAd->PortCfg.CurrentRxAntenna = (UCHAR)Antenna.field.RxDefaultAntenna;

		RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Tx_Configure, &TxValue);
		RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Rx_Configure, &RxValue);

		// Tx antenna select
		if(Antenna.field.TxDefaultAntenna == 0)
		{// Antenna Diversity
			TxValue = (TxValue & 0xFC) | 0x01;
		}
		else if(Antenna.field.TxDefaultAntenna == 1)
		{// Antenna A
			TxValue = (TxValue & 0xFC) | 0x00;
		}
		else if(Antenna.field.TxDefaultAntenna == 2)
		{// Antenna B
			TxValue = (TxValue & 0xFC) | 0x02;
		}

		// Rx antenna select
		if(Antenna.field.RxDefaultAntenna == 0)
		{// Antenna Diversity
			RxValue = (RxValue & 0xF9) | 0x02;
		}
		else if(Antenna.field.RxDefaultAntenna == 1)
		{// Antenna A
			RxValue = (RxValue & 0xF9) | 0x00;
		}
		else if(Antenna.field.RxDefaultAntenna == 2)
		{// Antenna B
			RxValue = (RxValue & 0xF9) | 0x04;
		}
		// Change to match microsoft definition, 0xff: diversity, 0: A, 1: B
		pAd->PortCfg.CurrentTxAntenna--;
		pAd->PortCfg.CurrentRxAntenna--;

		RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Tx_Configure, TxValue);
		RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Rx_Configure, RxValue);

        if (Antenna.field.RfType == RFIC_2420)
			pAd->PortCfg.RfType = RFIC_2420;
        else
			pAd->PortCfg.RfType = RFIC_2421;

        if (Antenna.field.LedMode == LED_MODE_TXRX_ACTIVITY)
                pAd->PortCfg.LedMode = LED_MODE_TXRX_ACTIVITY;
        else if (Antenna.field.LedMode == LED_MODE_SINGLE)
		{
			pAd->PortCfg.LedMode = LED_MODE_SINGLE;
			ASIC_LED_ACT_ON(pAd);
		}
        else			
			pAd->PortCfg.LedMode = LED_MODE_DEFAULT;

	        pAd->PortCfg.EnableRxAgcVgcTuning = (BOOLEAN)Antenna.field.EnableRxAgcVgcTuning;
	}

	DBGPRINT(RT_DEBUG_TRACE, "RF IC=%d, LED mode=%d, R13 tuning=%d\n", 
	    pAd->PortCfg.RfType, pAd->PortCfg.LedMode, pAd->PortCfg.EnableRxAgcVgcTuning);
	DBGPRINT(RT_DEBUG_TRACE, "<-- NICInitAsicFromEEPROM\n");
}

void NICInitializeAdapter(IN	PRTMP_ADAPTER	pAd)
{
	TXCSR2_STRUC	TxCSR2;
	RXCSR1_STRUC	RxCSR1;
	ULONG			Value;

	DBGPRINT(RT_DEBUG_TRACE, "--> NICInitializeAdapter\n");

	// Init spin locks
	spin_lock_init(&pAd->TxRingLock);
	spin_lock_init(&pAd->PrioRingLock);
	spin_lock_init(&pAd->AtimRingLock);
	spin_lock_init(&pAd->RxRingLock);
	spin_lock_init(&pAd->SendTxWaitQueueLock);
		
	// Write TXCSR2 register
	TxCSR2.field.TxDSize = RING_DESCRIPTOR_SIZE;
	TxCSR2.field.NumTxD  = TX_RING_SIZE;
	TxCSR2.field.NumAtimD  = ATIM_RING_SIZE;
	TxCSR2.field.NumPrioD  = PRIO_RING_SIZE;	
	RTMP_IO_WRITE32(pAd, TXCSR2, TxCSR2.word);

	// Write TXCSR3 register
	Value = pAd->TxRing[0].pa_addr;
	RTMP_IO_WRITE32(pAd, TX_RING_BASE_REG, Value);
	
	// Write TXCSR4 register
	Value = pAd->PrioRing[0].pa_addr;
	RTMP_IO_WRITE32(pAd, PRIO_RING_BASE_REG, Value);
	
	// Write TXCSR5 register
	Value = pAd->AtimRing[0].pa_addr;
	RTMP_IO_WRITE32(pAd, ATIM_RING_BASE_REG, Value);

	// Write TXCSR6 register
	Value = pAd->BeaconRing.pa_addr;
	RTMP_IO_WRITE32(pAd, BEACON_BASE_REG, Value);

	// Write RXCSR1 register
	RxCSR1.field.RxDSize = RING_DESCRIPTOR_SIZE;
	RxCSR1.field.NumRxD  = RX_RING_SIZE;
	RTMP_IO_WRITE32(pAd, RXCSR1, RxCSR1.word);
	
	// Write RXCSR2 register
	Value = pAd->RxRing[0].pa_addr;
	RTMP_IO_WRITE32(pAd, RX_RING_BASE_REG, Value);

	// Write CSR1 for host ready
	// Move Host reay to end of ASIC initialization 
	// to ensure no Rx will perform before ASIC init
	// RTMP_IO_WRITE32(pAd, CSR1, 0x4);

	// Initialze ASIC for TX & Rx operation
	NICInitializeAsic(pAd);
	
	DBGPRINT(RT_DEBUG_TRACE, "<-- NICInitializeAdapter\n");
}	

void	NICInitializeAsic(IN	PRTMP_ADAPTER	pAd)
{
	ULONG			index;

	DBGPRINT(RT_DEBUG_TRACE, "--> NICInitializeAsic\n");

	// Initialize MAC register to default value
	for (index = 0; index < NUM_MAC_REG_PARMS; index++)
	{
		RTMP_IO_WRITE32(pAd, MACRegTable[index].Register, MACRegTable[index].Value);
	}
	
	// Set Host ready before kicking Rx
	RTMP_IO_WRITE32(pAd, CSR1, 0x1); // reset MAC state machine, requested by Kevin 2003-2-11
	RTMP_IO_WRITE32(pAd, CSR1, 0x4);

	// Initialize BBP register to default value
	for (index = 0; index < NUM_BBP_REG_PARMS; index++)
	{
		RTMP_BBP_IO_WRITE32(pAd, BBPRegTable[index]);
	}

	// Initialize RF register to default value
    AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
	AsicLockChannel(pAd, pAd->PortCfg.Channel);

	// Kick Rx
	RTMP_IO_WRITE32(pAd, RXCSR0, 0x56);  // allow ToDs=1 frames, allow PSPOLL(CNTL) frame
	
	// Clear old FCS jitter before init ASIC
	RTMP_IO_READ32(pAd, CNT0, &index);
	
	// Clear old Rx FIFO error jitter before init ASIC
	RTMP_IO_READ32(pAd, CNT4, &index);
	
	DBGPRINT(RT_DEBUG_TRACE, "<-- NICInitializeAsic\n");
}

/*
	========================================================================
	
	Routine Description:
		Reset NIC Asics

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		Reset NIC to initial state AS IS system boot up time.
		
	========================================================================
*/
VOID	NICIssueReset(
	IN	PRTMP_ADAPTER	pAd)
{
	CSR3_STRUC		StaMacReg0;
	CSR4_STRUC		StaMacReg1;

	DBGPRINT(RT_DEBUG_TRACE, "--> NICIssueReset\n");

	// Abort Tx, prevent ASIC from writing to Host memory
	RTMP_IO_WRITE32(pAd, TXCSR0, 0x08);
	
	// Disable Rx, register value supposed will remain after reset
	RTMP_IO_WRITE32(pAd, RXCSR0, 0x1);

	// Issue reset and clear from reset state
	RTMP_IO_WRITE32(pAd, CSR1, 0x01);
	RTMP_IO_WRITE32(pAd, CSR1, 0x00);

	DBGPRINT(RT_DEBUG_TRACE, "<-- NICIssueReset\n");
}

/*
	========================================================================
	
	Routine Description:
		Check ASIC registers and find any reason the system might hang

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		
		
	========================================================================
*/
BOOLEAN	NICCheckForHang(
	IN	PRTMP_ADAPTER	pAd)
{
	ULONG	Value, OldValue;

	// Update FCS counters
 	RTMP_IO_READ32(pAd, CNT0, &Value);
	pAd->WlanCounters.FCSErrorCount += Value;
	
	// Update RxNoBuffer counters
 	RTMP_IO_READ32(pAd, CNT4, &Value);
	pAd->Counters.RxNoBuffer += Value;
	
	// Read CCA error count
	RTMP_IO_READ32(pAd, CNT3, &Value);
	pAd->PrivateInfo.CCAErrCnt = (Value & 0x0000ffff);
	DBGPRINT(RT_DEBUG_INFO, "CCA error count -> 0x%x\n", pAd->PrivateInfo.CCAErrCnt);

	// Read PLCP error count
	RTMP_IO_READ32(pAd, CNT1, &Value);
	pAd->PrivateInfo.PhyRxErrCnt = (Value & 0x0000ffff);
	DBGPRINT(RT_DEBUG_INFO, "PHY Rx error count -> 0x%x\n", pAd->PrivateInfo.PhyRxErrCnt);
	
	pAd->PrivateInfo.PhyTxErrCnt = (Value & 0xffff0000) >> 16;
	DBGPRINT(RT_DEBUG_INFO, "PHY Tx error count -> 0x%x\n", pAd->PrivateInfo.PhyTxErrCnt);

	return (FALSE);
}

/*
	========================================================================
	
	Routine Description:
		Reset NIC from error

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		Reset NIC from error state
		
	========================================================================
*/
VOID	NICResetFromError(
	IN	PRTMP_ADAPTER	pAdapter)
{
	UCHAR	Index;
	
	// Reset BBP (according to alex, reset ASIC will force reset BBP
	// Therefore, skip the reset BBP
	// RTMP_IO_WRITE32(pAdapter, CSR1, 0x2);
	// Release BBP reset
	// RTMP_IO_WRITE32(pAdapter, CSR1, 0x0);

	RTMP_IO_WRITE32(pAdapter, CSR1, 0x1);
	// Remove ASIC from reset state
	RTMP_IO_WRITE32(pAdapter, CSR1, 0x0);

	// Init send data structures and related parameters
	NICInitTransmit(pAdapter);

	NICInitializeAdapter(pAdapter);	
	NICInitAsicFromEEPROM(pAdapter);

	// Switch to current channel, since during reset process, the connection should remains on.	
	AsicSwitchChannel(pAdapter, pAdapter->PortCfg.Channel);
	AsicLockChannel(pAdapter, pAdapter->PortCfg.Channel);
	
}

/*
	========================================================================
	
	Routine Description:
		Verify section is valid for Get key parameter.

	Arguments:
		buffer						Pointer to the buffer to start find the key section
		ptr						    pointer to section

	Return Value:
		TRUE						Success
		FALSE						Fail
	========================================================================
*/
INT	RTMPIsFindSection(
	IN	PUCHAR	ptr,
	IN	PUCHAR	buffer)
{
	if(ptr == buffer)
		return TRUE;
	else if (ptr > buffer) 
	{
		while (ptr > buffer)
		{
			ptr--;
			if( *ptr == 0x0a)
				return TRUE;
			else if( (*ptr == ' ') || (*ptr == '\t'))
				continue;
			else
				return FALSE;
		}
		return TRUE;
	}
		
	return FALSE;
}
/*
	========================================================================
	
	Routine Description:
		Find key section for Get key parameter.

	Arguments:
		buffer						Pointer to the buffer to start find the key section
		section						the key of the secion to be find

	Return Value:
		NULL						Fail
		Others						Success
	========================================================================
*/
PUCHAR	RTMPFindSection(
	IN	PCHAR	buffer,
	IN	PCHAR	section)
{
	CHAR temp_buf[255];
	PUCHAR	ptr;

	strcpy(temp_buf, "[");                  /*  and the opening bracket [  */
	strcat(temp_buf, section);
	strcat(temp_buf, "]");

	if((ptr = strstr(buffer, temp_buf)) != NULL)
	{
		if(RTMPIsFindSection(ptr, buffer))
			return (ptr+strlen("\n"));
		else
			return NULL;
	}
	else
		return NULL;
}	
/*
	========================================================================
	
	Routine Description:
		Get key parameter.

	Arguments:
		section						the key of the secion
		key							Pointer to key string
		dest						Pointer to destination		
		destsize					The datasize of the destination
		buffer						Pointer to the buffer to start find the key

	Return Value:
		TRUE						Success
		FALSE						Fail

	Note:
		This routine get the value with the matched key (case case-sensitive)
	========================================================================
*/
INT RTMPGetKeyParameter(
	IN	PUCHAR	section,
	IN	PCHAR	key,
	OUT	PCHAR	dest,	
	IN	INT		destsize,
	IN	PCHAR	buffer)
{
	char temp_buf1[600];
	char temp_buf2[600];
	char *start_ptr;
	char *end_ptr;
	char *ptr;
	char *too_far_ptr;
	char *offset;
	int  len;

	//find section
	if((offset = RTMPFindSection(buffer, section)) == NULL)
		return (FALSE);

	strcpy(temp_buf1, "\n");
	strcat(temp_buf1, key);
	strcat(temp_buf1, "=");

	//search key
	if((start_ptr=strstr(offset, temp_buf1))==NULL)
		return FALSE;

	start_ptr+=strlen("\n");
	if((too_far_ptr=strstr(offset+1, "["))==NULL)
	   too_far_ptr=offset+strlen(offset);

	if((end_ptr=strstr(start_ptr, "\n"))==NULL)
	   end_ptr=start_ptr+strlen(start_ptr);

	if (too_far_ptr<start_ptr)
		return FALSE;

	memcpy(temp_buf2, start_ptr, end_ptr-start_ptr);
	temp_buf2[end_ptr-start_ptr]='\0';
	len = strlen(temp_buf2);
	strcpy(temp_buf1, temp_buf2);
	if((start_ptr=strstr(temp_buf1, "=")) == NULL)
		return FALSE;

	strcpy(temp_buf2, start_ptr+1);
	ptr = temp_buf2;
	//trim space or tab
	while(*ptr != 0x00)
	{
		if( (*ptr == ' ') || (*ptr == '\t') )
			ptr++;
	  	else
	  	   break;
	}

	len = strlen(ptr);	  
	memset(dest, 0x00, destsize);
	strncpy(dest, ptr, len >= destsize ?  destsize: len);

	return TRUE;
}

/*
	========================================================================
	
	Routine Description:
		In kernel mode read parameters from file

	Arguments:
		src						the location of the file.
		dest						put the parameters to the destination.
		Length					size to read.

	Return Value:
		None

	Note:
		
	========================================================================
*/

VOID RTMPReadParametersFromFile(
	IN	PRTMP_ADAPTER pAd)
{
	CHAR					buffer[MAX_INI_BUFFER_SIZE];
	CHAR					tmpbuf[600];
	UCHAR					keyMaterial[40];
	ULONG					BbpCsr1;
	UCHAR					TxValue;
	UCHAR					RxValue;
	ULONG					RtsThresh;
	ULONG					FragThresh;
	NDIS_802_11_WEP_STATUS	WepStatus;
	ULONG					KeyIdx;
	ULONG					KeyType;
	ULONG					KeyLen;
	MACADDR					macAddress;
	PUCHAR					macptr;						
	INT						i=0,j=0;

	RTMP_PROFILE_SETTING	profile;
	PUCHAR					src;
	struct file				*srcf;
	INT 					retval, orgfsuid, orgfsgid;
	UINT					Length;
   	mm_segment_t			orgfs;
   	
   	src = PROFILE_PATH;
   	Length = sizeof(RTMP_PROFILE_SETTING);
	DBGPRINT(RT_DEBUG_TRACE, "--> RTMPReadParametersFromFile\n");
	// Save uid and gid used for filesystem access.
	// Set user and group to 0 (root)	
	orgfsuid = current->fsuid;
	orgfsgid = current->fsgid;
	current->fsuid=current->fsgid = 0;
    orgfs = get_fs();
    set_fs(KERNEL_DS);

	if (src && *src) 
	{
		srcf = filp_open(src, O_RDONLY, 0);
		if (IS_ERR(srcf)) 
		{
			DBGPRINT(RT_DEBUG_TRACE, "--> Error %ld opening %s\n", -PTR_ERR(srcf),src);    
		}
		else 
		{
			/* The object must have a read method */
			if (srcf->f_op && srcf->f_op->read) 
			{
				memset(buffer, 0x00, MAX_INI_BUFFER_SIZE);
				retval=srcf->f_op->read(srcf, buffer, MAX_INI_BUFFER_SIZE, &srcf->f_pos);
				//retval=srcf->f_op->read(srcf, (PUCHAR)&profile, Length, &srcf->f_pos);
				if (retval < 0)
				{
					DBGPRINT(RT_DEBUG_TRACE, "--> Read %s error %d\n", src, -retval);
				}
				else
				{
					//SSID
					if(RTMPGetKeyParameter("Default", "SSID", pAd->PortCfg.Ssid, 32, buffer))
					{
						pAd->PortCfg.SsidLen = strlen(pAd->PortCfg.Ssid);
						DBGPRINT(RT_DEBUG_TRACE, "SSID=%s\n", pAd->PortCfg.Ssid);
					}
					//Channel
					if(RTMPGetKeyParameter("Default", "Channel", tmpbuf, 10, buffer))
					{
						pAd->PortCfg.Channel = (UCHAR) simple_strtol(tmpbuf, 0, 10);
						DBGPRINT(RT_DEBUG_TRACE, "Channel=%d\n", pAd->PortCfg.Channel);				
					}
					//BeaconPeriod
					if(RTMPGetKeyParameter("Default", "BeaconPeriod", tmpbuf, 10, buffer))
					{
						pAd->PortCfg.BeaconPeriod = (USHORT) simple_strtol(tmpbuf, 0, 10);
						DBGPRINT(RT_DEBUG_TRACE, "BeaconPeriod=%d\n", pAd->PortCfg.BeaconPeriod);
					}
					//RTSThreshold
					if(RTMPGetKeyParameter("Default", "RTSThreshold", tmpbuf, 10, buffer))
					{
						RtsThresh = simple_strtol(tmpbuf, 0, 10);
						if( (RtsThresh >= 1) && (RtsThresh <= MAX_RTS_THRESHOLD) )
							pAd->PortCfg.RtsThreshold  = (USHORT)RtsThresh;
						else
							pAd->PortCfg.RtsThreshold = MAX_RTS_THRESHOLD;
						
						DBGPRINT(RT_DEBUG_TRACE, "RTSThreshold=%d\n", pAd->PortCfg.RtsThreshold);
					}
					//FragThreshold
					if(RTMPGetKeyParameter("Default", "FragThreshold", tmpbuf, 10, buffer))
					{		
						FragThresh = simple_strtol(tmpbuf, 0, 10);
						pAd->PortCfg.bFragmentZeroDisable = FALSE;			

						if (FragThresh > MAX_FRAG_THRESHOLD || FragThresh < MIN_FRAG_THRESHOLD)
						{ //illegal FragThresh so we set it to default
							pAd->PortCfg.FragmentThreshold = MAX_FRAG_THRESHOLD;
							pAd->PortCfg.bFragmentZeroDisable = TRUE;
						}
						else
							pAd->PortCfg.FragmentThreshold = (USHORT)FragThresh;

						DBGPRINT(RT_DEBUG_TRACE, "FragThreshold=%d\n", pAd->PortCfg.FragmentThreshold);
					}
					//AuthMode
					if(RTMPGetKeyParameter("Default", "AuthMode", tmpbuf, 10, buffer))
					{
						if ((strcmp(tmpbuf, "OPEN") == 0) || (strcmp(tmpbuf, "open") == 0))
							pAd->PortCfg.AuthMode = Ndis802_11AuthModeOpen;
						else if ((strcmp(tmpbuf, "SHARED") == 0) || (strcmp(tmpbuf, "shared") == 0))
							pAd->PortCfg.AuthMode = Ndis802_11AuthModeShared;
						else if ((strcmp(tmpbuf, "WPAPSK") == 0) || (strcmp(tmpbuf, "wpapsk") == 0))
							pAd->PortCfg.AuthMode = Ndis802_11AuthModeWPAPSK;
//						else if ((strcmp(tmpbuf, "WPA") == 0) || (strcmp(tmpbuf, "wpa") == 0))
//							pAd->PortCfg.AuthMode = Ndis802_11AuthModeWPA;
						else //Default
							pAd->PortCfg.AuthMode = Ndis802_11AuthModeOpen;

                        RTMPMakeRSNIE(pAd, pAd->PortCfg.AuthMode, pAd->PortCfg.WepStatus);
						DBGPRINT(RT_DEBUG_TRACE, "AuthMode=%d\n", pAd->PortCfg.AuthMode);
					}
					//EncrypType
					if(RTMPGetKeyParameter("Default", "EncrypType", tmpbuf, 10, buffer))
					{
						if ((strcmp(tmpbuf, "NONE") == 0) || (strcmp(tmpbuf, "none") == 0))
							pAd->PortCfg.WepStatus = Ndis802_11WEPDisabled;
						else if ((strcmp(tmpbuf, "WEP") == 0) || (strcmp(tmpbuf, "wep") == 0))
							pAd->PortCfg.WepStatus = Ndis802_11WEPEnabled;
						else if ((strcmp(tmpbuf, "TKIP") == 0) || (strcmp(tmpbuf, "tkip") == 0))
                        {
							pAd->PortCfg.WepStatus = Ndis802_11Encryption2Enabled;
                        }
						else
							pAd->PortCfg.WepStatus = Ndis802_11WEPDisabled;

                        RTMPMakeRSNIE(pAd, pAd->PortCfg.AuthMode, pAd->PortCfg.WepStatus);
						DBGPRINT(RT_DEBUG_TRACE, "EncrypType=%d\n", pAd->PortCfg.WepStatus);
					}
                    //RekeyMethod
					if(RTMPGetKeyParameter("Default", "RekeyMethod", tmpbuf, 10, buffer))
					{
						if ((strcmp(tmpbuf, "TIME") == 0) || (strcmp(tmpbuf, "time") == 0))
							pAd->PortCfg.WPAREKEY.ReKeyMethod = TIME_REKEY;
						else if ((strcmp(tmpbuf, "PKT") == 0) || (strcmp(tmpbuf, "pkt") == 0))
							pAd->PortCfg.WPAREKEY.ReKeyMethod = PKT_REKEY;
						else if ((strcmp(tmpbuf, "DISABLE") == 0) || (strcmp(tmpbuf, "disable") == 0))
							pAd->PortCfg.WPAREKEY.ReKeyMethod = DISABLE_REKEY;
						else
							pAd->PortCfg.WPAREKEY.ReKeyMethod = DISABLE_REKEY;

						DBGPRINT(RT_DEBUG_TRACE, "ReKeyMethod=%d\n", pAd->PortCfg.WPAREKEY.ReKeyMethod);
					}
                    //RekeyInterval
					if(RTMPGetKeyParameter("Default", "RekeyInterval", tmpbuf, 10, buffer))
					{
						if((simple_strtol(tmpbuf, 0, 10) >= 0) && (simple_strtol(tmpbuf, 0, 10) < MAX_REKEY_INTER))
							pAd->PortCfg.WPAREKEY.ReKeyInterval = simple_strtol(tmpbuf, 0, 10);
						else //Default
							pAd->PortCfg.WPAREKEY.ReKeyInterval = 10;

						DBGPRINT(RT_DEBUG_TRACE, "ReKeyInterval=%d\n", pAd->PortCfg.WPAREKEY.ReKeyInterval);
					}
#ifdef	WDS
					//WdsList
					if(RTMPGetKeyParameter("Default", "WdsList", tmpbuf, 576, buffer))
					{
						RT_802_11_WDS	Wds;						
						
						memset(&Wds, 0x00, sizeof(RT_802_11_WDS));
						for (i=0, macptr = strtok(tmpbuf,";"); (macptr && i < 3); macptr = strtok(NULL,";"), i++) 
						{
							if(strlen(macptr) != 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
								continue; 

							for (j=0; j<ETH_LENGTH_OF_ADDRESS; j++)
							{
								AtoH(macptr, &macAddress.Octet[j], 2);
								macptr=macptr+3;
							}	
							memcpy(&Wds.Entry[i], &macAddress, ETH_LENGTH_OF_ADDRESS);
							Wds.Num ++;
						}

						pAd->WdsTab.Num = Wds.Num;						
						for(i = 0; i < pAd->WdsTab.Num; i++)
						{
							NdisMoveMemory(&pAd->WdsTab.WdsEntry[i].WdsAddr[0], &Wds.Entry[i], MAC_ADDR_LEN);
							DBGPRINT(RT_DEBUG_TRACE,"WDS-AP(%d)-%02x:%02x:%02x:%02x:%02x:%02x\n", i,
								pAd->WdsTab.WdsEntry[i].WdsAddr[0], pAd->WdsTab.WdsEntry[i].WdsAddr[1], pAd->WdsTab.WdsEntry[i].WdsAddr[2], 
								pAd->WdsTab.WdsEntry[i].WdsAddr[3], pAd->WdsTab.WdsEntry[i].WdsAddr[4], pAd->WdsTab.WdsEntry[i].WdsAddr[5]);
						}
					}
					//WdsKey					
					if(RTMPGetKeyParameter("Default", "WPAPSK", tmpbuf, 255, buffer))
					{
						if ((strlen(tmpbuf) >= 8) && (strlen(tmpbuf) <= 64))
						{					
							PasswordHash((char *)tmpbuf, RALINK, 6, keyMaterial);
						    //UCHAR   prefix[] = {'R', 'a', 'l', 'i', 'n', 'k'};
							//PasswordHash((char *)tmpbuf, prefix, 6, keyMaterial);
							NdisZeroMemory(&pAd->WdsTab.Wpa_key, sizeof(WPA_KEY));
							NdisMoveMemory(&pAd->WdsTab.Wpa_key.Key, keyMaterial, 16);
							pAd->WdsTab.Wpa_key.KeyLen = 16;
							NdisMoveMemory(&pAd->WdsTab.Wpa_key.RxMic, keyMaterial+16, 8);
							NdisMoveMemory(&pAd->WdsTab.Wpa_key.TxMic, keyMaterial+16, 8); 
                            // WPAPSK TKIP , require the same RXMIC and TXMIC . 
							DBGPRINT(RT_DEBUG_TRACE,"WDS-KEY (num=%d) %d\n", pAd->WdsTab.Num, sizeof(RALINK));
#ifdef RT2400_DBG
							printk("================WDS=======================\n");
							printk("WPAPSK=>\n");
							printk("WdsTab.Wpa_key=>\n");
							for (i=0; i<16; i++)
								printk("%02x:", pAd->WdsTab.Wpa_key.Key[i]);
							printk("\nRxMic Key=>\n");
							for (i=0; i<8; i++)
								printk("%02x:", pAd->WdsTab.Wpa_key.RxMic[i]);
							printk("\nTxMic Key=>\n");
							for (i=0; i<8; i++)
								printk("%02x:", pAd->WdsTab.Wpa_key.TxMic[i]);
							printk("\n");
#endif
						}
					}
					//WdsEnable
					if(RTMPGetKeyParameter("Default", "WdsEnable", tmpbuf, 255, buffer))
					{						
						if(simple_strtol(tmpbuf, 0, 10) == 0)
						{ //Disable
							pAd->WdsTab.Num = 0;
							pAd->WdsTab.Wpa_key.KeyLen = 0;							
							DBGPRINT(RT_DEBUG_TRACE,"WDS-Enable (num=%d, keylen=%d)\n", pAd->WdsTab.Num, pAd->WdsTab.Wpa_key.KeyLen);
						}
					}
#endif
                    //WPAPSK_KEY
					if(RTMPGetKeyParameter("Default", "WPAPSK", tmpbuf, 255, buffer))
					{
						DBGPRINT(RT_DEBUG_TRACE, "WPAPSK_KEY=%s\n", tmpbuf);
						if ((strlen(tmpbuf) >= 8) && (strlen(tmpbuf) <= 64))
						{
							PasswordHash((char *)tmpbuf, pAd->PortCfg.Ssid, pAd->PortCfg.SsidLen, keyMaterial);
							NdisMoveMemory(pAd->PortCfg.PMK, keyMaterial, 32);
#ifdef RT2400_DBG
							printk("%s() WPAPSK Key => \n", __FUNCTION__);
							for (i = 0; i < 32; i++)
							{
								printk("%02x:", pAd->PortCfg.PMK[i]);
								if ((i%16) == 15)
									printk("\n");
							}
							printk("\n");
#endif
						}
					}
					//DefaultKeyID
					if(RTMPGetKeyParameter("Default", "DefaultKeyID", tmpbuf, 10, buffer))
					{
						KeyIdx = simple_strtol(tmpbuf, 0, 10);
						if((KeyIdx >= 1 ) && (KeyIdx <= 4))
							pAd->PortCfg.DefaultKeyId = (UCHAR) (KeyIdx - 1 );
						else
							pAd->PortCfg.DefaultKeyId = 0;

						
						DBGPRINT(RT_DEBUG_TRACE, "DefaultKeyID=%d\n", pAd->PortCfg.DefaultKeyId);
					}
					//Key1Type
					if(RTMPGetKeyParameter("Default", "Key1Type", tmpbuf, 10, buffer))
					{
						KeyType = simple_strtol(tmpbuf, 0, 10);

						//Key1Str
						if(RTMPGetKeyParameter("Default", "Key1Str", tmpbuf, 255, buffer))
						{	
							KeyLen = strlen(tmpbuf);
							if(KeyType == 0)
							{//Hex type
								if( (KeyLen == 10) || (KeyLen == 26))
								{
									pAd->PortCfg.SharedKey[0].KeyLen = KeyLen / 2;
									AtoH(tmpbuf, pAd->PortCfg.SharedKey[0].Key, KeyLen / 2);

									DBGPRINT(RT_DEBUG_TRACE, "Key1Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key1Str is Invalid key length!\n");
								}								
							}
							else
							{//Ascii								
								if( (KeyLen == 5) || (KeyLen == 13))
								{
									pAd->PortCfg.SharedKey[0].KeyLen = KeyLen;
									NdisMoveMemory(pAd->PortCfg.SharedKey[0].Key, tmpbuf, KeyLen);
							
									DBGPRINT(RT_DEBUG_TRACE, "Key1Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");		
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key1Str is Invalid key length!\n");
								}
							}
						}
					}
					//Key2Type
					if(RTMPGetKeyParameter("Default", "Key2Type", tmpbuf, 10, buffer))
					{
						KeyType = simple_strtol(tmpbuf, 0, 10);
						//Key2Str
						if(RTMPGetKeyParameter("Default", "Key2Str", tmpbuf, 255, buffer))
						{	
							KeyLen = strlen(tmpbuf);
							if(KeyType == 0)
							{//Hex type
								if( (KeyLen == 10) || (KeyLen == 26))
								{
									pAd->PortCfg.SharedKey[1].KeyLen = KeyLen / 2;
									AtoH(tmpbuf, pAd->PortCfg.SharedKey[1].Key, KeyLen / 2);

									DBGPRINT(RT_DEBUG_TRACE, "Key2Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key2Str is Invalid key length!\n");
								}								
							}
							else
							{//Ascii								
								if( (KeyLen == 5) || (KeyLen == 13))
								{
									pAd->PortCfg.SharedKey[1].KeyLen = KeyLen;
									NdisMoveMemory(pAd->PortCfg.SharedKey[1].Key, tmpbuf, KeyLen);

									DBGPRINT(RT_DEBUG_TRACE, "Key2Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");	
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key2Str is Invalid key length!\n");
								}
							}
						}
					}
					//Key3Type
					if(RTMPGetKeyParameter("Default", "Key3Type", tmpbuf, 10, buffer))
					{
						KeyType = simple_strtol(tmpbuf, 0, 10);
						//Key3Str
						if(RTMPGetKeyParameter("Default", "Key3Str", tmpbuf, 255, buffer))
						{	
							KeyLen = strlen(tmpbuf);
							if(KeyType == 0)
							{//Hex type
								if( (KeyLen == 10) || (KeyLen == 26))
								{
									pAd->PortCfg.SharedKey[2].KeyLen = KeyLen / 2;
									AtoH(tmpbuf, pAd->PortCfg.SharedKey[2].Key, KeyLen / 2);

									DBGPRINT(RT_DEBUG_TRACE, "Key3Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");	
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key3Str is Invalid key length!\n");
								}								
							}
							else
							{//Ascii								
								if( (KeyLen == 5) || (KeyLen == 13))
								{
									pAd->PortCfg.SharedKey[2].KeyLen = KeyLen;
									NdisMoveMemory(pAd->PortCfg.SharedKey[2].Key, tmpbuf, KeyLen);
							
									DBGPRINT(RT_DEBUG_TRACE, "Key3Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");	
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key3Str is Invalid key length!\n");
								}
							}
						}
					}
					//Key4Type
					if(RTMPGetKeyParameter("Default", "Key4Type", tmpbuf, 10, buffer))
					{
						KeyType = simple_strtol(tmpbuf, 0, 10);
						//Key4Str
						if(RTMPGetKeyParameter("Default", "Key4Str", tmpbuf, 255, buffer))
						{	
							KeyLen = strlen(tmpbuf);
							if(KeyType == 0)
							{//Hex type
								if( (KeyLen == 10) || (KeyLen == 26))
								{
									pAd->PortCfg.SharedKey[3].KeyLen = KeyLen / 2;
									AtoH(tmpbuf, pAd->PortCfg.SharedKey[3].Key, KeyLen / 2);

									DBGPRINT(RT_DEBUG_TRACE, "Key4Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key4Str is Invalid key length!\n");
								}								
							}
							else
							{//Ascii								
								if( (KeyLen == 5) || (KeyLen == 13))
								{
									pAd->PortCfg.SharedKey[3].KeyLen = KeyLen;
									NdisMoveMemory(pAd->PortCfg.SharedKey[3].Key, tmpbuf, KeyLen);
							
									DBGPRINT(RT_DEBUG_TRACE, "Key4Str=%s and type=%s\n", tmpbuf, (KeyType==0) ? "Hex":"Ascii");		
								}
								else
								{ //Invalid key length
									DBGPRINT(RT_DEBUG_TRACE, "Key4Str is Invalid key length!\n");
								}
							}
						}
					}				
					// set file parameter to portcfg
#if 0
					DBGPRINT(RT_DEBUG_TRACE, "Ssid=%s SsidLen=%d\n", profile.SSID, profile.SsidLen);
					pAd->PortCfg.SsidLen = profile.SsidLen;
					NdisMoveMemory(pAd->PortCfg.Ssid, profile.SSID, pAd->PortCfg.SsidLen);
					pAd->PortCfg.Channel = profile.Channel;
					pAd->PortCfg.DefaultKeyId = profile.KeyIndex;
					pAd->PortCfg.SharedKey[0].KeyLen = profile.KeyLength;
					NdisMoveMemory(pAd->PortCfg.SharedKey[0].Key, &profile.Key1, profile.KeyLength);
					pAd->PortCfg.SharedKey[1].KeyLen = profile.KeyLength;
					NdisMoveMemory(pAd->PortCfg.SharedKey[1].Key, &profile.Key2, profile.KeyLength);
					pAd->PortCfg.SharedKey[2].KeyLen = profile.KeyLength;
					NdisMoveMemory(pAd->PortCfg.SharedKey[2].Key, &profile.Key3, profile.KeyLength);
					pAd->PortCfg.SharedKey[3].KeyLen = profile.KeyLength;
					NdisMoveMemory(pAd->PortCfg.SharedKey[3].Key, &profile.Key4, profile.KeyLength);
					pAd->PortCfg.FragmentThreshold = profile.Fragment;
					pAd->PortCfg.RtsThreshold = profile.RTS;
					pAd->PortCfg.WindowsTxPreamble = profile.PreamType;
					pAd->PortCfg.TxRate = profile.TransRate;
					pAd->PortCfg.TxPower = profile.TransPower;
					// pAd->PortCfg.Psm = profile.PSmode;
					pAd->PortCfg.CurrentTxAntenna = profile.AntennaTx;
					pAd->PortCfg.CurrentRxAntenna = profile.AntennaRx;
					pAd->PortCfg.CountryRegion = profile.CountryRegion;
#endif
				}
			}
			else
			{
				DBGPRINT(RT_DEBUG_TRACE, "--> %s does not have a write method\n", src);
			}
			
			retval=filp_close(srcf,NULL);
			
			if (retval)
			{
				DBGPRINT(RT_DEBUG_TRACE, "--> Error %d closing %s\n", -retval, src);
			}
		}
	}
	
#if 0
	src = FILTERFILE_PATH;
	if (src && *src) 
	{
		srcf = filp_open(src, O_RDONLY, 0);
		if (IS_ERR(srcf)) 
		{
			DBGPRINT(RT_DEBUG_TRACE, "--> Error %ld opening %s\n", -PTR_ERR(srcf),src);    
		}
		else 
		{
			/* The object must have a read method */
			if (srcf->f_op && srcf->f_op->read) 
			{
				retval=srcf->f_op->read(srcf, (PUCHAR)&pAd->FilterEntry, sizeof(RT_802_11_FILTER_ENTRY), &srcf->f_pos);
				if (retval < 0)
				{
					DBGPRINT(RT_DEBUG_TRACE, "--> Read %s error %d\n", src, -retval);
				}
			}
			else
			{
				DBGPRINT(RT_DEBUG_TRACE, "--> %s does not have a write method\n", src);
			}
			
			retval=filp_close(srcf,NULL);
			
			if (retval)
			{
				DBGPRINT(RT_DEBUG_TRACE, "--> Error %d closing %s\n", -retval, src);
			}
		}
	}
#endif	
	set_fs(orgfs);
	current->fsuid = orgfsuid;
	current->fsgid = orgfsgid;
}	

/*
	========================================================================
	
	Routine Description:
		Compare two memory block

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		1:			memory are equal
		0:			memory are not equal

	Note:
		
	========================================================================
*/
ULONG	RTMPEqualMemory(
	IN	PVOID	pSrc1,
	IN	PVOID	pSrc2,
	IN	ULONG	Length)
{
	PUCHAR	pMem1;
	PUCHAR	pMem2;
	ULONG	Index = 0;

	pMem1 = (PUCHAR) pSrc1;
	pMem2 = (PUCHAR) pSrc2;

	for (Index = 0; Index < Length; Index++)
	{
		if (pMem1[Index] != pMem2[Index])
		{
			break;
		}
	}

	if (Index == Length)
	{
		return (1);
	}
	else
	{
		return (0);
        }
}

VOID	NdisZeroMemory(
	IN	PVOID	pSrc,
	IN	ULONG	Length)
{
	PUCHAR	pMem;
	ULONG	Index = 0;

	pMem = (PUCHAR) pSrc;

	for (Index = 0; Index < Length; Index++)
	{
		pMem[Index] = 0x00;
	}
}

VOID	NdisFillMemory(
	IN	PVOID	pSrc,
	IN	ULONG	Length,
	IN	UCHAR	Fill)
{
	PUCHAR	pMem;
	ULONG	Index = 0;

	pMem = (PUCHAR) pSrc;

	for (Index = 0; Index < Length; Index++)
	{
		pMem[Index] = Fill;
	}
}

VOID	NdisMoveMemory(
	OUT	PVOID	pDest,
	IN	PVOID	pSrc,
	IN	ULONG	Length)
{
	PUCHAR	pMem1;
	PUCHAR	pMem2;
	ULONG	Index = 0;

	pMem1 = (PUCHAR) pDest;
	pMem2 = (PUCHAR) pSrc;

	for (Index = 0; Index < Length; Index++)
	{
		pMem1[Index] = pMem2[Index];
	}
}

/*
	========================================================================
	
	Routine Description:
		Initialize port configuration structure

	Arguments:
		Adapter						Pointer to our adapter

	Return Value:
		None

	Note:
		
	========================================================================
*/
VOID	PortCfgInit(
	IN	PRTMP_ADAPTER pAdapter)
{
    UINT i;
	ULONG	Now;
    // change the RSN_IE according to the initial value. RSN_IE[21]=1 -> WPA, RSN_IE[21]=2 ->WPAPSK...etc
    UCHAR	CipherSuiteWPAPSKTKIP[] = {
		0x00, 0x50, 0xf2, 0x01,	// oui
		0x01, 0x00,				// Version
		0x00, 0x50, 0xf2, 0x02,	// Multicast
		0x01, 0x00,				// Number of unicast
		0x00, 0x50, 0xf2, 0x02,	// unicast
		0x01, 0x00,				// number of authentication method
		0x00, 0x50, 0xf2, 0x02	// authentication
		};
    UCHAR	CipherSuiteWPAPSKTKIPLen = (sizeof(CipherSuiteWPAPSKTKIP) / sizeof(UCHAR));

    DBGPRINT(RT_DEBUG_TRACE, "--> PortCfgInit\n");    
    
    pAdapter->PortCfg.MediumOccupancyLimit = 10000;
    pAdapter->PortCfg.CapabilityInfo = 0x0000;
	pAdapter->PortCfg.BANClass3Data = FALSE;
	pAdapter->PortCfg.RSNIE_Len = 0x16;
	NdisZeroMemory(pAdapter->PortCfg.RSN_IE, MAX_LEN_OF_RSNIE);
	NdisMoveMemory(pAdapter->PortCfg.RSN_IE, CipherSuiteWPAPSKTKIP, CipherSuiteWPAPSKTKIPLen );
    NdisZeroMemory(pAdapter->PortCfg.R_Counter, LEN_KEY_DESC_REPLAY);
	NdisZeroMemory(pAdapter->PortCfg.Key_Counter, sizeof(pAdapter->PortCfg.Key_Counter));
	NdisZeroMemory(pAdapter->PortCfg.PMK, sizeof(pAdapter->PortCfg.PMK));

	pAdapter->PortCfg.WPAREKEY.ReKeyMethod = TIME_REKEY;
	pAdapter->PortCfg.WPAREKEY.ReKeyInterval = 88;

	pAdapter->PortCfg.AuthMode = Ndis802_11AuthModeOpen;
	pAdapter->PortCfg.WepStatus = Ndis802_11WEPDisabled;

	// if (pAdapter->PortCfg.AuthMode >= Ndis802_11AuthModeWPA)
	{	
		//  the same action case OID_AUTHMODE in rtmp_info.c 
		pAdapter->PortCfg.WpaGTKState = SETKEYS_DONE;
		pAdapter->PortCfg.GKeyDoneStations = 0;
		pAdapter->PortCfg.CapabilityInfo |= 0x10;    
		pAdapter->PortCfg.MICFailureCounter = 0;
	}
    pAdapter->PortCfg.WPAGKeyID = 1;
    for (i = 1; i < GROUP_KEY_NO; i++)
    {
        NdisZeroMemory(pAdapter->PortCfg.GroupKey[i].RxTsc, 6);
        pAdapter->PortCfg.GroupKey[i].RxTsc[0]=0x01;
        NdisZeroMemory(pAdapter->PortCfg.GroupKey[i].TxTsc, 6);
        pAdapter->PortCfg.GroupKey[i].KeyLen = 16;
    }
    pAdapter->PortCfg.AuthRspTimeout = AUTH_KEY_TIMEOUT;   // in msec
    pAdapter->PortCfg.PortSecured = WPA_802_1X_PORT_SECURED;
    pAdapter->PortCfg.PrivacyFilter = Ndis802_11PrivFilterAcceptAll;

    pAdapter->PortCfg.SupportedRates[0] = 0x82; // in units of 0.5 Mbps
    pAdapter->PortCfg.SupportedRates[1] = 0x84; // in units of 0.5 Mbps
    pAdapter->PortCfg.SupportedRates[2] = 0x8B; // in units of 0.5 Mbps
    pAdapter->PortCfg.SupportedRates[3] = 0x96; // in units of 0.5 Mbps
    pAdapter->PortCfg.SupportedRatesLen = 4;

    pAdapter->PortCfg.DesiredRates[0] = 2;  // in units of 0.5 Mbps
    pAdapter->PortCfg.DesiredRates[1] = 4;  // in units of 0.5 Mbps
    pAdapter->PortCfg.DesiredRates[2] = 11; // in units of 0.5 Mbps
    pAdapter->PortCfg.DesiredRates[3] = 22; // in units of 0.5 Mbps

    pAdapter->PortCfg.MaxBasicRate = RATE_2;
    pAdapter->PortCfg.MaxTxRate = RATE_11;  // intersection of DesiredRates[] and SupportedRates[]
    pAdapter->PortCfg.TxRate = RATE_11;

    pAdapter->PortCfg.BeaconPeriod = 100;     // in mSec
    //pAdapter->PortCfg.DtimPeriod = 1;         // never mind, decided by AP later
    pAdapter->PortCfg.AssocRspTimeout = ASSOC_TIMEOUT;  // in mSec

    pAdapter->PortCfg.CfpMaxDuration = 0;     // never mind, decided by AP later
    pAdapter->PortCfg.CfpDurRemain = 0;       // never mind, decided by AP later
    pAdapter->PortCfg.CfpCount = 0;           // never mind, decided by AP later
    pAdapter->PortCfg.CfpPeriod = 0;          // never mind, decided by AP later
    
    for(i = 0; i < SHARE_KEY_NO; i++) {
        pAdapter->PortCfg.SharedKey[i].KeyLen = 0;
    }

    pAdapter->PortCfg.DefaultKeyId = 0;

    //NdisMoveMemory(pAdapter->PortCfg.StaAddr.Octet, pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);

    pAdapter->PortCfg.RtsThreshold = 2347;
    pAdapter->PortCfg.ShortRetryLimit = 7;
    pAdapter->PortCfg.LongRetryLimit = 4;
    pAdapter->PortCfg.FragmentThreshold = 2346;
    pAdapter->PortCfg.bFragmentZeroDisable = FALSE;
    pAdapter->PortCfg.MaxTxLifeTime = 512;
    pAdapter->PortCfg.MaxRxLifeTime = 512;
    NdisMoveMemory(pAdapter->PortCfg.ManufactureId, "Ralink Technology", 17);
    pAdapter->PortCfg.ManufactureIdLen = 17;
    NdisMoveMemory(pAdapter->PortCfg.ProductId, "802.11b NIC", 11);
    pAdapter->PortCfg.ProductIdLen = 11;
    pAdapter->PortCfg.CurrentRegDomain = 0x10;

    pAdapter->PortCfg.CurrentTxAntenna = 0xff;		// To match OID definition
    pAdapter->PortCfg.CurrentRxAntenna = 0xff;		// To match OID definition
    pAdapter->PortCfg.NumberOfAntenna = 2;

    pAdapter->PortCfg.TxPowerLevel[0] = 100;
    pAdapter->PortCfg.NumOfTxPowerLevel = 1;
    pAdapter->PortCfg.TxPower = 100; //mW
    
    pAdapter->PortCfg.CcaModeSupported = 4;
    pAdapter->PortCfg.CurrentCcaMode = 4;
    pAdapter->PortCfg.EdThreshold = 0;
    //pAdapter->PortCfg.DomainSupported[0] = 0x10;
    //pAdapter->PortCfg.NumOfSupportedDomain = 1;
    pAdapter->PortCfg.AntennaSupportTx = TRUE;
    pAdapter->PortCfg.AntennaSupportRx = TRUE;
    pAdapter->PortCfg.AntennaSupportDiversityRx = TRUE;

    //pAdapter->PortCfg.CwMin = CW_MIN_IN_BITS;
    //pAdapter->PortCfg.CwMax = CW_MAX_IN_BITS;
    pAdapter->PortCfg.RecvDtim = TRUE;
    //pAdapter->PortCfg.DtimCount = 0;
    NdisZeroMemory(&pAdapter->PortCfg.Bssid, MAC_ADDR_LEN);
    NdisFillMemory(&pAdapter->PortCfg.Broadcast, MAC_ADDR_LEN, 0xff);
//  pAdapter->PortCfg.Pss = PWR_ACTIVE;
    pAdapter->PortCfg.RssiTrigger = 0;
    pAdapter->PortCfg.LastRssi = 60;
    pAdapter->PortCfg.RssiTriggerMode = RSSI_TRIGGERED_UPON_BELOW_THRESHOLD;
    pAdapter->PortCfg.AtimWin = 0;
    pAdapter->PortCfg.Channel = 1;

    pAdapter->PortCfg.Aid = 1;

    pAdapter->PortCfg.DefaultListenCount = 3;//default listen count;
    pAdapter->PortCfg.CurrListenCount = 0;  // in units of Beacon interval
    pAdapter->PortCfg.MinChnlTime = MIN_CHANNEL_TIME;   // minimum channel time (in mSEC)
    pAdapter->PortCfg.MaxChnlTime = MAX_CHANNEL_TIME; // maximum channel time (in mSEC)

    pAdapter->PortCfg.SsidLen = 0;
    NdisZeroMemory(pAdapter->PortCfg.Ssid, MAX_LEN_OF_SSID);  // NOT NULL-terminated

	if((pAdapter->PortCfg.SsidLen <= 0) || (pAdapter->PortCfg.SsidLen >= MAX_LEN_OF_SSID))
	{
		NdisMoveMemory(pAdapter->PortCfg.Ssid, "SoftAP-", 7);
		pAdapter->PortCfg.Ssid[7] = (((pAdapter->CurrentAddress[5] >> 4) > 9) ? ((pAdapter->CurrentAddress[5] >> 4) - 10 + 'A') : ((pAdapter->CurrentAddress[5] >> 4) + '0'));
		pAdapter->PortCfg.Ssid[8] = (((pAdapter->CurrentAddress[5] & 0x0F) > 9) ? ((pAdapter->CurrentAddress[5]& 0x0F) - 10 + 'A') : ((pAdapter->CurrentAddress[5]& 0x0F) + '0'));
		pAdapter->PortCfg.SsidLen = 9;
		DBGPRINT(RT_DEBUG_TRACE, "DefApSSID (Len=%d, ssid=%s...)\n", pAdapter->PortCfg.SsidLen, pAdapter->PortCfg.Ssid);
	}

    // global variables mXXXX used in MAC protocol state machines
    pAdapter->PortCfg.Mibss = FALSE;
    pAdapter->PortCfg.Massoc = TRUE;

    // PHY specification
    pAdapter->PortCfg.PhyMode = PHY_11B;    // PHY_11A, PHY_11G, PHY_11B
    pAdapter->PortCfg.Drxtx = 0;      // RF RX to TX turn around time, in units of usec
    pAdapter->PortCfg.Dswchnl = 0;    // in units of usec, time to switch to a new channel 
    pAdapter->PortCfg.Dpreamble = 144;  // in units of usec 
    pAdapter->PortCfg.Dplcp = 48;      // in units of usec 
    pAdapter->PortCfg.Dslot = 20;      // in units of usec 
    pAdapter->PortCfg.Dsifs = 10;      // in units of usec 
    pAdapter->PortCfg.Dpifs = 30;      // in units of usec 
    pAdapter->PortCfg.Ddifs = 50;      // in units of usec 
    pAdapter->PortCfg.Deifs = 364;     // in units of usec, change when BasicRateSet changes
    pAdapter->PortCfg.TxPreamble = Rt802_11PreambleLong; // use Long preamble on TX by defaut

    // user desired power mode
    pAdapter->PortCfg.WindowsPowerMode = Ndis802_11PowerModeCAM; // Ndis802_11PowerModeFast_PSP;
    pAdapter->PortCfg.WindowsBatteryPowerMode = Ndis802_11PowerModeCAM; // Ndis802_11PowerModeFast_PSP;
    pAdapter->PortCfg.WindowsTxPreamble = Rt802_11PreambleAuto; // use Long preamble on TX by defaut
    pAdapter->PortCfg.WindowsACCAMEnable = FALSE;
//    pAdapter->PortCfg.PacketFilter = NDIS_PACKET_TYPE_ALL_MULTICAST | NDIS_PACKET_TYPE_DIRECTED | NDIS_PACKET_TYPE_BROADCAST;
	pAdapter->bAcceptDirect = TRUE;
	pAdapter->bAcceptMulticast = FALSE;
	pAdapter->bAcceptBroadcast = TRUE;
	pAdapter->bAcceptAllMulticast = TRUE;

    pAdapter->PortCfg.RfType = RFIC_2420;
    pAdapter->PortCfg.LedMode = LED_MODE_DEFAULT;

    init_timer(&pAdapter->PortCfg.RfTuningTimer);
    pAdapter->PortCfg.RfTuningTimer.data = (unsigned long)pAdapter;
    pAdapter->PortCfg.RfTuningTimer.function = &AsicRfTuningExec;				/* timer handler */
	
	// Patch for Ndtest
	pAdapter->PortCfg.AutoReconnect = FALSE;
    pAdapter->PortCfg.IgnoredScanNumber = 0;

    pAdapter->PortCfg.CountryRegion = REGION_MKK1;
	
    pAdapter->bContinuousTx == FALSE;
    pAdapter->bContinuousRx == FALSE;
	
    pAdapter->SpinLockCount = 0;
	
    DBGPRINT(RT_DEBUG_TRACE, "<-- PortCfgInit\n");
}

UCHAR BtoH(char ch)
{
    if (ch >= '0' && ch <= '9') return (ch - '0');        // Handle numerals
    if (ch >= 'A' && ch <= 'F') return (ch - 'A' + 0xA);  // Handle capitol hex digits
    if (ch >= 'a' && ch <= 'f') return (ch - 'a' + 0xA);  // Handle small hex digits
    return(255);
}

//
//  FUNCTION: AtoH(char *, UCHAR *, int)
//
//  PURPOSE:  Converts ascii string to network order hex
//
//  PARAMETERS:
//    src    - pointer to input ascii string
//    dest   - pointer to output hex
//    destlen - size of dest
//
//  COMMENTS:
//
//    2 ascii bytes make a hex byte so must put 1st ascii byte of pair
//    into upper nibble and 2nd ascii byte of pair into lower nibble.
//

void AtoH(char * src, UCHAR * dest, int destlen)
{
    char *srcptr;
	PUCHAR destTemp;

    srcptr = src;	
	destTemp = (PUCHAR) dest; 

    while(destlen--)
    {
		*destTemp = BtoH(*srcptr++) << 4;    // Put 1st ascii byte in upper nibble.
		*destTemp += BtoH(*srcptr++);      // Add 2nd ascii byte to above.
		destTemp++;
    }
}

