/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************/

#include "rt_config.h"

#define MAX_NUM_OF_FREE_NDIS_PACKET     32

char const *pEventText[EVENT_MAX_EVENT_TYPE] = {
    "restart access point",
    "successfully associated",
    "has disassociated",
    "has been aged-out and disassociated",
    "Active CounterMeature",
    "Invalid PSK password"};

NDIS_STATUS ApInitialize(
    IN  PRTMP_ADAPTER   pAd)
{
    NDIS_STATUS     Status = NDIS_STATUS_SUCCESS;
    UCHAR       GTK[TKIP_GTK_LENGTH];

    DBGPRINT(RT_DEBUG_TRACE, "---> ApInitialize\n");

    // initialize MAC table and allocate spin lock
    NdisZeroMemory(&pAd->MacTab, sizeof(MAC_TABLE));
    spin_lock_init(&pAd->MacTabLock);

     // Init TKIP Group-Key-related variables
    GenRandom(pAd,pAd->PortCfg.GMK);
    GenRandom(pAd,pAd->PortCfg.GNonce);
    CountGTK(pAd->PortCfg.GMK,  (UCHAR*)pAd->PortCfg.GNonce, pAd->CurrentAddress, pAd->PortCfg.GroupKey[0].Key, TKIP_GTK_LENGTH);
    NdisMoveMemory(pAd->PortCfg.GroupKey[pAd->PortCfg.WPAGKeyID].Key, GTK, 16);
    NdisMoveMemory(pAd->PortCfg.GroupKey[pAd->PortCfg.WPAGKeyID].TxMic, &GTK[16], 8);
    NdisMoveMemory(pAd->PortCfg.GroupKey[pAd->PortCfg.WPAGKeyID].RxMic, &GTK[24], 8);            
    // Init Group key update timer, and countermeasures timer
    init_timer(&pAd->PortCfg.REKEYTimer);
    pAd->PortCfg.REKEYTimer.data = (unsigned long)pAd;
    pAd->PortCfg.REKEYTimer.function = &GREKEYPeriodicExec;

    init_timer(&pAd->PortCfg.CounterMeasureTimer);
    pAd->PortCfg.CounterMeasureTimer.data = (unsigned long)pAd;
    pAd->PortCfg.CounterMeasureTimer.function = &CMTimerExec;
    DBGPRINT(RT_DEBUG_TRACE, "<--- ApInitialize\n");

    return Status;
}

VOID ApShutdown(
    IN PRTMP_ADAPTER pAd)
{
    DBGPRINT(RT_DEBUG_TRACE, "---> ApShutdown\n");    
    ApStop(pAd);
    DBGPRINT(RT_DEBUG_TRACE, "<--- ApShutdown\n");
}

/*
    ==========================================================================
    Description:
        Start AP service. If any vital AP parameter is changed, a STOP-START
        sequence is required to disassociate all STAs.
    Note:
    ==========================================================================
 */
VOID ApStartUp(
    IN PRTMP_ADAPTER pAd) 
{
    UCHAR         Ssid[MAX_LEN_OF_SSID], SsidLen; 
    BOOLEAN       TimerCancelled;

    //NICEnableInterrupt(pAd);
    if ((pAd->PortCfg.SsidLen <= 0) || (pAd->PortCfg.SsidLen >= MAX_LEN_OF_SSID))
        return;
    
    DBGPRINT(RT_DEBUG_TRACE, "!!! ApStartUp !!!\n");

    pAd->PortCfg.CapabilityInfo = CAP_GENERATE(1, 0, 0, 0, (pAd->PortCfg.WepStatus != Ndis802_11EncryptionDisabled) );
    pAd->PortCfg.AtimWin           = 0;
        
    AsicSetBssid(pAd, (MACADDR *)pAd->CurrentAddress); 
    AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
    AsicLockChannel(pAd, pAd->PortCfg.Channel);
    MlmeSetTxPreamble(pAd, pAd->PortCfg.TxPreamble);
    MlmeUpdateTxRates(pAd);

    // start sending BEACON out        
    MakeBssBeacon(pAd);
    AsicEnableBssSync(pAd);
    ASIC_LED_ACT_ON(pAd);

    // Group rekey related
    if ((pAd->PortCfg.WPAREKEY.ReKeyInterval != 0) && ((pAd->PortCfg.WPAREKEY.ReKeyMethod == TIME_REKEY) || (pAd->PortCfg.WPAREKEY.ReKeyMethod == PKT_REKEY))) 
    {
        // Regularly check the timer
        if (pAd->PortCfg.REKEYTimerRunning == FALSE)
        {
            pAd->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
            add_timer(&pAd->PortCfg.REKEYTimer);
            pAd->PortCfg.REKEYTimerRunning = TRUE;
            pAd->PortCfg.REKEYCOUNTER = 0;
        }
    }
    else
        pAd->PortCfg.REKEYTimerRunning = FALSE;

	ApLogEvent(pAd, (PMACADDR)pAd->CurrentAddress, EVENT_RESET_ACCESS_POINT);
	// send SIGUSR1 to rt2500apd
	if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPA)
	{
        DBGPRINT(RT_DEBUG_TRACE, "!!!send SIGUSR1 to rt2400apd !!!\n");
		PUCHAR                  src;
		struct file             *srcf;
		INT                     retval, orgfsuid, orgfsgid;
		mm_segment_t            orgfs;
		CHAR                    buffer[MAX_INI_BUFFER_SIZE];
		CHAR                    tmpbuf[600];
		src = PROFILE_PATH;

		// Save uid and gid used for filesystem access.
		// Set user and group to 0 (root)   
		orgfsuid = current->fsuid;
		orgfsgid = current->fsgid;
		current->fsuid=current->fsgid = 0;
		orgfs = get_fs();
		set_fs(KERNEL_DS);

        if (src && *src) 
        {
            srcf = filp_open(src, O_RDONLY, 0);
            if (IS_ERR(srcf)) 
            {
                DBGPRINT(RT_DEBUG_ERROR, "--> Error %ld opening %s\n", -PTR_ERR(srcf),src);    
            }
            else 
            {
                /* The object must have a read method */
                if (srcf->f_op && srcf->f_op->read) 
                {
                    memset(buffer, 0x00, MAX_INI_BUFFER_SIZE);
                    retval=srcf->f_op->read(srcf, buffer, MAX_INI_BUFFER_SIZE, &srcf->f_pos);
                    if (retval < 0)
                    {
                        DBGPRINT(RT_DEBUG_ERROR, "--> Read %s error %d\n", src, -retval);
                    }
                    else
                    {
                        // set file parameter to portcfg
                        //CountryRegion
                        if(RTMPGetKeyParameter("Default", "pid", tmpbuf, 255, buffer))
                        {
                            pAd->PortCfg.ApdPid= (ULONG) simple_strtol(tmpbuf, 0, 10);
                            DBGPRINT(RT_DEBUG_TRACE, "ApdPid=%d\n", pAd->PortCfg.ApdPid);
                            kill_proc(pAd->PortCfg.ApdPid, SIGUSR1, 0);
                        }
                    }
                }
                else
                {
                    DBGPRINT(RT_DEBUG_ERROR, "--> %s does not have a write method\n", src);
                }
        
                retval=filp_close(srcf,NULL);
        
                if (retval)
                {
                    DBGPRINT(RT_DEBUG_ERROR, "--> Error %d closing %s\n", -retval, src);
                }
            }
        }
        set_fs(orgfs);
        current->fsuid = orgfsuid;
        current->fsgid = orgfsgid;
    }
}

/*
    ==========================================================================
    Description:
        disassociate all STAs and stop AP service.
    Note:
    ==========================================================================
 */
VOID ApStop(
    IN PRTMP_ADAPTER pAd) 
{
    DBGPRINT(RT_DEBUG_TRACE, "!!! ApStop !!!\n");
    //NICDisableInterrupt(pAd);
    if(pAd->PortCfg.REKEYTimerRunning==TRUE)
    {
        del_timer_sync(&pAd->PortCfg.REKEYTimer);
        pAd->PortCfg.REKEYTimerRunning=FALSE;
    }
    
    if(pAd->PortCfg.CMTimerRunning==TRUE)
    {
        del_timer_sync(&pAd->PortCfg.CounterMeasureTimer);
        pAd->PortCfg.CMTimerRunning = FALSE;
    }

    MacTableReset(pAd);
    AsicDisableSync(pAd);
    ASIC_LED_ACT_OFF(pAd);
}

VOID CleanupPsQueue(
    IN  PRTMP_ADAPTER   pAd,
    IN  PQUEUE_HEADER   pQueue)
{
    struct sk_buff  *skb;

    while (pQueue->Head)
    {
        skb = (struct sk_buff *) RemoveHeadQueue(pQueue);
        if(skb)
            RTMPFreeSkbBuffer(skb);
    }
}

VOID MacTableReset(
    IN  PRTMP_ADAPTER  pAd)
{
    int i;
    RTMPAcquireSpinLock(pAd, &pAd->MacTabLock);
    CleanupPsQueue(pAd, &pAd->MacTab.McastPsQueue);
    for (i=0; i<MAX_LEN_OF_MAC_TABLE; i++)
    {
        if (pAd->MacTab.Content[i].Valid)
        {
            if(pAd->MacTab.Content[i].RetryTimerRunning == TRUE)
            {
                DBGPRINT(RT_DEBUG_TRACE, " pAd->MacTab.Content[%d]  Cancel Retry Timer ! \n",i);                
                del_timer_sync(&pAd->MacTab.Content[i].RetryTimer);
                pAd->MacTab.Content[i].RetryTimerRunning = FALSE;
            }
            CleanupPsQueue(pAd, &pAd->MacTab.Content[i].PsQueue);
        }
    }
    NdisZeroMemory(&pAd->MacTab, sizeof(MAC_TABLE));
    InitializeQueueHeader(&pAd->MacTab.McastPsQueue);
    RTMPFreeSpinLock(pAd, &pAd->MacTabLock);
}

/*
    ==========================================================================
    Description:
        Add and new entry into MAC table
    ==========================================================================
 */
MAC_TABLE_ENTRY *MacTableInsertEntry(
    IN  PRTMP_ADAPTER   pAd, 
    IN  PMACADDR pAddr) 
{
    UCHAR Cidx, HashIdx;
    int i;
    MAC_TABLE_ENTRY *pEntry = NULL, *pCurrEntry;

    // if FULL, return
    if (pAd->MacTab.Size >= MAX_LEN_OF_MAC_TABLE) 
        return NULL;

    // allocate one MAC entry
    RTMPAcquireSpinLock(pAd, &pAd->MacTabLock);
    for (i = 0; i< MAX_LEN_OF_MAC_TABLE; i++)
    {
        // pick up the first available vacancy
        if (pAd->MacTab.Content[i].Valid == FALSE)
        {
            pEntry = &pAd->MacTab.Content[i];
            NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
            pEntry->Valid = TRUE;
            // Flag for judging if AP needs to disassoc the station.
            pEntry->Flag=0;
            // Indicator for retry timer running state, 
            pEntry->RetryTimerRunning=FALSE;
            // Indicator that which entry invokes counter measure attack
            pEntry->CMTimerRunning=FALSE;
            // Replay counter for the EAPOL packets
            NdisMoveMemory(pEntry->R_Counter, pAd->PortCfg.R_Counter, sizeof(pEntry->R_Counter));
            pEntry->ReTryCounter=0;
            pEntry->AuthMode=pAd->PortCfg.AuthMode;
            if(pEntry->AuthMode>=Ndis802_11AuthModeWPA)
            {
                pEntry->WpaState=AS_INITIALIZE;
                //  default is TKIP
                pEntry->WepStatus=Ndis802_11Encryption2Enabled;
            }
            else if(pAd->PortCfg.AuthMode==Ndis802_11AuthModeOpen)
            {
                pEntry->WpaState=AS_NOTUSE;
                pEntry->WepStatus=Ndis802_11WEPDisabled;
            }
            else    
            {
                pEntry->WpaState=AS_NOTUSE;
                pEntry->WepStatus=Ndis802_11Encryption1Enabled;
            }
            // per entry indicator of group key installed or not, if all entry have GTKState equals REKEY_ESTABLISHED, AP use groupkey to encrypt mcast
            pEntry->GTKState=REKRY_NEGOTIATING;
            pEntry->PrivacyFilter=Ndis802_11PrivFilterAcceptAll;        
            pEntry->PairwiseKey[0].KeyLen=0;
            pEntry->PortSecured=WPA_802_1X_PORT_NOT_SECURED;
            pEntry->RSNIE_Len=0;
            pEntry->Addr = *pAddr;
            pEntry->Sst = SST_NOT_AUTH;
            pEntry->AuthState = AS_NOT_AUTH;
            pEntry->PsMode = PWR_ACTIVE;
            pEntry->Rate = RATE_11;
            pEntry->PsQIdleCount = 0;
            pEntry->NoDataIdleCount = 0;
            InitializeQueueHeader(&pEntry->PsQueue);
            pAd->MacTab.Size ++;
            DBGPRINT(RT_DEBUG_TRACE, "MacTableInsertEntry - allocate entry #%d, Total= %d\n",i, pAd->MacTab.Size);
            break;
        }
    }

    // add this MAC entry into HASH table
    if (pEntry)
    {
        HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);
        if (pAd->MacTab.Hash[HashIdx] == NULL)
        {
            pAd->MacTab.Hash[HashIdx] = pEntry;
        }
        else
        {
            pCurrEntry = pAd->MacTab.Hash[HashIdx];
            while (pCurrEntry->pNext != NULL)
                pCurrEntry = pCurrEntry->pNext;
            pCurrEntry->pNext = pEntry;
        }
    }

    RTMPFreeSpinLock(pAd, &pAd->MacTabLock);
    return pEntry;
}

/*! \brief delete an entry inside the table
 *  \param Addr the mac address
 *  \return TRUE if the deletion is successful, FALSE otherwise
 *  \pre
 *  \post
 */
BOOLEAN MacTableDeleteEntry(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR pAddr) 
{
    USHORT HashIdx;
    MAC_TABLE_ENTRY *pEntry, *pPrevEntry;

    RTMPAcquireSpinLock(pAd, &pAd->MacTabLock);
    HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);
    pEntry = pAd->MacTab.Hash[HashIdx];
    if (pEntry)
    {
        if(pEntry->RetryTimerRunning==TRUE)
        {
            DBGPRINT(RT_DEBUG_TRACE, " MacTableDeleteEntry :   Cancel Retry Timer ! \n");
            del_timer_sync(&pEntry->RetryTimer);
            pEntry->RetryTimerRunning=FALSE;                
        }

        if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr))
        {
            pAd->MacTab.Hash[HashIdx] = pEntry->pNext;
            CleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
            NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
            pAd->MacTab.Size --;
            DBGPRINT(RT_DEBUG_TRACE, "MacTableDeleteEntry - Total= %d\n", pAd->MacTab.Size);
        }
        else
        {
            while (pEntry->pNext)
            {
                pPrevEntry = pEntry;
                pEntry = pEntry->pNext;
                if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr))
                {
                    pPrevEntry->pNext = pEntry->pNext;
                    CleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
                    NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
                    pAd->MacTab.Size --;
                    DBGPRINT(RT_DEBUG_TRACE, "MacTableDeleteEntry - Total= %d\n", pAd->MacTab.Size);
                    break;
                }
            }
        }
    }

    RTMPFreeSpinLock(pAd, &pAd->MacTabLock);
    return TRUE;
}

/*
    ==========================================================================
    Description:
        Look up the MAC address in the MAC table. Return NULL if not found.
    Return:
        pEntry - pointer to the MAC entry; NULL is not found
    ==========================================================================
*/
MAC_TABLE_ENTRY *MacTableLookup(
    IN PRTMP_ADAPTER pAd, 
    MACADDR *pAddr) 
{
    ULONG HashIdx, Pidx;
    MAC_TABLE_ENTRY *pEntry;

    HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);
    pEntry = pAd->MacTab.Hash[HashIdx];

    while (pEntry && pEntry->Valid)
    {
        if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr)) 
        {
            break;
        }
        else
            pEntry = pEntry->pNext;
    }
    return pEntry;
}

/*
    ==========================================================================
    Description:
        Look up a STA MAC table. Return its Sst to decide if an incoming
        frame from this STA or an outgoing frame to this STA is permitted.
    Return:
    ==========================================================================
*/

MAC_TABLE_ENTRY *SsPsInquiry(
    IN  PRTMP_ADAPTER   pAd, 
    IN MACADDR *pAddr, 
    OUT SST   *Sst, 
    OUT USHORT *Aid,
    OUT UCHAR *PsMode,
    OUT UCHAR *Rate) 
{
    MAC_TABLE_ENTRY *pEntry = NULL;

    if (MAC_ADDR_IS_GROUP(*pAddr)) // mcast & broadcast address
    {
        *Sst        = SST_ASSOC;
        *Aid        = 0;
        *PsMode     = PWR_ACTIVE;
        *Rate       = pAd->PortCfg.MaxBasicRate;
    } 
    else // unicast address
    {
        pEntry = MacTableLookup(pAd, pAddr);
        if (pEntry) 
        {
            *Sst        = pEntry->Sst;
            *Aid        = pEntry->Aid;
            *PsMode     = pEntry->PsMode;
            *Rate       = pEntry->Rate;
        } 
        else 
        {
            *Sst        = SST_NOT_AUTH;
            *Aid        = 0;
            *PsMode     = PWR_ACTIVE;
            *Rate       = 0;
        }
    }
    return pEntry;
}

/*! \brief set the station current power save mode
 *  \param Addr the Mac address
 *  \param psm power save mode (PWR_ACTIVE or PWR_SAVE)
 *  \return TRUE if the setting is successful FALSE if parameters are incorrect or is a group address
 *  \pre the address must inside the table already, otherwise it will not set any value
 *  \post
 */
BOOLEAN PsIndicate(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR pAddr, 
    IN UCHAR Psm) 
{
    MAC_TABLE_ENTRY *pEntry;

    pEntry = MacTableLookup(pAd, pAddr);
    if (pEntry) 
    {
        if ((pEntry->PsMode == PWR_SAVE) && (Psm == PWR_ACTIVE))
        {
            DBGPRINT(RT_DEBUG_TRACE, "PsIndicate - %02x:%02x:%02x:%02x:%02x:%02x wakes up, act like rx PS-POLL\n", pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]);
            // sleep station awakes, move all pending frames from PSQ to TXQ if any
            RTMPHandleRxPsPoll(pAd, pAddr, pEntry->Aid);
        }
        else if ((pEntry->PsMode != PWR_SAVE) && (Psm == PWR_SAVE))
        {
            DBGPRINT(RT_DEBUG_TRACE, "PsIndicate - %02x:%02x:%02x:%02x:%02x:%02x sleeps\n", pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]);
        }
        
        pEntry->NoDataIdleCount = 0;
        pEntry->PsMode = Psm;
    } 
    else 
    {
        // not in table, try to learn it ???? why bother?
    }
    return TRUE;
}


extern unsigned long get_cmos_time(void);
    
VOID ApLogEvent(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR pAddr,
    IN USHORT   Event)
{
    if (pAd->EventTab.Num < MAX_NUM_OF_EVENT)
    {
        RT_802_11_EVENT_LOG *pLog = &pAd->EventTab.Log[pAd->EventTab.Num];
        pLog->SystemTime = get_cmos_time();
        COPY_MAC_ADDR(&pLog->Addr, pAddr);
        pLog->Event = Event;
        DBGPRINT(RT_DEBUG_TRACE,"LOG#%d %02x:%02x:%02x:%02x:%02x:%02x %s\n",
            pAd->EventTab.Num, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], 
            pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], pEventText[Event]);

        pAd->EventTab.Num += 1;
    }
}

