/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************/

#include "rt_config.h"

/*
    ==========================================================================
    Description:
        The sync state machine, 
    Parameters:
        Sm - pointer to the state machine
    Note:
        the state machine looks like the following
        
                        SYNC_IDLE             
    MT2_PEER_PROBE_REQ  peer_probe_reqt_action    
    ==========================================================================
 */
VOID SyncStateMachineInit(
    IN PRTMP_ADAPTER pAd, 
    IN STATE_MACHINE *Sm, 
    OUT STATE_MACHINE_FUNC Trans[]) 
{
    StateMachineInit(Sm, (STATE_MACHINE_FUNC*)Trans, MAX_SYNC_STATE, MAX_SYNC_MSG, (STATE_MACHINE_FUNC)Drop, SYNC_IDLE, SYNC_MACHINE_BASE);

    StateMachineSetAction(Sm, SYNC_IDLE, MT2_PEER_PROBE_REQ, (STATE_MACHINE_FUNC)PeerProbeReqAction); 
}


VOID PeerProbeReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR       Addr2;
    CHAR          Ssid[MAX_LEN_OF_SSID];
    UCHAR         SsidLen, Rates[MAX_LEN_OF_SUPPORTED_RATES], RatesLen;
    MACHDR        ProbeRspHdr;
    NDIS_STATUS   NStatus;
    UCHAR         *OutBuffer = NULL;
    ULONG         FrameLen = 0;
    LARGE_INTEGER FakeTimestamp;
    UCHAR         SsidIe = IE_SSID, DsIe = IE_DS_PARM, IbssIe = IE_IBSS_PARM, SuppIe = IE_SUPP_RATES, 
                  DsLen = 1, IbssLen = 2, TimIe=IE_TIM, TimLen=1, // DtimCount=0, DtimPeriod=1,
                  BitmapControl=0, VirtualBitmap=0;
   	UCHAR			RSNIe=IE_WPA_RSN_IE;
	USHORT			CapabilityInfo = pAd->PortCfg.CapabilityInfo;

    if (PeerProbeReqSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, Ssid, &SsidLen, Rates, &RatesLen))
    {
        if ((SsidLen == 0) || RTMPEqualMemory(Ssid, pAd->PortCfg.Ssid, (ULONG) SsidLen))
        {
            // allocate and send out ProbeRsp frame

            OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
            if(OutBuffer == NULL)
            	return;
            
            DBGPRINT(RT_DEBUG_INFO, "SYNC - Send PROBE_RSP to %02x:%02x:%02x:%02x:%02x:%02x...\n", 
                Addr2.Octet[0],Addr2.Octet[1],Addr2.Octet[2],Addr2.Octet[3],Addr2.Octet[4],Addr2.Octet[5] );
            MgtMacHeaderInit(pAd, &ProbeRspHdr, SUBTYPE_PROBE_RSP, &Addr2);
    		if(pAd->PortCfg.AuthMode>=Ndis802_11AuthModeWPA)
    		{
    			CapabilityInfo |= 0x10;
    			
    			MakeOutgoingFrame(OutBuffer,          &FrameLen, 
    				MAC_HDR_LEN,                      (CHAR *)&ProbeRspHdr, 
    				TIMESTAMP_LEN,                    &FakeTimestamp,
    				2,                                &pAd->PortCfg.BeaconPeriod,
    				2,                                &CapabilityInfo,
    				1,                                &SsidIe, 
    				1,                                &pAd->PortCfg.SsidLen, 
    				pAd->PortCfg.SsidLen,             pAd->PortCfg.Ssid,
    				1,                                &SuppIe, 
    				1,                                &pAd->PortCfg.SupportedRatesLen,
    				pAd->PortCfg.SupportedRatesLen,   pAd->PortCfg.SupportedRates, 
    				1,                                &DsIe, 
    				1,                                &DsLen, 
    				1,                                &pAd->PortCfg.Channel,
    				1,                                &RSNIe,
    				1,                                &pAd->PortCfg.RSNIE_Len,
    				pAd->PortCfg.RSNIE_Len,			  pAd->PortCfg.RSN_IE,        
    				END_OF_ARGS);
    		}
    		else
    		{
                MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                                  MAC_HDR_LEN,                      (CHAR *)&ProbeRspHdr, 
                                  TIMESTAMP_LEN,                    &FakeTimestamp,
                                  2,                                &pAd->PortCfg.BeaconPeriod,
                                  2,                                &pAd->PortCfg.CapabilityInfo,
                                  1,                                &SsidIe, 
                                  1,                                &pAd->PortCfg.SsidLen, 
                                  pAd->PortCfg.SsidLen,             pAd->PortCfg.Ssid,
                                  1,                                &SuppIe, 
                                  1,                                &pAd->PortCfg.SupportedRatesLen,
                                  pAd->PortCfg.SupportedRatesLen,   pAd->PortCfg.SupportedRates, 
                                  1,                                &DsIe, 
                                  1,                                &DsLen, 
                                  1,                                &pAd->PortCfg.Channel,
                                  END_OF_ARGS);
            }
        
            MiniportMMRequest(pAd, OutBuffer, FrameLen);
        }
    }
}

