/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

	Module Name:
	rtmp_info.c

	Abstract:
	IOCTL related subroutines

	Revision History:
	Who         When          What
	--------    ----------    ----------------------------------------------
	Rory Chen   01-03-2003    created
	Paul Wu     29-07-2003    Support iwpriv

*/

#include	"rt_config.h"

#ifndef IW_ESSID_MAX_SIZE
/* Maximum size of the ESSID and NICKN strings */
#define IW_ESSID_MAX_SIZE	32
#endif

#define NR_WEP_KEYS 4
#define WEP_SMALL_KEY_LEN (40/8)
#define WEP_LARGE_KEY_LEN (104/8)

/* The frequency of each channel in MHz */
const long channel_frequency[] = {
        2412, 2417, 2422, 2427, 2432, 2437, 2442,
        2447, 2452, 2457, 2462, 2467, 2472, 2484
};
#define NUM_CHANNELS ( sizeof(channel_frequency) / sizeof(channel_frequency[0]) )

const char NULL_AP[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
const char BROADCAST_AP[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

UINT RssiQuality(
	IN	INT RSSI)
{
	unsigned int signal_quality;
    if (RSSI >= -40 )
        signal_quality = 100;
    else if (RSSI >= -80)    // between -40 ~ -80dbm
        signal_quality = 20 + (RSSI + 80) * 2;   
    else if (RSSI >= -84 )   // between -80 ~ -84dbm
        signal_quality = (RSSI + 84) * 5;   
    else    // < -84 dbm
        signal_quality = 0;

    return signal_quality;
}


INT isHex(char *src, int len)
{
	int i;

	for(i=0; i<len; i++)
	{
		 if( !(((*src >= '0') && (*src <= '9' )) ||
			 ((*src >= 'A') && (*src <= 'F' )) ||
			 ((*src >= 'a') && (*src <= 'f' ))) )
			return 0;
	}

	return 1;
}

INT RT2400_ioctl(
	IN	struct net_device	*net_dev, 
	IN	OUT	struct ifreq	*rq, 
	IN	INT					cmd)
{
	PRTMP_ADAPTER						pAd= net_dev->priv;
	struct iwreq						*wrq = (struct iwreq *) rq;
	UCHAR								TxValue,RxValue;
	USHORT								subcmd;
	BOOLEAN								StateMachineTouched = FALSE;
	ULONG								BssLen, ulInfo, KeyIdx, Rssi, Preamble, FcsValue;
	ULONG								Now;
	INT									Status = NDIS_STATUS_SUCCESS;
	UINT								i;
	UINT								ret;
	UINT				                val;
	INT                                 err = 0;
	NDIS_802_11_RTS_THRESHOLD			RtsThresh;
	NDIS_802_11_FRAGMENTATION_THRESHOLD	FragThresh;
	NDIS_802_11_BSSID_LIST				*pBssidList = NULL;
	NDIS_802_11_MAC_ADDRESS				Bssid;
	RT_802_11_LINK_STATUS				LinkStatus;
	NDIS_802_11_CONFIGURATION			Configuration;
	NDIS_802_11_SSID                    Ssid;
	NDIS_802_11_NETWORK_INFRASTRUCTURE	BssType;
	NDIS_802_11_ANTENNA					Antenna;
	NDIS_802_11_RATES					Rates;
	NDIS_802_11_TX_POWER_LEVEL			TxPowerLevel;
	NDIS_802_11_AUTHENTICATION_MODE		AuthMode;
	NDIS_802_11_WEP_STATUS				WepStatus;
	NDIS_802_11_WEP						WepKey;
	NDIS_802_11_CONFIGURATION			Config;
	NDIS_802_11_POWER_MODE				PowerMode;
	NDIS_802_11_STATISTICS				Statistics;
	RT_VERSION_INFO						DriverVersionInfo;
	RT_802_11_HARDWARE_REGISTER			HardwareRegister;
	BOOLEAN								RadioState;
	UCHAR   msg[255];
	PUCHAR  pinfo;
	UCHAR   mySsid[IW_ESSID_MAX_SIZE+1];
	PUCHAR  strptr;
	UINT	keylen;
	UCHAR   key[26+1];

	if(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE))
	{
		DBGPRINT(RT_DEBUG_TRACE, "INFO::Network is down!\n");
		return -ENETDOWN;
	}
	subcmd=wrq->u.data.flags;
	
	switch(cmd) {
		case SIOCGIWNAME:
			DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCGIWNAME\n");
			strcpy(wrq->u.name, "RT2400PCI");   
			break;
			
		case SIOCGIWESSID:  //Get ESSID
			{
				char *essid = NULL;
				struct iw_point *erq = &wrq->u.essid;

				if(pAd->MediaState == NdisMediaStateConnected)
				{
					erq->flags=1;
					erq->length = pAd->PortCfg.SsidLen;
					essid = pAd->PortCfg.Ssid;
				}
				else if(pAd->SsidLen) 
				{// not the ANY ssid
					erq->flags = 1;
					erq->length = pAd->SsidLen;
					essid = pAd->Ssid;
				}
				else
				{//the ANY ssid was specified
					/* report ANY back */
					erq->flags=0;
					erq->length=0;
				}

				if(erq->pointer)
				{
					if(copy_to_user(erq->pointer, essid, erq->length))
					{
						Status = -EFAULT;
						break;
					}
				}
				DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCGIWESSID (Len=%d, ssid=%s...)\n", erq->length, erq->pointer);
			}			
			break;

		case SIOCSIWESSID:  //Set ESSID
			{
				struct iw_point *erq = &wrq->u.essid;

				memset(&Ssid, 0x00, sizeof(NDIS_802_11_SSID));

				if (erq->flags) 
				{
					if (erq->length > IW_ESSID_MAX_SIZE)
					{
						Status = -E2BIG;
						break;
					}

					if (copy_from_user(Ssid.Ssid, erq->pointer, erq->length-1))
					{
						Status = -EFAULT;
						break;
					}
					Ssid.SsidLength = erq->length-1;	 //minus null character.
				}
				else
					Ssid.SsidLength = 0;  // ANY ssid 

				if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)
				{
					DBGPRINT(RT_DEBUG_TRACE, "!!! MLME busy, ignore this SIOCSIWESSID !!!");
					Status = -EBUSY;
					break;
				}

				pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE;
				MlmeEnqueue(&pAd->Mlme.Queue, 
				        MLME_CNTL_STATE_MACHINE, 
				        OID_802_11_SSID, 
						sizeof(NDIS_802_11_SSID),
						(VOID *)&Ssid
						);
				memcpy(pAd->Ssid, Ssid.Ssid, Ssid.SsidLength);
				pAd->SsidLen = Ssid.SsidLength;				            
 				pAd->PortCfg.AutoReconnect = TRUE;
				StateMachineTouched = TRUE;
				DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWESSID (Len=%d,Ssid=%s)\n", Ssid.SsidLength, Ssid.Ssid);	
				break;
			}
		case SIOCGIWNWID:
			Status = -EOPNOTSUPP;
			break;

		case SIOCSIWNWID:
			Status = -EOPNOTSUPP;
			break;

		case SIOCGIWRTS:  // Get the current RTS threshold
			wrq->u.rts.value = (INT) pAd->PortCfg.RtsThreshold;
			wrq->u.rts.disabled = (wrq->u.rts.value == MAX_RTS_THRESHOLD);
			wrq->u.rts.fixed = 1;			

			DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCGIWRTS (=%d)\n",wrq->u.rts.value);
			break;

		case SIOCSIWRTS:
			RtsThresh = wrq->u.rts.value;
			if(wrq->u.rts.disabled)
				RtsThresh = MAX_RTS_THRESHOLD;
			if((RtsThresh < 0) || (RtsThresh > MAX_RTS_THRESHOLD)) 
			{
				Status = -EINVAL;
				break;
			}
			else
				pAd->PortCfg.RtsThreshold = (USHORT) RtsThresh;
			DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWRTS (=%d)\n",RtsThresh);
			break;

		case SIOCGIWFRAG: // Get the current fragmentation threshold
			wrq->u.frag.value = pAd->PortCfg.FragmentThreshold;
			wrq->u.frag.disabled = (wrq->u.frag.value >= MAX_FRAG_THRESHOLD);
			wrq->u.frag.fixed = 1;
			break;
			
		case SIOCSIWFRAG:  // Set the desired fragmentation threshold
			FragThresh = wrq->u.frag.value;

			if(wrq->u.frag.disabled)
				FragThresh = MAX_FRAG_THRESHOLD;

			if((FragThresh < MIN_FRAG_THRESHOLD) || (FragThresh > MAX_FRAG_THRESHOLD))
			{
				Status = -EINVAL;
				break;
			}

			pAd->PortCfg.FragmentThreshold = (USHORT)FragThresh;
			DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWFRAG (=%d)\n", FragThresh);
			break;
#if WIRELESS_EXT > 11
		case SIOCGIWSTATS:
			// TODO: Get /proc/net/wireless stats
			Status = -EOPNOTSUPP;
			break;

		case SIOCSIWSTATS:
			Status = -EOPNOTSUPP;
			break;
#endif
		case SIOCGIWAP:
			if (INFRA_ON(pAd) || ADHOC_ON(pAd)) 
                        {
				wrq->u.ap_addr.sa_family = ARPHRD_ETHER;
				memcpy(wrq->u.ap_addr.sa_data, &pAd->PortCfg.Bssid, ETH_ALEN);
                                DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCGIWAP(=%02x:%02x:%02x:%02x:%02x:%02x)\n",
                                pAd->PortCfg.Bssid.Octet[0],pAd->PortCfg.Bssid.Octet[1],pAd->PortCfg.Bssid.Octet[2],
                                pAd->PortCfg.Bssid.Octet[3],pAd->PortCfg.Bssid.Octet[4],pAd->PortCfg.Bssid.Octet[5]);
                        } 
                        else 
                        {
                                DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCGIWAP(=EMPTY)\n");
                                Status = -ENOTCONN;
                        }
			break;
		case SIOCSIWNICKN:
			{
				struct iw_point *erq = &wrq->u.data;
				char nickn[IW_ESSID_MAX_SIZE+1];

				memset(nickn, 0, sizeof(nickn));
				if (erq->flags) 
				{
					if (erq->length > IW_ESSID_MAX_SIZE)
					{
						Status = -E2BIG;
						break;
					}
					if (copy_from_user(nickn, erq->pointer, erq->length))
					{
						Status = -EFAULT;
			break;
					}
					strcpy(pAd->nickn, nickn);
					DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWNICKN(=%s)\n",pAd->nickn);
				}
			}
			break;
		case SIOCGIWNICKN:
			{
				struct iw_point *erq = &wrq->u.data;

				erq->length = strlen(pAd->nickn);
				if(copy_to_user(erq->pointer, pAd->nickn, erq->length))
					Status = -EFAULT;
			}
			break;
		case SIOCSIWAP:
			{
				PUCHAR ptr=NULL;
				ptr = (PUCHAR) &(wrq->u.ap_addr.sa_data);

				//memmove(&Bssid, ptr+2, sizeof(NDIS_802_11_MAC_ADDRESS));
				memmove(&Bssid, ptr, sizeof(NDIS_802_11_MAC_ADDRESS));

				if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)
				{
					DBGPRINT(RT_DEBUG_TRACE, "INFO:: MLME busy, ignore this ioctl SIOCSIWAP !!!");
					Status = -EBUSY;
				}
				else if((memcmp(NULL_AP, &Bssid, sizeof(NDIS_802_11_MAC_ADDRESS)) == 0) ||
					(memcmp(BROADCAST_AP, &Bssid, sizeof(NDIS_802_11_MAC_ADDRESS)) == 0))
				{ //Any
					// tell CNTL state machine to call NdisMSetInformationComplete() after completing
					// this request, because this request is initiated by NDIS.
					memset(&Ssid, 0x00, sizeof(NDIS_802_11_SSID));
					pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE; 
					MlmeEnqueue(&pAd->Mlme.Queue, 
								MLME_CNTL_STATE_MACHINE, 
								OID_802_11_SSID, 
								sizeof(NDIS_802_11_SSID),
								(VOID *)&Ssid
								);
					memcpy(pAd->Ssid, Ssid.Ssid, Ssid.SsidLength);
					pAd->SsidLen = Ssid.SsidLength;				            
 					pAd->PortCfg.AutoReconnect = TRUE;
					StateMachineTouched = TRUE;
					DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWAP (Len=%d,Ssid=%s)\n", Ssid.SsidLength, Ssid.Ssid);	
				}
				else 
				{
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTPRIV_IOCTL_SET_BSSID\n");
					pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE; 
				
					MlmeEnqueue(&pAd->Mlme.Queue, 
								MLME_CNTL_STATE_MACHINE, 
								RT_OID_802_11_BSSID, 
								sizeof(NDIS_802_11_MAC_ADDRESS), 
								&Bssid);
					
					StateMachineTouched = TRUE;
					DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWAP Bssid[=%02x:%02x:%02x:%02x:%02x:%02x]\n",
								Bssid[0], Bssid[1], Bssid[2], Bssid[3], Bssid[4], Bssid[5]);
				}
			}
			break;

		case SIOCGIWFREQ:	//Get frequency / channel 	
			if (INFRA_ON(pAd) || ADHOC_ON(pAd))
				wrq->u.freq.m = pAd->PortCfg.Channel;
			else
				wrq->u.freq.m = pAd->PortCfg.IbssConfig.Channel;
			wrq->u.freq.e = 0;
			wrq->u.freq.i = 0;
			break;

		case SIOCSIWFREQ:  //Set frequency / channel
			{
				struct iw_freq *frq = &wrq->u.freq;
				int chan = -1;

				if((frq->e == 0) && (frq->m <= 1000))
				{
					/* Setting by channel number */
					chan = frq->m;
				}
				else
				{
					/* Setting by frequency - search the table */
					int mult = 1;
					int i;

					for(i = 0; i < (6 - frq->e); i++)
						mult *= 10;

					for(i = 0; i < NUM_CHANNELS; i++)
						if(frq->m == (channel_frequency[i] * mult))
							chan = i+1;
				}

				if(!ChannelSanity(pAd, chan))
				{
					Status = -EINVAL;
					break;
				}
				pAd->PortCfg.IbssConfig.Channel = chan;
				DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWFREQ (Channel=%d)\n", pAd->PortCfg.IbssConfig.Channel);
			}
			break;
		case SIOCGIWRATE:  //Get Bit-Rate
			if (pAd->PortCfg.TxRate == RATE_1)
				wrq->u.bitrate.value = 1000000;
			else if (pAd->PortCfg.TxRate == RATE_2)
				wrq->u.bitrate.value = 2000000;
			else if (pAd->PortCfg.TxRate == RATE_5_5)
				wrq->u.bitrate.value = 5500000;
			else
				wrq->u.bitrate.value = 11000000;
			wrq->u.bitrate.fixed = (pAd->PortCfg.TxRate != RATE_11);
			wrq->u.bitrate.disabled = 0;
			break;

		case SIOCSIWRATE: //Set Bit-Rate
			switch (wrq->u.bitrate.value)
			{
				case -1: 
					pAd->PortCfg.TxRate = RATE_11; 
					break; /* auto rate */ 
				case 1000000: 
					pAd->PortCfg.TxRate = RATE_1; 
			break;
				case 2000000:
					pAd->PortCfg.TxRate = RATE_2; 
					break;
				case 5500000:
					pAd->PortCfg.TxRate = RATE_5_5; 
					break;
				case 11000000:
					pAd->PortCfg.TxRate = RATE_11; 
					break; 
				default:
					Status = -EINVAL;
					break;
			}
			DBGPRINT(RT_DEBUG_TRACE, "INFO::SIOCSIWRATE (rates=%d)\n", pAd->PortCfg.TxRate);
			break;
		case SIOCSIWENCODE:  //Set encryption stuff
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
					/* take the old default key if index is invalid */
				if((index < 0) || (index >= NR_WEP_KEYS))					
					index = pAd->PortCfg.DefaultKeyId;  	// Default key for tx (shared key)
				if(wrq->u.encoding.pointer)
				{
					int len = wrq->u.encoding.length;
					
					if(len > WEP_LARGE_KEY_LEN)
						len = WEP_LARGE_KEY_LEN;

					memset(pAd->PortCfg.SharedKey[index].Key, 0, MAX_LEN_OF_KEY);
					if(copy_from_user(pAd->PortCfg.SharedKey[index].Key, 
						wrq->u.encoding.pointer, len)) 
					{						
						pAd->PortCfg.SharedKey[index].KeyLen = 0;
						Status = -EFAULT;
						break;
					}
					else
					{
						pAd->PortCfg.SharedKey[index].KeyLen = 
							len <= WEP_SMALL_KEY_LEN ?
							WEP_SMALL_KEY_LEN : WEP_LARGE_KEY_LEN;
					}
				}
				pAd->PortCfg.DefaultKeyId = (UCHAR) index;
				if (wrq->u.encoding.flags & IW_ENCODE_DISABLED)
					pAd->PortCfg.WepStatus = Ndis802_11WEPDisabled;
				else
					pAd->PortCfg.WepStatus = Ndis802_11WEPEnabled;

				if (wrq->u.encoding.flags & IW_ENCODE_RESTRICTED)
					pAd->PortCfg.AuthMode = Ndis802_11AuthModeShared;
				if (wrq->u.encoding.flags & IW_ENCODE_OPEN)
					pAd->PortCfg.AuthMode = Ndis802_11AuthModeOpen;

				if(pAd->PortCfg.WepStatus == Ndis802_11WEPDisabled)		        
					pAd->PortCfg.AuthMode = Ndis802_11AuthModeOpen;
			}
			break;
		case SIOCGIWENCODE: // Get the WEP keys and mode
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				if ((index < 0) || (index >= NR_WEP_KEYS))
					index = pAd->PortCfg.DefaultKeyId;  	// Default key for tx (shared key)

				if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeOpen)
					wrq->u.encoding.flags = IW_ENCODE_OPEN;
				else if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeShared)
					wrq->u.encoding.flags = IW_ENCODE_RESTRICTED;

				if(pAd->PortCfg.WepStatus)
					wrq->u.encoding.flags |= IW_ENCODE_DISABLED;
				if(wrq->u.encoding.pointer)
				{
					wrq->u.encoding.length = pAd->PortCfg.SharedKey[index].KeyLen;
					if (copy_to_user(wrq->u.encoding.pointer, 
						pAd->PortCfg.SharedKey[index].Key,
						pAd->PortCfg.SharedKey[index].KeyLen))
					{
						Status = -EFAULT;
						break;
					}
					wrq->u.encoding.flags |= (index + 1);
				}
			}
			break;
		case SIOCGIWMODE: // Get operation mode 
			if (INFRA_ON(pAd))
				wrq->u.mode = IW_MODE_INFRA;
			else if(ADHOC_ON(pAd))
				wrq->u.mode = IW_MODE_ADHOC;
			else
			{
				if (pAd->PortCfg.BssType == BSS_INDEP)
					wrq->u.mode = IW_MODE_ADHOC;
				else if (pAd->PortCfg.BssType == BSS_INFRA)
					wrq->u.mode = IW_MODE_INFRA;
				else
					wrq->u.mode = IW_MODE_AUTO;
			}				
			break;
			
		case SIOCSIWMODE:  //Set operation mode
			if(wrq->u.mode == IW_MODE_ADHOC)
			{
				pAd->PortCfg.BssType = BSS_INDEP;
				DBGPRINT(RT_DEBUG_TRACE, "INFO::RTPRIV_IOCTL_SET_INFRASTRUCTURE_MODE (AD-HOC)\n");
			}
			else if(wrq->u.mode == IW_MODE_INFRA)
			{
				pAd->PortCfg.BssType = BSS_INFRA;
				DBGPRINT(RT_DEBUG_TRACE, "INFO::RTPRIV_IOCTL_SET_INFRASTRUCTURE_MODE (INFRA)\n");
			}
			else
				Status = -EINVAL;
			break;
		case SIOCGIWPOWER:
		case SIOCSIWPOWER:
		case SIOCGIWTXPOW:
		case SIOCSIWTXPOW:
			Status = -EOPNOTSUPP;
			break;
		case SIOCDEVPRIVATE:
			switch(subcmd)
			{
				case RTLINUX_SET_OID_802_11_BSSID_LIST_SCAN:
					Now = jiffies;
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_BSSID_LIST_SCAN \n");
					if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)
					{
						MlmeRestartStateMachine(pAd);
		               	DBGPRINT(RT_DEBUG_TRACE, "INFO::!!! MLME busy, reset MLME state machine !!!");
					}
						
						// tell CNTL state machine to call NdisMSetInformationComplete() after completing
						// this request, because this request is initiated by NDIS.
		            pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE;
						// Reset Missed scan number
						pAd->PortCfg.IgnoredScanNumber = 0;
						pAd->PortCfg.LastScanTime = Now;
                
						MlmeEnqueue(&pAd->Mlme.Queue, 
									MLME_CNTL_STATE_MACHINE, 
									OID_802_11_BSSID_LIST_SCAN, 
									0, 
									NULL);

		            Status = NDIS_STATUS_SUCCESS;
						StateMachineTouched = TRUE;
					break;

				case RTLINUX_GET_OID_802_11_BSSID_LIST:
					BssLen = pAd->PortCfg.BssTab.BssNr * sizeof(NDIS_WLAN_BSSID) + sizeof(ULONG);
					pBssidList = (PNDIS_802_11_BSSID_LIST) kmalloc(BssLen, GFP_KERNEL);
					if(pBssidList == NULL)
					{
		                Status = -ENOMEM;
		                break;
		            }
		            memset(pBssidList, 0, BssLen);
		            
					pBssidList->NumberOfItems = pAd->PortCfg.BssTab.BssNr;
										
					for (i = 0; i < pAd->PortCfg.BssTab.BssNr; i++) 
					{
						pBssidList->Bssid[i].Length = sizeof(NDIS_WLAN_BSSID);
		                memmove(&pBssidList->Bssid[i].MacAddress, &pAd->PortCfg.BssTab.BssEntry[i].Bssid, MAC_ADDR_LEN);
		                pBssidList->Bssid[i].Ssid.SsidLength = pAd->PortCfg.BssTab.BssEntry[i].SsidLen;
		                memmove(pBssidList->Bssid[i].Ssid.Ssid, pAd->PortCfg.BssTab.BssEntry[i].Ssid, pAd->PortCfg.BssTab.BssEntry[i].SsidLen);
		                pBssidList->Bssid[i].Privacy = pAd->PortCfg.BssTab.BssEntry[i].Privacy;
		                pBssidList->Bssid[i].Rssi = pAd->PortCfg.BssTab.BssEntry[i].Rssi - RSSI_TO_DBM_OFFSET;
		                pBssidList->Bssid[i].NetworkTypeInUse = Ndis802_11DS;
		                pBssidList->Bssid[i].Configuration.Length = sizeof(NDIS_802_11_CONFIGURATION);
		                pBssidList->Bssid[i].Configuration.BeaconPeriod = pAd->PortCfg.BssTab.BssEntry[i].BeaconPeriod;
		                pBssidList->Bssid[i].Configuration.ATIMWindow = pAd->PortCfg.BssTab.BssEntry[i].AtimWin;
						if (pAd->PortCfg.BssTab.BssEntry[i].Channel == 14)
							pBssidList->Bssid[i].Configuration.DSConfig = 2484000; 
						else
							pBssidList->Bssid[i].Configuration.DSConfig = 2412000 + (pAd->PortCfg.BssTab.BssEntry[i].Channel - 1) * 5000; 
				
		                if(pAd->PortCfg.BssTab.BssEntry[i].BssType == BSS_INFRA) 
		                    pBssidList->Bssid[i].InfrastructureMode = Ndis802_11Infrastructure;
		                else 
		                    pBssidList->Bssid[i].InfrastructureMode = Ndis802_11IBSS;

		                memmove(pBssidList->Bssid[i].SupportedRates, pAd->PortCfg.BssTab.BssEntry[i].Rates, pAd->PortCfg.BssTab.BssEntry[i].RatesLen);

		                DBGPRINT(RT_DEBUG_INFO,"TRACE::RTLINUX_GET_OID_802_11_BSSID_LIST (BSS[%d], MAC = %02x:%02x:%02x:%02x:%02x:%02x, Ssid:%s)\n",
							i,
							pBssidList->Bssid[i].MacAddress[0],
                            pBssidList->Bssid[i].MacAddress[1],
                            pBssidList->Bssid[i].MacAddress[2],
                            pBssidList->Bssid[i].MacAddress[3],
                            pBssidList->Bssid[i].MacAddress[4],
                            pBssidList->Bssid[i].MacAddress[5],
                            pBssidList->Bssid[i].Ssid.Ssid);
					}
					
					wrq->u.data.length = BssLen;
					copy_to_user(wrq->u.data.pointer, pBssidList, wrq->u.data.length);
					break;

				case RTLINUX_GET_MEDIA_CONNECT_STATUS:
		            DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_MEDIA_CONNECT_STATUS \n");
					wrq->u.data.length=sizeof(NDIS_MEDIA_STATE);
					copy_to_user(wrq->u.data.pointer, &pAd->MediaState, wrq->u.data.length);
					break;

				case RTLINUX_GET_OID_802_11_BSSID:
					if (INFRA_ON(pAd) || ADHOC_ON(pAd)) 
		            {
		                COPY_MAC_ADDR(&Bssid, &pAd->PortCfg.Bssid);
		                DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_BSSID(=%02x:%02x:%02x:%02x:%02x:%02x)\n",
		                         Bssid[0],Bssid[1],Bssid[2],Bssid[3],Bssid[4],Bssid[5]);
		                wrq->u.data.length = MAC_ADDR_LEN;
		                copy_to_user(wrq->u.data.pointer, &Bssid, wrq->u.data.length);
		            } 
		            else 
		            {
		                DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_BSSID(=EMPTY)\n");
		                Status = -ENOTCONN;
		            }
		            break;

				case RTLINUX_SET_OID_802_11_BSSID:
					if (wrq->u.data.length != sizeof(NDIS_802_11_MAC_ADDRESS))
			            Status = -EINVAL;
		            else if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)
		            {
		                DBGPRINT(RT_DEBUG_TRACE, "INFO:: MLME busy, ignore this SET RTLINUX_SET_OID_802_11_BSSID !!!");
			            Status = -EBUSY;
		            }
			        else 
			        {
						DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_BSSID\n");
						pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE; 
		            
			            MlmeEnqueue(&pAd->Mlme.Queue, 
			                        MLME_CNTL_STATE_MACHINE, 
			                        RT_OID_802_11_BSSID, 
			                        wrq->u.data.length, 
			                        wrq->u.data.pointer);
			            
			            StateMachineTouched = TRUE;
			        }
		            break;

				case RTLINUX_GET_OID_802_11_SSID:
					Ssid.SsidLength = pAd->PortCfg.SsidLen;
					NdisZeroMemory(Ssid.Ssid, MAX_LEN_OF_SSID);
        		    memmove(Ssid.Ssid, pAd->PortCfg.Ssid, Ssid.SsidLength);
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_SSID (Len=%d, ssid=%s)\n", Ssid.SsidLength,Ssid.Ssid);

					wrq->u.data.length=sizeof(NDIS_802_11_SSID);
					copy_to_user(wrq->u.data.pointer, &Ssid, wrq->u.data.length);
					break;

				case RTLINUX_GET_OID_802_11_CONFIGURATION:
					Configuration.Length = sizeof(NDIS_802_11_CONFIGURATION);
					Configuration.BeaconPeriod = pAd->PortCfg.BeaconPeriod;
					Configuration.ATIMWindow = pAd->PortCfg.AtimWin;
					if (pAd->PortCfg.Channel == 14)
						Configuration.DSConfig = 2484000;
					else
						Configuration.DSConfig = 2412000 + (pAd->PortCfg.Channel - 1) * 5000;
					
					DBGPRINT(RT_DEBUG_TRACE, "INFO::OID_802_11_CONFIGURATION(BeaconPeriod=%d,AtimW=%d,Channel=%d) \n",
						Configuration.BeaconPeriod, Configuration.ATIMWindow, pAd->PortCfg.Channel);
					wrq->u.data.length = sizeof(NDIS_802_11_CONFIGURATION);
					copy_to_user(wrq->u.data.pointer, &Configuration, wrq->u.data.length);
					break;
				
				case RTLINUX_GET_RT_OID_802_11_QUERY_LINK_STATUS:
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_RT_OID_802_11_QUERY_LINK_STATUS\n");
					if (pAd->PortCfg.TxRate == RATE_1)
		                LinkStatus.CurrTxRate = 2;  // unit: 500 kbps
		            else if (pAd->PortCfg.TxRate == RATE_2)
		                LinkStatus.CurrTxRate = 4;  // unit: 500 kbps
		            else if (pAd->PortCfg.TxRate == RATE_5_5)
		                LinkStatus.CurrTxRate = 11; // unit: 500 kbps
		            else
		                LinkStatus.CurrTxRate = 22; // unit: 500 kbps
		            LinkStatus.ChannelQuality = pAd->Mlme.RoamCqi;
		            LinkStatus.RxByteCount = pAd->RalinkCounters.ReceivedByteCount;
		            LinkStatus.TxByteCount = pAd->RalinkCounters.TransmittedByteCount;
							
					wrq->u.data.length = sizeof(RT_802_11_LINK_STATUS);
					copy_to_user(wrq->u.data.pointer, &LinkStatus, wrq->u.data.length);
					break;

				case RTLINUX_GET_OID_802_11_RSSI:
					Rssi = pAd->PortCfg.LastRssi - RSSI_TO_DBM_OFFSET;
					wrq->u.data.length = sizeof(NDIS_802_11_RSSI);
					copy_to_user(wrq->u.data.pointer, &Rssi, wrq->u.data.length);
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_RSSI(=%d)\n", Rssi);
					break;

				case RTLINUX_GET_OID_802_11_INFRASTRUCTURE_MODE:
					if (ADHOC_ON(pAd)) 
		                ulInfo = Ndis802_11IBSS;
		            else if (INFRA_ON(pAd)) 
		                ulInfo = Ndis802_11Infrastructure;
		            else 
		                ulInfo = Ndis802_11AutoUnknown;
										
					wrq->u.data.length = sizeof(NDIS_802_11_NETWORK_INFRASTRUCTURE);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);

					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_INFRASTRUCTURE_MODE(=%d)\n", ulInfo);
					break;

				case RTLINUX_SET_OID_802_11_INFRASTRUCTURE_MODE:
					if (wrq->u.data.length != sizeof(NDIS_802_11_NETWORK_INFRASTRUCTURE))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&BssType, wrq->u.data.pointer, wrq->u.data.length);		                
		                if (BssType == Ndis802_11IBSS) 
		                {
							if (pAd->PortCfg.BssType != BSS_INDEP)
							{
								// Config has changed
								pAd->bConfigChanged = TRUE;
							}
		                    pAd->PortCfg.BssType = BSS_INDEP;
		                    DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_INFRASTRUCTURE_MODE (AD-HOC)\n");
		                } 
		                else if (BssType == Ndis802_11Infrastructure) 
		                {
							if (pAd->PortCfg.BssType != BSS_INFRA)
							{
								// Config has changed
								pAd->bConfigChanged = TRUE;
							}
		                    pAd->PortCfg.BssType = BSS_INFRA;
		                    DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_INFRASTRUCTURE_MODE (INFRA)\n");
		                } 
		                else 
		                {
		                    //if(BssType != Ndis802_11AutoUnknown)  // not supported???????????????
		                    Status = -EINVAL;
		                    DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_INFRASTRUCTURE_MODE (unknown)\n");
		                }
		            }
					break;

				case RTLINUX_GET_OID_802_11_TX_ANTENNA_SELECTED:
					if (pAd->PortCfg.CurrentTxAntenna == 0xff)
						ulInfo = 0xffffffff;
					else
						ulInfo = pAd->PortCfg.CurrentTxAntenna;

					wrq->u.data.length = sizeof(ulInfo);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);

					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_TX_ANTENNA_SELECTED(=%d)\n", ulInfo);
					break;

				case RTLINUX_SET_OID_802_11_TX_ANTENNA_SELECTED:
					if (wrq->u.data.length != sizeof(NDIS_802_11_ANTENNA)) 
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&Antenna, wrq->u.data.pointer, wrq->u.data.length);
		                if (Antenna == 0xFFFFFFFF) 
						{// Diversity
							pAd->PortCfg.CurrentTxAntenna = (UCHAR)Antenna;
							RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Tx_Configure, &TxValue);
                			TxValue = (TxValue & 0xFC) | 0x01;
                			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Tx_Configure, TxValue);
						} 
						else if (Antenna >= pAd->PortCfg.NumberOfAntenna)
							Status = -EINVAL;
						else
						{
							pAd->PortCfg.CurrentTxAntenna = (UCHAR)Antenna;

							RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Tx_Configure, &TxValue);
							if(Antenna == 0)
							{// Antenna A
								TxValue = (TxValue & 0xFC) | 0x00;
							}
							else if(Antenna == 1)
							{// Antenna B
								TxValue = (TxValue & 0xFC) | 0x02;
							}
                			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Tx_Configure, TxValue);
						}
		            }
		            DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_SET_OID_802_11_TX_ANTENNA_SELECTED (=%d) \n",Antenna);
		            break;

				case RTLINUX_GET_OID_802_11_RX_ANTENNA_SELECTED:
					if (pAd->PortCfg.CurrentRxAntenna == 0xff)
						ulInfo = 0xffffffff;
					else
						ulInfo = pAd->PortCfg.CurrentRxAntenna;

					wrq->u.data.length = sizeof(ulInfo);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);

					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_RX_ANTENNA_SELECTED(=%d)\n", ulInfo);
            		break;

            	case RTLINUX_SET_OID_802_11_RX_ANTENNA_SELECTED:
					if (wrq->u.data.length != sizeof(NDIS_802_11_ANTENNA))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&Antenna, wrq->u.data.pointer, wrq->u.data.length);
		                if(Antenna == 0xFFFFFFFF) 
						{// Diversity
							pAd->PortCfg.CurrentRxAntenna = (UCHAR)Antenna;
                			RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Rx_Configure, &RxValue);
                			RxValue = (RxValue & 0xF9) | 0x02;
                			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Rx_Configure, RxValue);
						} 
						else if(Antenna >= pAd->PortCfg.NumberOfAntenna)
							Status = -EINVAL;
						else
						{
							pAd->PortCfg.CurrentRxAntenna = (UCHAR)Antenna;
                    
							RTMP_BBP_IO_READ32_BY_REG_ID(pAd, BBP_Rx_Configure, &RxValue);
							if(Antenna == 0)
							{// Antenna A
								RxValue = (RxValue & 0xF9) | 0x00;
							}
							else if(Antenna == 1)
							{// Antenna B
								RxValue = (RxValue & 0xF9) | 0x04;
							}
                			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, BBP_Rx_Configure, RxValue);
						}
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_RX_ANTENNA_SELECTED (=%d)\n",Antenna);
		            break;

				case RTLINUX_GET_OID_802_11_DESIRED_RATES:
					wrq->u.data.length = sizeof(NDIS_802_11_RATES);
					copy_to_user(wrq->u.data.pointer, &pAd->PortCfg.DesiredRates, wrq->u.data.length);
		            DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_OID_802_11_DESIRED_RATES (%d,%d,%d,%d,%d,%d,%d,%d)\n",
		                Rates[0],Rates[1],Rates[2],Rates[3],Rates[4],Rates[5],Rates[6],Rates[7]);
					break;

				case RTLINUX_SET_OID_802_11_DESIRED_RATES:
					if (wrq->u.data.length != sizeof(NDIS_802_11_RATES)) 
		                Status = -EINVAL;
		            else
		            {
		            	copy_from_user(&pAd->PortCfg.DesiredRates, wrq->u.data.pointer, wrq->u.data.length);
		                // Changing DesiredRate may affect the MAX TX rate we used to TX frames out
		                MlmeUpdateTxRates(pAd);
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_DESIRED_RATES (%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x)\n",
		                pAd->PortCfg.DesiredRates[0],pAd->PortCfg.DesiredRates[1],
		                pAd->PortCfg.DesiredRates[2],pAd->PortCfg.DesiredRates[3],
		                pAd->PortCfg.DesiredRates[4],pAd->PortCfg.DesiredRates[5],
		                pAd->PortCfg.DesiredRates[6],pAd->PortCfg.DesiredRates[7] );
					break;

				case RTLINUX_GET_OID_802_11_QUERY_PREAMBLE:
					ulInfo = pAd->PortCfg.TxPreamble;
					wrq->u.data.length = sizeof(RT_802_11_PREAMBLE);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
            		DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_QUERY_PREAMBLE(=%d)\n", ulInfo);
            		break;

				case RTLINUX_SET_OID_802_11_SET_PREAMBLE:
					if (wrq->u.data.length != sizeof(ULONG))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&Preamble, wrq->u.data.pointer, wrq->u.data.length);
		                if (Preamble == Rt802_11PreambleShort)
		                {
		                    pAd->PortCfg.WindowsTxPreamble = Preamble;
		                    MlmeSetTxPreamble(pAd, Rt802_11PreambleShort);
		                }
		                else if ((Preamble == Rt802_11PreambleLong) || (Preamble == Rt802_11PreambleAuto))
		                {
		                    // if user wants AUTO, initialize to LONG here, then change according to AP's
		                    // capability upon association.
		                    pAd->PortCfg.WindowsTxPreamble = Preamble;
		                    MlmeSetTxPreamble(pAd, Rt802_11PreambleLong);
		                }
		                else
		                    Status = -EINVAL;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_SET_PREAMBLE (=%d)\n", Preamble);
					break;

				case RTLINUX_GET_OID_802_11_TX_POWER_LEVEL:
					ulInfo = (ULONG)pAd->PortCfg.TxPower;
					wrq->u.data.length = sizeof(NDIS_802_11_TX_POWER_LEVEL);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
            		DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_TX_POWER_LEVEL(=%d)\n", ulInfo);
            		break;

            	case RTLINUX_SET_OID_802_11_TX_POWER_LEVEL:
            		if (wrq->u.data.length != sizeof(NDIS_802_11_TX_POWER_LEVEL)) 
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&TxPowerLevel, wrq->u.data.pointer, wrq->u.data.length);
		                if (TxPowerLevel > MAX_TX_POWER_LEVEL)
		                    Status = -EINVAL;
		                else
		                    pAd->PortCfg.TxPower = (UCHAR)TxPowerLevel;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_TX_POWER_LEVEL (=%d) \n",TxPowerLevel);
		            break;

				case RTLINUX_GET_OID_802_11_RTS_THRESHOLD:
					ulInfo = pAd->PortCfg.RtsThreshold;
					wrq->u.data.length = sizeof(NDIS_802_11_RTS_THRESHOLD);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_RTS_THRESHOLD(=%d)\n", ulInfo);
					break;

				case RTLINUX_SET_OID_802_11_RTS_THRESHOLD:
					if (wrq->u.data.length != sizeof(NDIS_802_11_RTS_THRESHOLD))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&RtsThresh, wrq->u.data.pointer, wrq->u.data.length);
						if (RtsThresh > MAX_RTS_THRESHOLD)
		                    Status = -EINVAL;
		                else 
		                    pAd->PortCfg.RtsThreshold = (USHORT)RtsThresh;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_RTS_THRESHOLD (=%d)\n",RtsThresh);
		            break;

            	case RTLINUX_GET_OID_802_11_FRAGMENTATION_THRESHOLD:
            		ulInfo = pAd->PortCfg.FragmentThreshold;
		            if (pAd->PortCfg.bFragmentZeroDisable == TRUE)
		            	ulInfo = 0;
		            wrq->u.data.length = sizeof(NDIS_802_11_FRAGMENTATION_THRESHOLD);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_FRAGMENTATION_THRESHOLD(=%d)\n", ulInfo);
		            break;

				case RTLINUX_SET_OID_802_11_FRAGMENTATION_THRESHOLD:
					if (wrq->u.data.length != sizeof(NDIS_802_11_FRAGMENTATION_THRESHOLD))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&FragThresh, wrq->u.data.pointer, wrq->u.data.length);
		                pAd->PortCfg.bFragmentZeroDisable = FALSE;
		                if (FragThresh > MAX_FRAG_THRESHOLD || FragThresh < MIN_FRAG_THRESHOLD)
		                	if (FragThresh == 0)
		                	{
		                		pAd->PortCfg.FragmentThreshold = MAX_FRAG_THRESHOLD;
		                		pAd->PortCfg.bFragmentZeroDisable = TRUE;
		                	}
		                	else
		                    	Status = -EINVAL;
		                else
		                    pAd->PortCfg.FragmentThreshold = (USHORT)FragThresh;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_FRAGMENTATION_THRESHOLD (=%d) \n",FragThresh);
		            break;

				case RTLINUX_GET_OID_802_11_AUTHENTICATION_MODE:
					if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeOpen) 
		                ulInfo = Ndis802_11AuthModeOpen;
		            else if (pAd->PortCfg.AuthMode == Ndis802_11AuthModeShared) 
		                ulInfo = Ndis802_11AuthModeShared;
		            else 
		                ulInfo = Ndis802_11AuthModeAutoSwitch;
		            wrq->u.data.length = sizeof(NDIS_802_11_AUTHENTICATION_MODE);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);

		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_AUTHENTICATION_MODE(=%d)\n", ulInfo);
		            break;

				case RTLINUX_SET_OID_802_11_AUTHENTICATION_MODE:
					if (wrq->u.data.length != sizeof(NDIS_802_11_AUTHENTICATION_MODE)) 
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&AuthMode, wrq->u.data.pointer, wrq->u.data.length);
		                if (pAd->PortCfg.AuthMode > Ndis802_11AuthModeMax)
		                    Status = -EINVAL;
		            else 
		            {
							if (pAd->PortCfg.AuthMode != AuthMode)
		            		{
								// Config has changed
								pAd->bConfigChanged = TRUE;
		            		}
							pAd->PortCfg.AuthMode = AuthMode;			
						}
		            }
					pAd->PortCfg.PortSecured = WPA_802_1X_PORT_NOT_SECURED;
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_AUTHENTICATION_MODE (=%d) \n",pAd->PortCfg.AuthMode);
		            break;
					
				case RTLINUX_GET_OID_802_11_WEP_STATUS:
					wrq->u.data.length = sizeof(NDIS_802_11_WEP_STATUS);
					copy_to_user(wrq->u.data.pointer, &pAd->PortCfg.WepStatus, wrq->u.data.length);
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_WEP_STATUS(=%d)\n", pAd->PortCfg.WepStatus);
		            break;

				case RTLINUX_SET_OID_802_11_WEP_STATUS:
					if (wrq->u.data.length != sizeof(NDIS_802_11_WEP_STATUS))
						Status = -EINVAL;
					else 
					{
						copy_from_user(&WepStatus, wrq->u.data.pointer, wrq->u.data.length);
						if (pAd->PortCfg.WepStatus != WepStatus)
						{
							// Config has changed
							pAd->bConfigChanged = TRUE;
						}

						if (WepStatus == Ndis802_11WEPEnabled)
							pAd->PortCfg.WepStatus = Ndis802_11WEPEnabled;
						else
							pAd->PortCfg.WepStatus = Ndis802_11WEPDisabled;
					}
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_WEP_STATUS (=%d)\n",WepStatus);
					break;

				case RTLINUX_SET_OID_802_11_REMOVE_WEP:
					if (wrq->u.data.length != sizeof(NDIS_802_11_KEY_INDEX))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&KeyIdx, wrq->u.data.pointer, wrq->u.data.length);
						if (KeyIdx & 0x80000000)
						{
                			// Should never set default bit when remove key
							Status = -EINVAL;
						}
						else
		                {
		                	KeyIdx = KeyIdx & 0x0fffffff;
		                    if (KeyIdx > 4)
		                        Status = -EINVAL;
		                    else
		                        pAd->PortCfg.SharedKey[KeyIdx].KeyLen = 0;
		                }
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_REMOVE_WEP (id=%d)\n",KeyIdx);
		            break;

				case RTLINUX_SET_OID_802_11_ADD_WEP:
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_ADD_WEP\n");
					copy_from_user(&WepKey, wrq->u.data.pointer, wrq->u.data.length);
					KeyIdx = WepKey.KeyIndex & 0x0fffffff;
           			
					// it is a shared key
					if (KeyIdx > 4)
						Status = -EINVAL;
					else 
					{
						pAd->PortCfg.SharedKey[KeyIdx].KeyLen = (UCHAR) WepKey.KeyLength;
						NdisMoveMemory(pAd->PortCfg.SharedKey[KeyIdx].Key, &WepKey.KeyMaterial, WepKey.KeyLength);
            			if (WepKey.KeyIndex & 0x80000000)
           				{
             				// Default key for tx (shared key)
               				pAd->PortCfg.DefaultKeyId = (UCHAR) KeyIdx;
            			}
				   }
            
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_ADD_WEP (id=%d, Len=%d-byte)\n", 
						KeyIdx, WepKey.KeyLength);
					break;

				case RTLINUX_SET_OID_802_11_CONFIGURATION:
					if (wrq->u.data.length != sizeof(NDIS_802_11_CONFIGURATION))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&Config, wrq->u.data.pointer, wrq->u.data.length);
		                pAd->PortCfg.IbssConfig.BeaconPeriod = (USHORT) Config.BeaconPeriod;
		                pAd->PortCfg.IbssConfig.AtimWin = (USHORT) Config.ATIMWindow;
		                if (Config.DSConfig == 2484000)
                			pAd->PortCfg.IbssConfig.Channel = 14;
						else if (Config.DSConfig >= 2412000 && Config.DSConfig < 2484000) 
                			pAd->PortCfg.IbssConfig.Channel = (UCHAR)((Config.DSConfig - 2412000) / 5000) + 1;
						else 
		                    Status = -EINVAL;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_CONFIGURATION (BeacnPeriod=%d,AtimW=%d,Ch=%d)\n",
		                Config.BeaconPeriod, Config.ATIMWindow, pAd->PortCfg.IbssConfig.Channel);
		            break;

				case RTLINUX_GET_OID_802_11_POWER_MODE:
					ulInfo = pAd->PortCfg.WindowsPowerMode;
					wrq->u.data.length = sizeof(NDIS_802_11_POWER_MODE);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_POWER_MODE(=%d)\n", ulInfo);
		            break;

				case RTLINUX_SET_OID_802_11_POWER_MODE:
					if (wrq->u.data.length != sizeof(NDIS_802_11_POWER_MODE))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&PowerMode, wrq->u.data.pointer, wrq->u.data.length);

		                // save user's policy here, but not change PortCfg.Psm immediately
		                if (PowerMode == Ndis802_11PowerModeCAM) 
		                {
		                    // clear PSM bit immediately
		                    MlmeSetPsmBit(pAd, PWR_ACTIVE);
		                    
		                    pAd->PortCfg.RecvDtim = TRUE;
							if (pAd->PortCfg.WindowsACCAMEnable == FALSE)
			                    pAd->PortCfg.WindowsPowerMode = PowerMode;
			                pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
		                } 
		                else if (PowerMode == Ndis802_11PowerModeMAX_PSP) 
		                {
		                    // do NOT turn on PSM bit here, wait until MlmeCheckForPsmChange()
		                    // to exclude certain situations.
		                    //     MlmeSetPsmBit(pAd, PWR_SAVE);
							if (pAd->PortCfg.WindowsACCAMEnable == FALSE)
			                    pAd->PortCfg.WindowsPowerMode = PowerMode;
			                pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
		                    pAd->PortCfg.RecvDtim = FALSE;
		                    pAd->PortCfg.DefaultListenCount = 5;
		                } 
		                else if (PowerMode == Ndis802_11PowerModeFast_PSP) 
		                {
		                    // do NOT turn on PSM bit here, wait until MlmeCheckForPsmChange()
		                    // to exclude certain situations.
		                    //     MlmeSetPsmBit(pAd, PWR_SAVE);
		                    pAd->PortCfg.RecvDtim = TRUE;
							if (pAd->PortCfg.WindowsACCAMEnable == FALSE)
			                    pAd->PortCfg.WindowsPowerMode = PowerMode;
			                pAd->PortCfg.WindowsBatteryPowerMode = PowerMode;
		                    pAd->PortCfg.DefaultListenCount = 3;
		                } 
		                else
		                    Status = -EINVAL;
		            }
		            DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_POWER_MODE (=%d)\n",PowerMode);
		            break;

				case RTLINUX_SET_OID_802_11_SSID:
					if ((wrq->u.data.length > MAX_LEN_OF_SSID + sizeof(ULONG)) || (wrq->u.data.length < sizeof(ULONG)))
			            Status = -EINVAL;
		            else 
		            {
		                copy_from_user(&Ssid, wrq->u.data.pointer, wrq->u.data.length);
		                
						if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)
						{
							MlmeRestartStateMachine(pAd);
		                	DBGPRINT(RT_DEBUG_TRACE, "!!! MLME busy, reset MLME state machine !!!");
						}
		                                
		    	        if(Ssid.SsidLength > 32)
				            Status = -EINVAL;
			            else 
			            {
		                    // tell CNTL state machine to call NdisMSetInformationComplete() after completing
		                    // this request, because this request is initiated by NDIS.
		                    pAd->Mlme.CntlAux.CurrReqIsFromNdis = FALSE; 
		                    
				            MlmeEnqueue(&pAd->Mlme.Queue, 
				                        MLME_CNTL_STATE_MACHINE, 
				                        OID_802_11_SSID, 
				                        wrq->u.data.length, 
				                        wrq->u.data.pointer);
				            
				            Status = NDIS_STATUS_SUCCESS;
				            StateMachineTouched = TRUE;
			            }
		                DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_OID_802_11_SSID (Len=%d,Ssid=%s)\n", Ssid.SsidLength, Ssid.Ssid);
			        }
			        break;

				case RTLINUX_GET_OID_802_11_STATISTICS:
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_GET_OID_802_11_STATISTICS \n");
		            // Update FCS counters
		            RTMP_IO_READ32(pAd, CNT0, &FcsValue);
		            pAd->WlanCounters.FCSErrorCount.LowPart += FcsValue;
		            
		            Statistics.TransmittedFragmentCount = pAd->WlanCounters.TransmittedFragmentCount;
		            Statistics.MulticastTransmittedFrameCount = pAd->WlanCounters.MulticastTransmittedFrameCount;
		            Statistics.FailedCount = pAd->WlanCounters.FailedCount;
		            Statistics.RetryCount = pAd->WlanCounters.RetryCount;
		            Statistics.MultipleRetryCount = pAd->WlanCounters.MultipleRetryCount;
		            Statistics.RTSSuccessCount = pAd->WlanCounters.RTSSuccessCount;
		            Statistics.RTSFailureCount = pAd->WlanCounters.RTSFailureCount;
		            Statistics.ACKFailureCount = pAd->WlanCounters.ACKFailureCount;
		            Statistics.FrameDuplicateCount = pAd->WlanCounters.FrameDuplicateCount;
		            Statistics.ReceivedFragmentCount = pAd->WlanCounters.ReceivedFragmentCount;
		            Statistics.MulticastReceivedFrameCount = pAd->WlanCounters.MulticastReceivedFrameCount;
		            Statistics.FCSErrorCount = pAd->WlanCounters.FCSErrorCount;
		            wrq->u.data.length = sizeof(NDIS_802_11_STATISTICS);
					copy_to_user(wrq->u.data.pointer, &Statistics, wrq->u.data.length);					
		            break;

				case RTLINUX_GET_OID_GEN_RCV_OK:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_OID_GEN_RCV_OK \n");
					ulInfo = pAd->Counters.GoodReceives;
					wrq->u.data.length = sizeof(ulInfo);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
					break;

				case RTLINUX_GET_OID_GEN_RCV_NO_BUFFER:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_OID_GEN_RCV_NO_BUFFER \n");
					ulInfo = pAd->Counters.RxNoBuffer;
					wrq->u.data.length = sizeof(ulInfo);
					copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
					break;

				case RTLINUX_SET_RT_OID_802_11_RESET_COUNTERS:
					DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_RT_OID_802_11_RESET_COUNTERS!!!\n");
					memset(&pAd->WlanCounters, 0, sizeof(COUNTER_802_11));
        		    memset(&pAd->Counters, 0, sizeof(COUNTER_802_3));
    	        	memset(&pAd->RalinkCounters, 0 ,sizeof(COUNTER_RALINK));
		            memset(&pAd->Mlme.PrevWlanCounters, 0, sizeof(COUNTER_802_11));
					break;

				case RTLINUX_GET_OID_802_3_CURRENT_ADDRESS:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_OID_802_3_CURRENT_ADDRESS \n");
					wrq->u.data.length = ETH_LENGTH_OF_ADDRESS;
					copy_to_user(wrq->u.data.pointer, &pAd->CurrentAddress, wrq->u.data.length);
					break;

				case RTLINUX_GET_VERSION_INFO:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_VERSION_INFO \n");
					DriverVersionInfo.DriverMajorVersion=1;
					DriverVersionInfo.DriverMinorVersion=2;
					DriverVersionInfo.DriverSubVersion=0;
					DriverVersionInfo.DriverBuildYear=2003;
					DriverVersionInfo.DriverBuildMonth=12;
					DriverVersionInfo.DriverBuildDay=4;
					wrq->u.data.length = sizeof(RT_VERSION_INFO);
					copy_to_user(wrq->u.data.pointer, &DriverVersionInfo, wrq->u.data.length);
					break;

#if	RT2400_DBG
				case RTLINUX_GET_RT_OID_802_11_QUERY_HARDWARE_REGISTER:
		        	if (wrq->u.data.length != sizeof(RT_802_11_HARDWARE_REGISTER))
		                Status = -EINVAL;
		            else 
		            {
			        	copy_from_user(&HardwareRegister, wrq->u.data.pointer, sizeof(RT_802_11_HARDWARE_REGISTER));
		                if (HardwareRegister.HardwareType == 0)
		                {
		                    // MAC register read procedure
		                    RTMP_IO_READ32(pAd, HardwareRegister.Offset, &HardwareRegister.Data);
		                    wrq->u.data.length = sizeof(RT_802_11_HARDWARE_REGISTER);
							copy_to_user(wrq->u.data.pointer, &HardwareRegister, wrq->u.data.length);
		                }
		                else if (HardwareRegister.HardwareType == 1)
		                {
		                    // BBP register read procedure
		                    RTMP_BBP_IO_READ32_BY_REG_ID(pAd, HardwareRegister.Offset, &HardwareRegister.Data);
		                    wrq->u.data.length = sizeof(RT_802_11_HARDWARE_REGISTER);
		                    copy_to_user(wrq->u.data.pointer, &HardwareRegister, wrq->u.data.length);
		                }
		                else
		                {
			                wrq->u.data.length = 0;
		                    Status = -EINVAL;
		                }
		                DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_QUERY_HARDWARE_REGISTER (type=%d, offset=%d, data=%d)\n", HardwareRegister.HardwareType, HardwareRegister.Offset, HardwareRegister.Data);
		            }
		            break;

		        case RTLINUX_SET_RT_OID_802_11_SET_HARDWARE_REGISTER:
		        	if (wrq->u.data.length != sizeof(RT_802_11_HARDWARE_REGISTER))
		                Status = -EINVAL;
		            else 
		            {
		            	copy_from_user(&HardwareRegister, wrq->u.data.pointer, sizeof(RT_802_11_HARDWARE_REGISTER));
		                if (HardwareRegister.HardwareType == 0)
		                {
		                    // MAC register write procedure
		                    RTMP_IO_WRITE32(pAd, HardwareRegister.Offset, HardwareRegister.Data);
		                }
		                else if (HardwareRegister.HardwareType == 1)
		                {
		                    // BBP register write procedure
		                    RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, HardwareRegister.Offset, HardwareRegister.Data);
		                }
		                else if (HardwareRegister.HardwareType == 2)
		                {
		                    // RF register write procedure (haven't been verified, like bit order)
#if 0                    
		                    RfCsr.word = 0;
		                    RfCsr.field.Busy = 1;
		                    RfCsr.field.RFRegValue = pHardwareRegister->Data;
		                    RfCsr.field.NumberOfBits= 20;
		                    RfCsr.field.IFSelect= 0;
		                    RfCsr.field.PLL_LD= 0;
		                    RTMP_RF_IO_WRITE32(pAd, RfCsr.word);
#endif
		                    Status = -EOPNOTSUPP;
		                }
		                DBGPRINT(RT_DEBUG_TRACE, "INFO::RTLINUX_SET_RT_OID_802_11_SET_HARDWARE_REGISTER (type=%d, offset=%d, data=%d)\n", HardwareRegister.HardwareType, HardwareRegister.Offset, HardwareRegister.Data);
		            }
		        	break;
#endif
				case RTLINUX_SET_RT_OID_802_11_SET_RADIO:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_SET_RT_OID_802_11_SET_RADIO \n");
					copy_from_user(&RadioState, wrq->u.data.pointer, wrq->u.data.length);
					if (pAd->PortCfg.bRadio != RadioState)
					{
						pAd->PortCfg.bRadio = RadioState;
						if (RadioState == TRUE)
							MlmeRadioOn(pAd);
						else
							MlmeRadioOff(pAd);
					}
					break;
					
				case RTLINUX_GET_RT_OID_802_11_QUERY_RADIO:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_GET_RT_OID_802_11_QUERY_RADIO \n");
					wrq->u.data.length = sizeof(pAd->PortCfg.bRadio);
					copy_to_user(wrq->u.data.pointer, &pAd->PortCfg.bRadio, wrq->u.data.length);
					break;
					
				case RTLINUX_RT_OID_802_11_SET_COUNTRY_REGION:
					DBGPRINT(RT_DEBUG_INFO, "INFO::RTLINUX_RT_OID_802_11_SET_COUNTRY_REGION \n");
					copy_from_user(&pAd->PortCfg.CountryRegion, wrq->u.data.pointer, wrq->u.data.length);
					break;
					
				default:
					DBGPRINT(RT_DEBUG_TRACE, "INFO::unknown IOCTL's subcmd = 0x%08x\n", subcmd);
					Status = -EOPNOTSUPP;
					break;
			}
			break;		
			
		default:
			DBGPRINT(RT_DEBUG_TRACE, "INFO::unknown IOCTL's cmd = 0x%08x\n", cmd);
			Status = -EOPNOTSUPP;
			break;
	}

	if(Status == NDIS_STATUS_SUCCESS)
	{
		if((cmd == SIOCDEVPRIVATE) && (subcmd == RTLINUX_GET_OID_802_11_BSSID_LIST))
            kfree(pBssidList);
	}
	if(StateMachineTouched) // Upper layer sent a MLME-related operations
	    MlmeHandler(pAd);
	
	return Status;
}


UCHAR	BCAST[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
/*
	========================================================================
	
	Routine Description:
		Add WPA key process

	Arguments:
		pAdapter						Pointer to our adapter
		pBuf							Pointer to the where the key stored

	Return Value:
		NDIS_SUCCESS					Add key successfully

	Note:
		
	========================================================================
*/
NDIS_STATUS	RTMPWPAAddKeyProc(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuf)
{
	PNDIS_802_11_KEY	pKey;
	ULONG				KeyIdx;
	NDIS_STATUS			Status;
	
	PUCHAR		pTxMic, pRxMic;
	BOOLEAN 	bTxKey; 		// Set the key as transmit key
	BOOLEAN 	bPairwise;		// Indicate the key is pairwise key
	BOOLEAN 	bKeyRSC;		// indicate the receive  SC set by KeyRSC value.
								// Otherwise, it will set by the NIC.
	BOOLEAN 	bAuthenticator; // indicate key is set by authenticator.
	INT 		i, PairwiseIdx;
	
	pKey = (PNDIS_802_11_KEY) pBuf;
	KeyIdx = pKey->KeyIndex & 0xff;
	// Bit 31 of Add-key, Tx Key
	bTxKey		   = (pKey->KeyIndex & 0x80000000) ? TRUE : FALSE;
	// Bit 30 of Add-key PairwiseKey
	bPairwise	   = (pKey->KeyIndex & 0x40000000) ? TRUE : FALSE;
	// Bit 29 of Add-key KeyRSC
	bKeyRSC 	   = (pKey->KeyIndex & 0x20000000) ? TRUE : FALSE;
	// Bit 28 of Add-key Authenticator
	bAuthenticator = (pKey->KeyIndex & 0x10000000) ? TRUE : FALSE;

	// 1. Check Group / Pairwise Key
	if (bPairwise)	// Pairwise Key
	{
		// 1. KeyIdx must be 0, otherwise, return NDIS_STATUS_INVALID_DATA
		if (KeyIdx != 0)
			return(-EINVAL);
		
		// 2. Check bTx, it must be true, otherwise, return NDIS_STATUS_INVALID_DATA
		if (bTxKey == FALSE)
			return(-EINVAL);

		// 3. If BSSID is not all 0xff, return NDIS_STATUS_INVALID_DATA
		if (RTMPEqualMemory(pKey->BSSID, BCAST, 6))
			return(-EINVAL);
			
		// 4. Selct RxMic / TxMic based on Supp / Authenticator
		if (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPANone)
		{
			// for WPA-None Tx, Rx MIC is the same
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pRxMic = pTxMic;
		}
		else if (bAuthenticator == TRUE)
		{
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pRxMic = (PUCHAR) (&pKey->KeyMaterial) + 24;
		}
		else
		{
			pRxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 24;
		}
		
		// 5. Find the old entry to overwrite or find an empty entry.
		PairwiseIdx = 0;
		for (i = 0; i < PAIRWISE_KEY_NO; i++)
		{
			if (pAdapter->PortCfg.PairwiseKey[i].KeyLen == 0)
			{
				PairwiseIdx = i;
				break;
			}
			else if (RTMPEqualMemory(pAdapter->PortCfg.PairwiseKey[i].BssId, pKey->BSSID, 6))
			{
				// Found the old entry
				PairwiseIdx = i;
				break;
			}
		}
		// If there is no match and no empty pairwise key, we have to replace an old one
		// which will be index 0 in our case.

		// 6. Check RxTsc
		if (bKeyRSC == TRUE)
		{
			NdisMoveMemory(&pAdapter->PortCfg.PairwiseKey[PairwiseIdx].RxTsc, &pKey->KeyRSC, 6);			
		}
		else
		{
			NdisZeroMemory(pAdapter->PortCfg.PairwiseKey[PairwiseIdx].RxTsc, 6);		
		}
		
		// 7. Copy information into Pairwise Key structure.
		// pKey->KeyLength will include TxMic and RxMic, therefore, we use 16 bytes hardcoded.
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].KeyLen = 16;		
		NdisMoveMemory(pAdapter->PortCfg.PairwiseKey[PairwiseIdx].Key, &pKey->KeyMaterial, 16);
		NdisMoveMemory(pAdapter->PortCfg.PairwiseKey[PairwiseIdx].RxMic, pRxMic, 8);
		NdisMoveMemory(pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxMic, pTxMic, 8);
		NdisMoveMemory(pAdapter->PortCfg.PairwiseKey[PairwiseIdx].BssId, pKey->BSSID, 6);
		// Init TxTsc to one based on WiFi WPA specs
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[0] = 1;
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[1] = 0;
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[2] = 0;
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[3] = 0;
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[4] = 0;
		pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxTsc[5] = 0;
		Status = NDIS_STATUS_SUCCESS;
		
		DBGPRINT(RT_DEBUG_INFO, "TKIP Key = ");
		for (i = 0; i < 16; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.PairwiseKey[PairwiseIdx].Key[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP Rx MIC Key = ");
		for (i = 0; i < 8; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.PairwiseKey[PairwiseIdx].RxMic[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP Tx MIC Key = ");
		for (i = 0; i < 8; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.PairwiseKey[PairwiseIdx].TxMic[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP RxTSC = ");
		for (i = 0; i < 6; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.PairwiseKey[PairwiseIdx].RxTsc[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");
		DBGPRINT(RT_DEBUG_INFO, "BSSID:%02x:%02x:%02x:%02x:%02x:%02x \n",
			pKey->BSSID[0],pKey->BSSID[1],pKey->BSSID[2],pKey->BSSID[3],pKey->BSSID[4],pKey->BSSID[5]);
		
	}
	else	// Group Key
	{
		// 1. Check BSSID, if not current BSSID or Bcast, return NDIS_STATUS_INVALID_DATA
		if ((!RTMPEqualMemory(&pKey->BSSID, &BCAST, 6)) &&
			(!RTMPEqualMemory(&pKey->BSSID, &pAdapter->PortCfg.Bssid, 6)))
			return(-EINVAL);

		// 2. Check Key index for supported Group Key
		if (KeyIdx >= GROUP_KEY_NO)
			return(-EINVAL);
		
		// 3. Set as default Tx Key if bTxKey is TRUE
		if (bTxKey == TRUE)
			pAdapter->PortCfg.DefaultKeyId = (UCHAR) KeyIdx;
		
		// 4. Selct RxMic / TxMic based on Supp / Authenticator
		if (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPANone)
		{
			// for WPA-None Tx, Rx MIC is the same
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pRxMic = pTxMic;
		}
		else if (bAuthenticator == TRUE)
		{
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pRxMic = (PUCHAR) (&pKey->KeyMaterial) + 24;
		}
		else
		{
			pRxMic = (PUCHAR) (&pKey->KeyMaterial) + 16;
			pTxMic = (PUCHAR) (&pKey->KeyMaterial) + 24;
		}
		
		// 5. Check RxTsc
		if (bKeyRSC == TRUE)
		{
			NdisMoveMemory(pAdapter->PortCfg.GroupKey[KeyIdx].RxTsc, &pKey->KeyRSC, 6);			
		}
		else
		{
			NdisZeroMemory(pAdapter->PortCfg.GroupKey[KeyIdx].RxTsc, 6);		
		}

		// 6. Copy information into Group Key structure.
		// pKey->KeyLength will include TxMic and RxMic, therefore, we use 16 bytes hardcoded.
		pAdapter->PortCfg.GroupKey[KeyIdx].KeyLen = 16;		
		NdisMoveMemory(pAdapter->PortCfg.GroupKey[KeyIdx].Key, &pKey->KeyMaterial, 16);
		NdisMoveMemory(pAdapter->PortCfg.GroupKey[KeyIdx].RxMic, pRxMic, 8);
		NdisMoveMemory(pAdapter->PortCfg.GroupKey[KeyIdx].TxMic, pTxMic, 8);
		NdisMoveMemory(pAdapter->PortCfg.GroupKey[KeyIdx].BssId, pKey->BSSID, 6);
		// Init TxTsc to one based on WiFi WPA specs
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[0] = 1;
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[1] = 0;
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[2] = 0;
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[3] = 0;
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[4] = 0;
		pAdapter->PortCfg.GroupKey[KeyIdx].TxTsc[5] = 0;
		// 802.1x port control
		pAdapter->PortCfg.PortSecured = WPA_802_1X_PORT_SECURED;
		Status = NDIS_STATUS_SUCCESS;

		DBGPRINT(RT_DEBUG_INFO, "TKIP Key = ");
		for (i = 0; i < 16; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.GroupKey[KeyIdx].Key[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP Rx MIC Key = ");
		for (i = 0; i < 8; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.GroupKey[KeyIdx].RxMic[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP Tx MIC Key = ");
		for (i = 0; i < 8; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.GroupKey[KeyIdx].TxMic[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "TKIP RxTSC = ");
		for (i = 0; i < 6; i++)
		{
			DBGPRINT(RT_DEBUG_INFO, "%02x:", pAdapter->PortCfg.GroupKey[KeyIdx].RxTsc[i]);
		}
		DBGPRINT(RT_DEBUG_INFO, "\n");						
		DBGPRINT(RT_DEBUG_INFO, "BSSID:%02x:%02x:%02x:%02x:%02x:%02x \n",
			pKey->BSSID[0],pKey->BSSID[1],pKey->BSSID[2],pKey->BSSID[3],pKey->BSSID[4],pKey->BSSID[5]);
	
		if (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPANone)
		{
			pAdapter->PortCfg.SharedKey[KeyIdx].KeyLen = (UCHAR) pKey->KeyLength;
			NdisMoveMemory(pAdapter->PortCfg.SharedKey[KeyIdx].Key, &pKey->KeyMaterial, pKey->KeyLength);
		}
	}
	return (Status);
}

/*
	========================================================================
	
	Routine Description:
		Remove WPA Key process

	Arguments:
		pAdapter						Pointer to our adapter
		pBuf							Pointer to the where the key stored

	Return Value:
		NDIS_SUCCESS					Add key successfully

	Note:
		
	========================================================================
*/
NDIS_STATUS	RTMPWPARemoveKeyProc(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuf)
{
	PNDIS_802_11_REMOVE_KEY	pKey;
	ULONG					KeyIdx;
	NDIS_STATUS				Status = NDIS_STATUS_FAILURE;
	
	PUCHAR		pTxMic, pRxMic;
	BOOLEAN 	bTxKey; 		// Set the key as transmit key
	BOOLEAN 	bPairwise;		// Indicate the key is pairwise key
	BOOLEAN 	bKeyRSC;		// indicate the receive  SC set by KeyRSC value.
								// Otherwise, it will set by the NIC.
	BOOLEAN 	bAuthenticator; // indicate key is set by authenticator.
	INT 		i, PairwiseIdx;
	
	pKey = (PNDIS_802_11_REMOVE_KEY) pBuf;
	KeyIdx = pKey->KeyIndex & 0xff;
	// Bit 31 of Add-key, Tx Key
	bTxKey		   = (pKey->KeyIndex & 0x80000000) ? TRUE : FALSE;
	// Bit 30 of Add-key PairwiseKey
	bPairwise	   = (pKey->KeyIndex & 0x40000000) ? TRUE : FALSE;
	// Bit 29 of Add-key KeyRSC
	bKeyRSC 	   = (pKey->KeyIndex & 0x20000000) ? TRUE : FALSE;
	// Bit 28 of Add-key Authenticator
	bAuthenticator = (pKey->KeyIndex & 0x10000000) ? TRUE : FALSE;

	// 1. If bTx is TRUE, return failure information
	if (bTxKey == TRUE)
		return(-EINVAL);

	// 2. Check Pairwise Key
	if (bPairwise)
	{
		// a. If BSSID is broadcast, remove all pairwise keys.
		if (RTMPEqualMemory(&pKey->BSSID, &BCAST, 6))
		{
			for (i = 0; i < PAIRWISE_KEY_NO; i++)
			{
				pAdapter->PortCfg.PairwiseKey[i].KeyLen = 0;
			}
			Status = NDIS_STATUS_SUCCESS;
		}

		// b. If not broadcast, remove the pairwise specified by BSSID
		else
		{
			for (i = 0; i < PAIRWISE_KEY_NO; i++)
			{
				if (RTMPEqualMemory(pAdapter->PortCfg.PairwiseKey[i].BssId, pKey->BSSID, 6))
				{
					pAdapter->PortCfg.PairwiseKey[i].KeyLen = 0;
					Status = NDIS_STATUS_SUCCESS;
					break;
				}
			}
			
		}
		// c. If no pairwise supported, delete Group Key 0.
		//    The will be false since we do support pairwise keys.
	}
	// 3. Group Key
	else
	{		
		// a. If BSSID is broadcast, remove all group keys indexed
		if (RTMPEqualMemory(&pKey->BSSID, &BCAST, 6))
		{
			pAdapter->PortCfg.GroupKey[KeyIdx].KeyLen = 0;
			Status = NDIS_STATUS_SUCCESS;
		}

		// b. If BSSID matched, delte the group key indexed.
		else if (RTMPEqualMemory(pAdapter->PortCfg.GroupKey[KeyIdx].BssId, pKey->BSSID, 6))
		{
			pAdapter->PortCfg.GroupKey[KeyIdx].KeyLen = 0;
			Status = NDIS_STATUS_SUCCESS;
		}
		// c. For WEP compatibility
		if (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPANone)
		{
			pAdapter->PortCfg.SharedKey[KeyIdx].KeyLen = 0;
		}
	}
	
	return (Status);
}


/*
	========================================================================
	
	Routine Description:
		Remove All WPA Keys

	Arguments:
		pAdapter						Pointer to our adapter

	Return Value:
		None

	Note:
		
	========================================================================
*/
VOID	RTMPWPARemoveAllKeys(
	IN	PRTMP_ADAPTER	pAdapter)
{
	INT i;

	// For WPA-None, there is no need to remove it, since WinXP won't set it again after
	// Link up. And it will be replaced if user changed it.
	if (pAdapter->PortCfg.AuthMode == Ndis802_11AuthModeWPANone)
		return;
	
	for (i = 0; i < PAIRWISE_KEY_NO; i++)
	{
		pAdapter->PortCfg.PairwiseKey[i].KeyLen = 0;
	}
	
	for (i = 0; i < GROUP_KEY_NO; i++)
	{
		pAdapter->PortCfg.GroupKey[i].KeyLen = 0;
	}
}


